#include <stdio.h>                      // fprintf()
#include "../pci/pci.h"                 // pci_get_device_handle() pci_get_base_address_n()
#include "../dpmi32/phys.h"             // map_physical() unmap_physical()
#include "../apptel/point.h"



AtmCard::AtmCard( Bit8 card_index0 )
{
    card_index = card_index0;

    // find the PCI card we want
    //
    card_handle = pci_get_device_handle( APPTEL_PCI_622_DEVICE_ID, APPTEL_VENDOR_ID, card_index );
    if( card_handle == RETCODE_NO_MORE_PCI_DEVICES )
    {
        fprintf( stderr, "cannot find vendor ID 0x%.4X, device ID 0x%.4X, index %d\n", APPTEL_VENDOR_ID, APPTEL_PCI_622_DEVICE_ID, card_index );
        init_result = ERROR_POINT_NO_SUCH_CARD;
        return;
    }
    fprintf( stdout, "PCI BIOS handle for POINT card %d is 0x%.4X\n", card_index, card_handle );

    // get the physical addresses of the regions we care about
    //
    amcc_phys_address = pci_get_base_address_n( card_handle, AMCC_BASE_ADDR_REG );
    if( amcc_phys_address == 0 )
    {
        fprintf( stderr, "POINT AMCC is not mapped into host memory!\n" );
        init_result = ERROR_POINT_AMCC_UNMAPPED;
        return;
    }
    fprintf( stdout, "POINT AMCC is mapped into host memory at phys 0x%.8X\n", amcc_phys_address );

    fpga_phys_address = pci_get_base_address_n( card_handle, FPGA_BASE_ADDR_REG );
    if( fpga_phys_address == 0 )
    {
        fprintf( stderr, "POINT FPGA's are not mapped into host memory!\n" );
        init_result = ERROR_POINT_FPGA_UNMAPPED;
        return;
    }
    fprintf( stdout, "POINT TX and RX FPGA's are mapped into host memory at phys 0x%.8X\n", fpga_phys_address );

    slave_burst_phys_address = pci_get_base_address_n( card_handle, SLAVE_BURST_BASE_ADDR_REG );
    if( slave_burst_phys_address == 0 )
    {
        fprintf( stderr, "POINT slave bursting is not mapped into host memory!\n" );
        init_result = ERROR_POINT_SLAVE_BURST_UNMAPPED;
        return;
    }
    fprintf( stdout, "POINT slave bursting is mapped into host memory at phys 0x%.8X\n", slave_burst_phys_address );

    suni_phys_address = pci_get_base_address_n( card_handle, SUNI_BASE_ADDR_REG );
    if( suni_phys_address == 0 )
    {
        fprintf( stderr, "AIM SUNI-622 is not mapped into host memory!\n" );
        init_result = ERROR_POINT_AIM_SUNI622_UNMAPPED;
        return;
    }
    fprintf( stdout, "AIM SUNI-622 is mapped into host memory at phys 0x%.8X\n", suni_phys_address );

    aim_and_cpld_phys_address = pci_get_base_address_n( card_handle, AIM_AND_CPLD_BASE_ADDR_REG );
    if( aim_and_cpld_phys_address == 0 )
    {
        fprintf( stderr, "AIM FPGA and POINT CPLD are not mapped into host memory!\n" );
        init_result = ERROR_POINT_AIM_CPLD_UNMAPPED;
        return;
    }
    fprintf( stdout, "AIM FPGA and POINT CPLD are mapped into host memory at phys 0x%.8X\n", aim_and_cpld_phys_address );

    // get accessibility to the memory regions
    //
    amcc = (Amcc_registers *) map_physical( amcc_phys_address, sizeof(Amcc_registers), &amcc_mem_handle );
    if( amcc == NULL )
    {
        fprintf( stderr, "cannot map %d bytes at phys address 0x%.8X into flat memory for POINT AMCC\n", sizeof(Amcc_registers), amcc_phys_address );
        init_result = ERROR_POINT_CANNOT_MAP_PHYS;
        return;
    }

    both_fpgas = (Fpga_registers *) map_physical( fpga_phys_address, sizeof(Fpga_registers), &fpga_mem_handle );
    if( both_fpgas == NULL )
    {
        fprintf( stderr, "cannot map %d bytes at phys address 0x%.8X into flat memory for POINT TX and RX FPGA's\n", sizeof(Fpga_registers), fpga_phys_address );
        init_result = ERROR_POINT_CANNOT_MAP_PHYS;
        return;
    }

    suni = (Suni_registers *) map_physical( suni_phys_address, sizeof(Suni_registers), &suni_mem_handle );
    if( suni == NULL )
    {
        fprintf( stderr, "cannot map %d bytes at phys address 0x%.8X into flat memory for AIM SUNI-622\n", sizeof(Suni_registers), suni_phys_address );
        init_result = ERROR_POINT_CANNOT_MAP_PHYS;
        return;
    }

    aim_and_cpld = (AimAndCpld_registers *) map_physical( aim_and_cpld_phys_address, sizeof(AimAndCpld_registers), &aim_and_cpld_mem_handle );
    if( both_fpgas == NULL )
    {
        fprintf( stderr, "cannot map %d bytes at phys address 0x%.8X into flat memory for AIM FPGA and POINT CPLD\n", sizeof(AimAndCpld_registers), aim_and_cpld_phys_address );
        init_result = ERROR_POINT_CANNOT_MAP_PHYS;
        return;
    }

    init_result = 0;
}



AtmCard::~AtmCard()
{
    // it is quite unsafe to exit if we never initialized!
    //
    if( init_result )
        return;

    // unmap the physical memory we mapped above
    //
    // warning: must be done in the reverse of the order we allocated!
    //
    unmap_physical( sizeof(AimAndCpld_registers), aim_and_cpld_mem_handle );
    unmap_physical( sizeof(Suni_registers), suni_mem_handle );
    unmap_physical( sizeof(Fpga_registers), fpga_mem_handle );
    unmap_physical( sizeof(Amcc_registers), amcc_mem_handle );
}



