// data structures for PMC/Sierra S/UNI-622 (a.k.a. PM5355) chip
//



// abbreviations used here:
//
// RSOP         receive section overhead processor
// RLOP         receive line overhead processor
// RPOP         receive path overhead processor
// RACP         receive ATM cell processor
// TACP         transmit ATM cell processor
//



#ifndef _SUNI622_H
#define _SUNI622_H



#define SUNI_BASE_ADDR_REG      3



// any write loads all performance meters in
// RSOP, RLOP, RPOP, RACP, and TACP blocks
//
struct SuniMasterResetIdentityAndLoadPerf_register
{
    Bit8 Identity       : 4;                    // SUNI revision number
    Bit8 Type           : 3;                    // 1 means SUNI-622
    Bit8 Reset          : 1;                    // not self-clearing
};



struct SuniRacpControl_register
{
    Bit8 ResetFifo                  : 1;
    Bit8 DisablePayloadDescrambling : 1;
    Bit8 AddCosetPolynomialToHcs    : 1;
    Bit8 PassUncorrectableHcsErrors : 1;
    Bit8 DisableHcsErrorCorrection  : 1;
    Bit8 PassFilteredCells          : 1;        // usually means idle/unassigned
    Bit8 RxParityEven               : 1;
    Bit8 EnableFixedStuffing        : 1;        // for SONET payload
};



struct SuniGfcAndMiscControl_register
{
    Bit8 RxGfcEnable0               : 1;        // copy GFC bit 0 to RGFC output
    Bit8 RxGfcEnable1               : 1;        // copy GFC bit 1 to RGFC output
    Bit8 RxGfcEnable2               : 1;        // copy GFC bit 2 to RGFC output
    Bit8 RxGfcEnable3               : 1;        // copy GFC bit 3 to RGFC output
    Bit8 Unused                     : 2;
    Bit8 RxByteParity               : 1;        // RXPRTY[1:0] is byte if 1, word if 0
    Bit8 CellDelineationDisable     : 1;        // use 1 for full SONET payload access (PPP mode)
};



// merged registers
//
struct Suni_registers
{
    union
    {
        SuniMasterResetIdentityAndLoadPerf_register Master;
        Bit32 dword_0;
    };

    Bit32 Reserved1[ 79 ];

    union
    {
        SuniRacpControl_register RacpControl;
        Bit32 dword_80;
    };

    Bit32 Reserved2[ 11 ];

    union
    {
        SuniGfcAndMiscControl_register GfcAndMiscControl;
        Bit32 dword_92;
    };

    Bit32 Reserved3[ 163 ];
};



#endif      // not included yet
