/*
data structures which describe registers common to the
Xilinx Field Programmable Gate Arrays (FPGA's) attached
to the receive and transmit sides of the Applied Telecom
AIM-622 OC12c ATM/PPP optics module.



Note 1: these data structures use bit fields, which are compiler-
and Central Processing Unit (CPU) specific, so you will probably
have to rewrite these data structures if you port to another
CPU or compiler

On the Intel CPU's using the Borland C++ 4.5 compiler, bits are
assigned from least significant to most significant within a word,
and less significant bytes of a word are stored at lower memory
addresses

No padding is inserted, and bitfields can straddle word boundaries

The width of each bitfield must be from 1 to 16 (see page 62 of the
Borland C++ version 4.5 Programmer's Guide).



Note 2: due to the use of anonymous unions, this header can only be
included from C++ code, not C code



Note 3: definitions of readability/writability abbreviations:
    W/O     writable only
    R/O     readable only
    W/R     writable and readable
    CW/R    readable, but writable only after
            CWRE bit set in Control_register
*/



#ifndef _XFPGA_H
#define _XFPGA_H



#include "../types/types.h"



// just a scratchpad to read from or write to during early
// hardware testing
//
struct Diagnostic_register
{
    Bit8 Value;
};



struct AddressXor_register
{
    // hold the XOR of address bits 9:2 for the previous
    // passthru read or write that selected the RX FPGA
    //
    // used for early hardware testing
    //
    Bit8 Value              : 1;

    Bit8 Reserved           : 7;
};



#endif      // not included yet
