/* conditional compile switches all go here
 *
 * manifest constants are still sprinkled
 * all through the code (sorry)
 */
#define ALWAYS_DUMP_CELL_COUNT  FALSE
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define DEBUG_CHAIN_LENGTH      FALSE
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_COPY_DATA         FALSE           // stupid debugger!
#define TALK_WHEN_ALLOC_FAILS   FALSE           // for debugging when overflow occurs only!  too slow
#define CLEAN_IFACES_TOGETHER   FALSE
#define QUERY_CALLS_EXPIRE      FALSE           // screws up offline analysis, redundant when caller gives us partial blocks at query time
#define HASHINDEX_PARANOIA      FALSE
#define CELLS_NEW_PARANOIA      FALSE
#define CELLS_EXPIRE_PARANOIA   FALSE
#define CELLS_CLEANUP_PARANOIA  TRUE
#define USE_CELL_COUNT_FIRST    TRUE            // zero count signals start of packet
#define USE_CELL_COUNT_LAST     TRUE            // max  count signals end of packet if card won't also send last cell
#define AVERAGE_OF_AVERAGES     FALSE
#define SHOW_CPU_UTILIZATION    FALSE
#define DELTA_TIME_STATS        FALSE           // checks whether we housecleaned on time
#define BLOCK_USAGE_STATS       TRUE
#define COUNT_THINGS            TRUE
#define USE_EXPIRATION          TRUE






#include <sys/types.h>

#if __unix__
#   include <sys/time.h>
#   include <sys/resource.h>
#endif
#include "../wattcp/tcp.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <float.h>              /* DBL_MIN DBL_MAX */

#include "../types/types.h"
#if !__unix__
#   include "..\intel\timestmp.h"
#   include "..\intel\byteswap.h"
#   include "..\intel\fpu.h"    /* get_fpu_status() clear_fpu_status() */
#endif

#include "..\integrat\buffer.h"             /* class BufferBlock */
#include "..\integrat\fs2flows.h"           /* make sure we comply with our own prototypes */
#include "..\integrat\aal5.h"               /* our own datatypes and prototypes */
#include "..\integrat\dostats.h"            /* nonzero() */
#include "..\integrat\dump.h"               /* hex_dump_char() */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



/*
 * general hashing routines
 *
 */


#define NPKTBUCKETS    256
#define PKTBUCKET_POOL_SIZE (NPKTBUCKETS * sizeof(struct cellhashrec *))
static struct cellhashrec **celltab = NULL;
static struct cellhashrec **celltab_tails = NULL;

#if DEBUG_HASHINDEX
    u_int hashindex_trace = 102;
#endif


#if DEBUG_CHAIN_LENGTH
#   define CHAIN_POOL_SIZE (NPKTBUCKETS * sizeof(int))
    static int *chainlen = NULL;
    static int longest_chain = 0;
    static u_int chainlen_sum = 0;
    static int chains_used = 0;
#endif

/* generate an 8-bit hash key
 */
#define cellhash(tr,if) fold_32_to_8( NNI_VPI_VCI( tr->atm.whole ) )


#define CELLHASH_RECORDS 100000L        // how long will this last at WTN?
#define CELLHASH_POOL_SIZE (CELLHASH_RECORDS * sizeof(struct cellhashrec))
static struct cellhashrec *hash_free = NULL;
static struct cellhashrec *hash_pool = NULL;
static u_int hash_alloc_failures;


#if COUNT_THINGS
    static u_int created_cells_trace = -1;
    static u_int prev_created_cells[ MAX_ATM_NICS ];
    static u_int created_cells[ MAX_ATM_NICS ];
    static u_int deleted_cells[ MAX_ATM_NICS ];
    static u_int cleaned_cells[ MAX_ATM_NICS ];
    static u_int sum_pending_cells[ MAX_ATM_NICS ];
    static u_int max_cleaned_cells[ MAX_ATM_NICS ];
    static u_int max_pending_cells[ MAX_ATM_NICS ];
    static u_int max_new_cells[ MAX_ATM_NICS ];
#   if AVERAGE_OF_AVERAGES
        static u_int sum_new_cells[ MAX_ATM_NICS ];
#   endif

    static u_int prev_created_packets[ MAX_ATM_NICS ];
    static u_int created_packets[ MAX_ATM_NICS ];
    static u_int deleted_packets[ MAX_ATM_NICS ];
    static u_int cleaned_packets[ MAX_ATM_NICS ];
    static u_int sum_pending_packets[ MAX_ATM_NICS ];
    static u_int max_cleaned_packets[ MAX_ATM_NICS ];
    static u_int max_pending_packets[ MAX_ATM_NICS ];
    static u_int max_new_packets[ MAX_ATM_NICS ];
#   if AVERAGE_OF_AVERAGES
        static u_int sum_new_packets[ MAX_ATM_NICS ];
#   endif

    static u_int max_cells_per_packet[ MAX_ATM_NICS ];
#endif  // COUNT_THINGS

#if BLOCK_USAGE_STATS
    static Bit32 max_pending_blocks[ MAX_ATM_NICS ];
    static Bit32 sum_pending_blocks[ MAX_ATM_NICS ];
    static Bit32 max_cleaned_blocks[ MAX_ATM_NICS ];
    static Bit32 sum_cleaned_blocks[ MAX_ATM_NICS ];
#endif

#if COUNT_THINGS || BLOCK_USAGE_STATS || DELTA_TIME_STATS
    static u_int cleaning_count[ MAX_ATM_NICS ];
#endif

static CardClock last_cell_time[ MAX_ATM_NICS ];
static CardClock last_query_time;
static int is_query_set;

// housecleaning control structures
//
#define DEFAULT_HOUSECLEAN_INTERVAL     0.25   /* one quarter of a second */
#define DEFAULT_PKT_ACCESS_TIMEOUT      1.0    /* one whole second */
static CardClock houseclean_alarm[ MAX_ATM_NICS ];
static CardClock last_houseclean_time[ MAX_ATM_NICS ];
static CardClock houseclean_interval = DEFAULT_HOUSECLEAN_INTERVAL * CARD_CLOCK_RATE;
static CardClock packet_access_timeout = DEFAULT_PKT_ACCESS_TIMEOUT * CARD_CLOCK_RATE;
static int is_alarm_set[ MAX_ATM_NICS ];
#if DELTA_TIME_STATS
    static double min_houseclean_interval;
    static double max_houseclean_interval;
    static double sum_houseclean_interval;
#endif
static void cells_expire( tracerec *record, Bit8 interface );

#if SHOW_CPU_UTILIZATION
    static PentiumClock before, elapsed;
    static PentiumClock startup;
    static PentiumClock min_elapsed, max_elapsed, sum_elapsed;
#endif




// clears all statistics during initialization of module
// or during query
//
int cells_clear( void )
{
    int result;
    unsigned interface;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        is_alarm_set[ interface ] = FALSE;

#if COUNT_THINGS
        prev_created_cells[ interface ] -= deleted_cells[ interface ];
        created_cells[ interface ] -= deleted_cells[ interface ];
        deleted_cells[ interface ] = 0;
        cleaned_cells[ interface ] = 0;
        sum_pending_cells[ interface ] = 0;
        max_cleaned_cells[ interface ] = 0;
        max_pending_cells[ interface ] = 0;
        max_new_cells[ interface ] = 0;
#if AVERAGE_OF_AVERAGES
        sum_new_cells[ interface ] = 0;
#endif

        prev_created_packets[ interface ] -= deleted_packets[ interface ];
        created_packets[ interface ] -= deleted_packets[ interface ];
        deleted_packets[ interface ] = 0;
        cleaned_packets[ interface ] = 0;
        sum_pending_packets[ interface ] = 0;
        max_cleaned_packets[ interface ] = 0;
        max_pending_packets[ interface ] = 0;
        max_new_packets[ interface ] = 0;
#if AVERAGE_OF_AVERAGES
        sum_new_packets[ interface ] = 0;
#endif

        max_cells_per_packet[ interface ] = 0;
#endif  // COUNT_THINGS

#if BLOCK_USAGE_STATS
        max_pending_blocks[ interface ] = 0;
        sum_pending_blocks[ interface ] = 0;
        max_cleaned_blocks[ interface ] = 0;
        sum_cleaned_blocks[ interface ] = 0;
#endif

#if COUNT_THINGS || BLOCK_USAGE_STATS || DELTA_TIME_STATS
        cleaning_count[ interface ] = 0;
#endif
    }

#if DELTA_TIME_STATS
    sum_houseclean_interval = 0;
    max_houseclean_interval = DBL_MIN;
    min_houseclean_interval = DBL_MAX;
#endif

#if SHOW_CPU_UTILIZATION
    sum_elapsed = Bit32( 0 );
    max_elapsed = Bit32( 0 );
    min_elapsed.high = -1;
    min_elapsed.low = -1;
    startup.set();
#endif

    hash_alloc_failures = 0;
    is_query_set = FALSE;

    return 0;
}



int cells_init( void )
{
    unsigned count;
    unsigned interface;
    CardClock timeout;

    timeout = packet_access_timeout;
    timeout.strip_bitfields();
    printf( "packets time out when accessed >%.1f seconds ago\n", timeout.get_seconds() );

    /* allocate buckets to point at head of each hash chain */
    celltab = (struct cellhashrec **) flat_alloc( PKTBUCKET_POOL_SIZE );
    if( celltab == NULL )
    {
        printf( "error %d allocating %d bytes (%d pointers) for packet hash bucket heads\n", errno, PKTBUCKET_POOL_SIZE, NPKTBUCKETS );
        return 1;
    }

    /* allocate buckets to point at tail of each hash chain */
    celltab_tails = (struct cellhashrec **) flat_alloc( PKTBUCKET_POOL_SIZE );
    if( celltab_tails == NULL )
    {
        printf( "error %d allocating %d bytes (%d pointers) for packet hash bucket tails\n", errno, PKTBUCKET_POOL_SIZE, NPKTBUCKETS );
        return 1;
    }

    /* nothing in any of the buckets yet
     */
    memset( celltab, 0, PKTBUCKET_POOL_SIZE );
    memset( celltab_tails, 0, PKTBUCKET_POOL_SIZE );

#if DEBUG_CHAIN_LENGTH
    /* allocate chain length counters for each bucket */
    chainlen = (int *) flat_alloc( CHAIN_POOL_SIZE );
    if( chainlen == NULL )
    {
        printf( "error %d allocating %d bytes (%d integers) for chain length counters\n", errno, CHAIN_POOL_SIZE, NPKTBUCKETS );
        return 1;
    }

    /* initialize chain length of all hashtable buckets to zero
     */
    for( count=0; count<NPKTBUCKETS; count++ )
        chainlen[ count ] = 0;
#endif

    /* allocate a pool of hash records */
    hash_pool = (struct cellhashrec *) flat_alloc( CELLHASH_POOL_SIZE );
    if( hash_pool == NULL )
    {
        printf( "error %d allocating %d bytes (%d records) for cell hash record pool\n", errno, CELLHASH_POOL_SIZE, CELLHASH_RECORDS );
        return 1;
    }

    /* put all the hash records on the singly-linked free list */
    hash_free = hash_pool;
    for( count=0; count<CELLHASH_RECORDS-1; count++ )
    {
        hash_pool[ count ].next = hash_pool + count + 1;
    }
    hash_pool[count].next = NULL;

    houseclean_interval.insert_bitfields( -1 );
    packet_access_timeout.insert_bitfields( -1 );
#if COUNT_THINGS
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        deleted_cells[ interface ] = 0;
        deleted_packets[ interface ] = 0;
    }
#endif

    return cells_clear();
}



static int cells_compare( struct cellhashrec *hr, struct tracerec *tr, Bit8 interface )
{
#if DEBUG_COPY_DATA
    cellhashrec hr2 = *hr;
    tracerec_last known_cell = *hr->last;
    tracerec cell = *tr;
    cell = *hr->first;
#endif

    return hr->interface == interface
        && hr->last->atm.bits.VCI==tr->atm.bits.VCI
        && hr->last->atm.bits.VPI==tr->atm.bits.VPI;
}



static int cells_delete_cells( cellhashrec *hr, BufferBlock *block )
{
    tracerec *record;
    tracerec *next;
    int result;
#if COUNT_THINGS
    Bit8 interface;
    u_int cells_per_packet = 0;
#endif

    // for all the trace records in this packet
    //
    for( record=hr->first; TRUE; record=next )
    {
        next = (tracerec *)record->next;

        if( !block )
            block = all_buffers->block_from_record( record );

#if COUNT_THINGS
        interface = block->cardindex;
        deleted_cells[ interface ]++;
        cells_per_packet++;
#endif

        block->freed++;

        // we cannot compare freed count to used count while block
        // is still being filled by the card, since used count is
        // still increasing
        //
        // if the block got onto the filled list and we somehow still
        // got called for the last cell, if we wanted to take care of
        // freeing it here, we would need to be able to get it off of
        // th filled list, but that is impossible in the general case,
        // since that list is singly-linked and the block might be
        // neither at the head nor the tail
        //
        // and if it is already on the freed list, we obviously have
        // nothing to do (although something is fouled up for us to
        // be called with such a block in the first place)
        //
        if( block->state==is_consuming && block->freed==block->used )
        {
            result = all_buffers->free( block );
            if( result )
                return result;
        }

        if( record == (tracerec *)hr->last )
            break;

        block = NULL;
    }

#if COUNT_THINGS
    if( max_cells_per_packet[ interface ] < cells_per_packet )
        max_cells_per_packet[ interface ] = cells_per_packet;
#endif

    return 0;
}



/*
 *  decrement chain length for that hash index
 *  and delete an entry from the hash table
 *
 *  assumes cells_delete_cells() has already been called
 */
static void cells_delete_packet( u_int indx, cellhashrec *hr, cellhashrec *prev, Bit8 interface )
{
#if DEBUG_CHAIN_LENGTH
    chainlen[indx]--;
    chainlen_sum--;
    if( chainlen[indx] < 0 )
    {
        fprintf( stderr, "warning: %d chain length for hash index %d\n", chainlen[indx], indx );
    }

#if DEBUG_HASHINDEX
    if( indx == hashindex_trace )
        fprintf( stderr, "hash index %d chain length decreased to %d\n", indx, chainlen[indx] );
#endif
#endif

    // take this packet off the collision-resolution list
    //
    if( !hr->next )
        celltab_tails[indx] = prev;
    if( prev != NULL )
        prev->next = hr->next;
    else
        celltab[indx] = hr->next;

    // put this packet on the free list
    //
    hr->next = hash_free;
    hash_free = hr;

#if COUNT_THINGS
    deleted_packets[ interface ]++;
#endif
}



static int cells_summarize( cellhashrec *hr, Bit8 interface )
{
    int result;
#if SHOW_CPU_UTILIZATION
    PentiumClock middle;
#endif

#if SHOW_CPU_UTILIZATION
    middle.set();
#endif

    if( hr->first == (tracerec *)hr->last )
        hr->second = NULL;
    else
        hr->second = hr->first->next;

    result = flows_per_packet( hr, interface );

#if SHOW_CPU_UTILIZATION
    // don't count flows time as reassembly
    //
    elapsed.set();
    elapsed -= middle;
    before += elapsed;
#endif

    cells_delete_cells( hr, NULL );

    return result;
}



// called when program is about to exit
// to walk through still-pending cells
// and force them to be accumulated statistically
//
int cells_cleanup( void )
{
    u_int indx;
    cellhashrec *hr;
    cellhashrec *next;

#if DEBUG_CHAIN_LENGTH
    fprintf( stderr, "chain length is %d max, %.3f avg\n", longest_chain, chainlen_sum/double(chains_used) );
    fprintf( stderr, "%u (%.1f%%) of %u buckets used\n", chains_used,
        (100.0*chains_used)/NPKTBUCKETS, NPKTBUCKETS );
#endif

#if CELLS_CLEANUP_PARANOIA
    int result;

    result = cells_check_all_packets( stdout );
    if( result )
    {
        cells_dump_all_packets( stdout );
        return result;
    }
#endif

    for( indx=0; indx<NPKTBUCKETS; indx++ )
    {
        for( hr=celltab[indx]; hr; hr=next )
        {
            cells_summarize( hr, hr->interface );

            next = hr->next;

            // always pass NULL as prev hashrecord pointer,
            // because we start at the head of the list and
            // delete every one, so the one we are at now
            // must always be the head of the list
            //
            cells_delete_packet( indx, hr, NULL, hr->interface );
        }
    }

    return 0;
}



/*
 *  establish a new packet
 *  1> insert into hash table [maintained as stack]
 *  2> increment chain length
 */

static int cells_new_packet( u_int indx, tracerec *rp, Bit8 interface,
    cellhashrec **hr0, cellhashrec **prev )
{
    struct cellhashrec *hr;
    u_int pending_packets;

    // allocate a hash record for the packet
    //
    hr = hash_free;
    if( hr == NULL )
    {
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new cellhashrec\n");
#endif
        hash_alloc_failures++;
        return 2;
    }
    hash_free = hr->next;
    *hr0 = hr;

#if COUNT_THINGS
    created_packets[ interface ]++;
    pending_packets = created_packets[ interface ] - deleted_packets[ interface ];
    sum_pending_packets[ interface ] += pending_packets;
    if( max_pending_packets[ interface ] < pending_packets )
        max_pending_packets[ interface ] = pending_packets;
#endif

#if DEBUG_CHAIN_LENGTH
    /*
     * increment chain length for this hash index
     */
    chainlen[indx]++;
    chainlen_sum++;

#if DEBUG_HASHINDEX
    if( indx == hashindex_trace )
        fprintf( stderr, "hash index %d chain length increased to %d\n", indx, chainlen[indx] );
#endif
#endif

#if CELLS_NEW_PARANOIA
    if( rp == NULL )
    {
        fprintf( stderr, "error!  cells_new_packet( NULL )\n" );
        return 123;
    }
#endif

    // there is only one cell in this packet so far
    //
    hr->first = rp;
    hr->last = (tracerec_last *) hr->first;
    hr->interface = interface;

    hr->next = NULL;
    *prev = celltab_tails[ indx ];

    /*
     * handle the case where no cells existed
     * on this hash chain until now
     */
    if( celltab[ indx ] == NULL )
    {
#if DEBUG_CHAIN_LENGTH
        chains_used++;
#endif
        celltab[ indx ] = hr;
        celltab_tails[ indx ] = hr;
        return 0;
    }

    /*
     * at least one packet already there (we collided earlier)
     * insert at the tail [this maintains a stack]
     */

    celltab_tails[indx]->next = hr;
    celltab_tails[indx] = hr;

#if DEBUG_CHAIN_LENGTH
    /*
     * if this is the longest chain we have ever seen,
     * update our global, and notify the user
     *
     * notifications should be asymptotically decreasing
     */
    if( longest_chain < chainlen[indx] )
    {
        longest_chain = chainlen[indx];
        fprintf( stderr, "hash index %d is now longest chain at %d\n", indx, longest_chain );
        fprintf( stderr, "calced from iface=%d VPI=%d VCI=%d\n", interface, rp->atm.bits.VPI, rp->atm.bits.VCI );
        fprintf( stderr, "average chain length is %.3f\n", chainlen_sum/double(chains_used) );
        fprintf( stderr, "%u (%.1f%%) of %u buckets used\n", chains_used,
            (100.0*chains_used)/NPKTBUCKETS, NPKTBUCKETS );
    }
#endif

    return 0;
}



#if SHOW_CPU_UTILIZATION
PentiumClock cells_get_last_reassemble_timing( void )
{
    return elapsed;
}
#endif



int cells_reassemble( struct tracerec *record, Bit8 interface )
{
    u_int hashindex;
    u_int pending_cells;
    struct cellhashrec *hr;
    struct cellhashrec *prev;
    struct cellhashrec *next;
    int result;

#if SHOW_CPU_UTILIZATION
    // start the timer of cell processing
    //
    before.set();
#endif

#if COUNT_THINGS
    created_cells[ interface ]++;
    pending_cells = created_cells[ interface ] - deleted_cells[ interface ];
    sum_pending_cells[ interface ] += pending_cells;
    if( max_pending_cells[ interface ] < pending_cells )
        max_pending_cells[ interface ] = pending_cells;
#endif

#if !CARD_SWAPS_ATM_HEADER
    // byteswap the ATM header
    //
    record->atm.whole = ntohl( record->atm.whole );
#endif

#if (CELL_COUNT_MAX == 0) && !CARD_COPIES_LAST_CELL
    cellhashrec dummy;
    dummy.first = record;
    dummy.last = (tracerec_last *) record;
    dummy.interface = interface;
    result = cells_summarize( &dummy, interface );
#else
    // time out old packets if necessary
    //
    cells_expire( record, interface );

    hashindex = cellhash( record, interface );

#if HASHINDEX_PARANOIA
    if( hashindex >= NPKTBUCKETS )
        printf( "internal error: cellhash %u > %d\n", hashindex, NPKTBUCKETS );
#endif

#if ALWAYS_DUMP_CELL_COUNT
    fprintf( stderr, " %d", created_cells[interface] );
#endif

#if !ALWAYS_DUMP_HASH_INFO && COUNT_THINGS
    if( created_cells[ interface ] == created_cells_trace )
#endif
#if ALWAYS_DUMP_HASH_INFO || COUNT_THINGS
        fprintf( stderr, "cell %5d hash index %3d to iface=%d VPI=%-3d VCI=%d\n",
            created_cells[interface], hashindex, interface,
            record->atm.bits.VPI, record->atm.bits.VCI );
#endif

    /* look for match or collision on hash chain
     */
    for( hr=celltab[hashindex],prev=NULL; hr; prev=hr,hr=next )
    {
        next = hr->next;

        // if we found an existing VPI/VCI/card...
        //
        int cmp = cells_compare( hr, record, interface );
        if( cmp )
        {
#if USE_CELL_COUNT_FIRST
            // if this is the first cell on the VPI/VCI...
            //
            if( (record->when.low & CELL_COUNT_MASK) == 0 )
            {
                // there will be no more cells for the old packet,
                // so hand it off to the statistics module
                //
                cells_summarize( hr, interface );

                // now reset the fields we won't touch below
                //
                hr->first = record;
            }
            else
#endif
                // chain the previously-last record to this one
                //
                hr->last->next = (tracerec_2 *) record;

            // update the fields of the old record
            //
            hr->last = (tracerec_last *)record;

            // move the hash record to the tail of the list,
            // so it will be easier for the housecleaner to
            // know which are old
            //

            // if not already the tail...
            //
            if( hr->next )
            {
#if DEBUG_HASHINDEX
                if( hashindex == hashindex_trace )
                    fprintf( stderr, "hash index %d being altered\n", hashindex );
#endif

                // remove from old spot
                //
                if( prev )
                    prev->next = hr->next;
                else
                    celltab[ hashindex ] = hr->next;

                // insert at tail
                //

                // remember who is currently the last,
                // in case we need to delete below
                //
                prev = celltab_tails[ hashindex ];

                // no need to check list tail pointer for NULL, since
                // there must be some other packet on the list,
                // since we would have been the tail if we were
                // alone on that list before
                //
                prev->next = hr;
                celltab_tails[ hashindex ] = hr;

                // make sure our next point is NULL,
                // to indicate (redundantly) that we
                // are the tail of the list
                //
                hr->next = NULL;

                // no need to change the head pointer for reinsertion,
                // since we know there is somebody else on the list,
                // since we were not the tail when this all started
            }

            result = 0;
            break;
        }
    }

    if( !hr )
        result = cells_new_packet( hashindex, record, interface, &hr, &prev );

    // if this is the last cell the card is allowed
    // to give us, or truly the last cell...
    //
    if(
#if USE_CELL_COUNT_LAST && !CARD_COPIES_LAST_CELL
        (record->when.low & CELL_COUNT_MASK) >= CELL_COUNT_MAX ||
#endif
        record->atm.bits.LAST
    )
    {
        // there will be no more cells for the current packet,
        // so hand it off to the statistics module
        //
        cells_summarize( hr, interface );

        // we don't have an immediate use for the hash record,
        // so unlink it
        //
        cells_delete_packet( hashindex, hr, prev, interface );
    }
#endif  // (CELL_COUNT_MAX != 0) || CARD_COPIES_LAST_CELL

#if SHOW_CPU_UTILIZATION
    // stop the timer of cell processing
    //
    elapsed.set();

    // update the cell processing time statistics
    //
    elapsed -= before;
    sum_elapsed += elapsed;
    if( max_elapsed < elapsed )
        max_elapsed = elapsed;
    if( min_elapsed > elapsed )
        min_elapsed = elapsed;
#endif

    return result;
}



// called every time a raw cell is being processed
//
// only triggers a timeout based on the current cell's
// time versus the alarm time for the current interface,
// because we cannot know what the current time is on
// any other interfaces
//
// we could have remembered the last time a cell was
// seen on other interfaces, but we would have triggered
// a timeout when we saw that other cell, so why repeat
// that effort now?
//
// only times out packets for the current interface,
// for the same reasons as above
//
// expects the collision-resolution list through the packet
// hash records to also keep recently-accessed packets at
// the tail, so we can do early-out here
//
static void cells_expire( tracerec *record, Bit8 interface )
{
    cellhashrec *hr;
    cellhashrec *prev;
    cellhashrec *next;
    u_int indx;
    u_int deleted_packets_before[ MAX_ATM_NICS ];
    u_int deleted_cells_before[ MAX_ATM_NICS ];
    u_int cleaned_packets_now;
    u_int cleaned_cells_now;
    u_int new_cells;
    u_int new_packets;
    Bit8 temp_interface;
    CardClock temp_card_time;
    CardClock this_cell_time;
    double new_houseclean_interval;
#if BLOCK_USAGE_STATS
    Bit32 filled_blocks, free_blocks;
    Bit32 new_blocks, after_blocks;
#endif

    this_cell_time = record->when;
#if CELLS_EXPIRE_PARANOIA
    if( last_cell_time[ interface ] >= this_cell_time )
        fprintf( stderr, "iface %d prev time 0x%.8X%.8X >= curr 0x%.8X%.8X\n",
            interface,
            last_cell_time[ interface ].high,
            last_cell_time[ interface ].low,
            this_cell_time.high,
            this_cell_time.low
        );
#endif
    last_cell_time[ interface ] = this_cell_time;
    if( !is_alarm_set[ interface ] )
    {
        is_alarm_set[ interface ] = TRUE;
        houseclean_alarm[ interface ] = this_cell_time + houseclean_interval;
        last_houseclean_time[ interface ] = this_cell_time;
        if( !is_query_set )
        {
            is_query_set = TRUE;
            last_query_time = this_cell_time;
        }
        return;
    }
    if( this_cell_time < houseclean_alarm[ interface ] )
        return;

#if COUNT_THINGS || DELTA_TIME_STATS
    // calculate how long it has been since our last housecleaning
    //
    temp_card_time = this_cell_time - last_houseclean_time[ interface ];
    temp_card_time.strip_bitfields();
    new_houseclean_interval = temp_card_time.get_seconds();
#endif

#if DELTA_TIME_STATS
    if( max_houseclean_interval < new_houseclean_interval )
        max_houseclean_interval = new_houseclean_interval;

    if( min_houseclean_interval > new_houseclean_interval )
        min_houseclean_interval = new_houseclean_interval;

    sum_houseclean_interval += new_houseclean_interval;
#endif

#if COUNT_THINGS || DELTA_TIME_STATS
    // remember when we did this housecleaning
    //
    last_houseclean_time[ interface ] = this_cell_time;
#endif

#if COUNT_THINGS
    new_cells = (created_cells[interface] - prev_created_cells[interface])
                / new_houseclean_interval;
    prev_created_cells[ interface ] = created_cells[ interface ];
    if( max_new_cells[ interface ] < new_cells )
        max_new_cells[ interface ] = new_cells;
#if AVERAGE_OF_AVERAGES
    sum_new_cells[ interface ] += new_cells;
#endif

    new_packets = (created_packets[interface] - prev_created_packets[interface])
                / new_houseclean_interval;
    prev_created_packets[ interface ] = created_packets[ interface ];
    if( max_new_packets[ interface ] < new_packets )
        max_new_packets[ interface ] = new_packets;
#if AVERAGE_OF_AVERAGES
    sum_new_packets[ interface ] += new_packets;
#endif

    // remember how many packets are deleted before starting
    // cleanup, so we can calculate afterwards how many were
    // due to our efforts
    //
    for( temp_interface=0; temp_interface<MAX_ATM_NICS; temp_interface++ )
    {
        deleted_packets_before[ temp_interface ] = deleted_packets[ temp_interface ];
        deleted_cells_before[ temp_interface ] = deleted_cells[ temp_interface ];
    }
#endif //COUNT_THINGS

#if BLOCK_USAGE_STATS
    // we are calculating pending blocks as
    //     total - free - filled
    // but really that also counts the 2 blocks
    // that each card is in the process of filling
    // at any given moment
    //
    // the only way around this would be to detect
    // whether each cell coming in belongs to a new
    // block or not, and increment per-card counters,
    // but I would rather not call block_from_buffer()
    // every time a cell comes in...
    //
    all_buffers->amount_filled( filled_blocks );
    all_buffers->amount_free( free_blocks );
    new_blocks = SW_RINGS - filled_blocks - free_blocks;
    sum_pending_blocks[ interface ] += new_blocks;
    if( max_pending_blocks[ interface ] < new_blocks )
        max_pending_blocks[ interface ] = new_blocks;
#endif

#if COUNT_THINGS || BLOCK_USAGE_STATS || DELTA_TIME_STATS
#if CLEAN_IFACES_TOGETHER
    for( temp_interface=0; temp_interface<MAX_ATM_NICS; temp_interface++ )
        cleaning_count[ temp_interface ]++;
#else
    cleaning_count[ interface ]++;
#endif
#endif

    // for all hash buckets...
    //
    for( indx=0; indx<NPKTBUCKETS; indx++ )
    {
        // for all packets in this bucket...
        //
        for( prev=NULL,hr=celltab[indx]; hr; hr=next )
        {
            next = hr->next;

#if !CLEAN_IFACES_TOGETHER
            // if this isn't the interface we care about...
            //
            if( hr->interface != interface )
            {
                // since we didn't delete this hash record,
                // it now becomes the previous
                //
                prev = hr;

                // ignore the packet
                //
                continue;
            }
#endif

            // if this packet saw cells recently...
            //
            if( this_cell_time <= hr->last->when + packet_access_timeout )

                // early out, because we have already seen any
                // less recently accessed cells, since they were
                // sorted to front of this list
                //
                break;

            // time to expire a packet
            //
            cells_summarize( hr, hr->interface );

            // move this packet from its collision-resolution list
            // to the free list
            //
            cells_delete_packet( indx, hr, prev, hr->interface );

            // do NOT assign to pointer to previous hashrecord,
            // since whoever was previous to this hash record is
            // now previous to the next hash record, since we just
            // deleted this puppy!
            //
        }
    }

#if COUNT_THINGS
    // calculate how many packets were deleted due to our efforts
    //
    for( temp_interface=0; temp_interface<MAX_ATM_NICS; temp_interface++ )
    {
        cleaned_packets_now = deleted_packets[ interface ] - deleted_packets_before[ interface ];
        cleaned_packets[ interface ] += cleaned_packets_now;
        if( max_cleaned_packets[ interface ] < cleaned_packets_now )
            max_cleaned_packets[ interface ] = cleaned_packets_now;

        cleaned_cells_now = deleted_cells[ interface ] - deleted_cells_before[ interface ];
        cleaned_cells[ interface ] += cleaned_cells_now;
        if( max_cleaned_cells[ interface ] < cleaned_cells_now )
            max_cleaned_cells[ interface ] = cleaned_cells_now;
    }

    // don't bother to keep separate counters for cells and
    // packets pending; just use the created_xxx counters
    // at query time
    //

    // note that we calculate max_pending_cells as cells are seen,
    // and we calculate max_pending_packets as packets are started,
    // not here, since those aren't per-second measures and we want
    // them to reflect the true high-water marks, not sampled ones
    //
#endif //COUNT_THINGS

#if BLOCK_USAGE_STATS
    all_buffers->amount_filled( filled_blocks );
    all_buffers->amount_free( free_blocks );
    after_blocks = new_blocks - (SW_RINGS - filled_blocks - free_blocks);
    sum_cleaned_blocks[ interface ] += after_blocks;
    if( max_cleaned_blocks[ interface ] < after_blocks )
        max_cleaned_blocks[ interface ] = after_blocks;
#endif

#if CLEAN_IFACES_TOGETHER
    for( temp_interface=0; temp_interface<MAX_ATM_NICS; temp_interface++ )
        houseclean_alarm[ temp_interface ] = this_cell_time + houseclean_interval;
#else
    houseclean_alarm[ interface ] = this_cell_time + houseclean_interval;
#endif
}



void cells_query( tcp_Socket *sock )
{
    Bit8 interface;
    Bit8 later_interface;
    double trace_duration;
    double trace_duration_temp;
    CardClock now2;
#if DELTA_TIME_STATS
    u_int sum_cleaning_count;
#endif
#if SHOW_CPU_UTILIZATION
    u_int sum_created;
#endif

#if SHOW_CPU_UTILIZATION
    // finalize cell processing time statistics
    // and print them
    //
    elapsed.set();
    elapsed -= startup;
    sum_created = 0;
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sum_created += created_cells[ interface ];
    if( sum_created )
    {
        sock_printf( sock, "cells_reassemble() took min %.4f max %.4f avg %.4f mS\n",
            min_elapsed.get_seconds() * 1000,
            max_elapsed.get_seconds() * 1000,
            sum_elapsed.get_seconds() * 1000 / sum_created
        );
        sock_printf( sock, "%.2f seconds (%.2f%% of %.2f) spent in cells_reassemble() during run\n",
            sum_elapsed.get_seconds(),
            100.0*(double)sum_elapsed/(double)elapsed,
            elapsed.get_seconds()
        );
    }
    else
        sock_printf( sock, "cells_reassemble() hasn't been called since last query\n" );
#endif

    if( !is_query_set )
    {
        sock_printf( sock, "cells_expire() has never been called\n" );
        return;
    }

#if QUERY_CALLS_EXPIRE
    PentiumClock now;

    // do one more housecleaning to trigger a timeout
    // on any "pregnant" packets, in case we haven't seen
    // a packet in more than houseclean_interval ticks
    //
    // this fake timestamp will not cause a negative
    // delta time to be calculated during next cells_expire(),
    // even if it is later than the next cell we will see,
    // because cells_clear() will cause last_houseclean_time
    // to be recalculated when we see the next cell
    //
    now.set();
    now2 = now;
    now2.insert_bitfields( -1 );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        cells_expire( (tracerec *)&now2, interface );
#endif      // QUERY_CALLS_EXPIRE

    trace_duration = 0.0;
    later_interface = 0;
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        if( is_alarm_set[ interface ] )
        {
            now2 = last_cell_time[ interface ];
            now2 -= last_query_time;
            now2.strip_bitfields();
            trace_duration_temp = now2.get_seconds();
            sock_printf( sock, "interface %d seconds since last cells query: %f\n", interface, trace_duration_temp );
            if( trace_duration < trace_duration_temp )
            {
                later_interface = interface;
                trace_duration = trace_duration_temp;
            }
        }
    }
    last_query_time = last_cell_time[ later_interface ];

#if COUNT_THINGS || BLOCK_USAGE_STATS || DELTA_TIME_STATS
#if DELTA_TIME_STATS
    sum_cleaning_count = 0;
#endif
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
#if DELTA_TIME_STATS
        sum_cleaning_count += cleaning_count[ interface ];
#endif
        sock_printf( sock, "interface %d housecleaning count: %u\n",
            interface, cleaning_count[interface] );
    }
#endif

#if DELTA_TIME_STATS
    sock_printf( sock, "seconds since last packet housecleaning: %f min, %f max, %f avg\n",
        min_houseclean_interval, max_houseclean_interval,
        sum_houseclean_interval/sum_cleaning_count );
#endif

    sock_printf( sock, "following statistics were calculated at packet expiration time:"
        " (first cell >.25 second after prev packet expiration)\n" );

#if COUNT_THINGS
    // cell stats
    //

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d pending cells: %d max, %.1f avg\n",
            interface, max_pending_cells[interface],
            sum_pending_cells[interface]/double(created_cells[interface]) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d new cells per second: %u max, %.1f avg, %u tot\n",
            interface, max_new_cells[interface],
#if AVERAGE_OF_AVERAGES
            sum_new_cells[interface]/double(cleaning_count[interface]),
            sum_new_cells[interface]
#else
            created_cells[interface]/nonzero(trace_duration),
            created_cells[interface]
#endif
        );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d cleaned cells per cleaning: %u max, %.1f avg, %u tot\n",
            interface, max_cleaned_cells[interface],
            double(cleaned_cells[interface])/cleaning_count[interface],
            cleaned_cells[interface]
        );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d deleted cells per cleaning: %.1f avg, %u tot\n",
            interface, double(deleted_cells[interface])/cleaning_count[interface],
            deleted_cells[interface]
        );

    // packet stats
    //

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d pending packets: %d max, %.1f avg\n",
            interface, max_pending_packets[interface],
            sum_pending_packets[interface]/double(created_packets[interface]) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d new packets per second: %u max, %.1f avg, %u tot\n",
            interface, max_new_packets[interface],
#if AVERAGE_OF_AVERAGES
            sum_new_packets[interface]/double(cleaning_count[interface]),
            sum_new_packets[interface]
#else
            created_packets[interface]/trace_duration,
            created_packets[interface]
#endif
        );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d cleaned packets per cleaning: %u max, %.1f avg, %u tot\n",
            interface, max_cleaned_packets[interface],
            double(cleaned_packets[interface])/cleaning_count[interface],
            cleaned_packets[interface]
        );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d deleted packets per cleaning: %.1f avg, %u tot\n",
            interface, double(deleted_packets[interface])/cleaning_count[interface],
            deleted_packets[interface]
        );

    // cells versus packets stats
    //

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d cells per packet: %u max, %.1f avg\n",
            interface, max_cells_per_packet[interface],
            double(created_cells[interface])/created_packets[interface]
        );

#endif //COUNT_THINGS

#if BLOCK_USAGE_STATS
    // block stats
    //

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d pending blocks per cleaning: %u max, %.1f avg, %u tot\n",
            interface, max_pending_blocks[interface],
            double(sum_pending_blocks[interface])/cleaning_count[interface],
            sum_pending_blocks[interface]
        );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %d cleaned blocks per cleaning: %u max, %.1f avg, %u tot\n",
            interface, max_cleaned_blocks[interface],
            double(sum_cleaned_blocks[interface])/cleaning_count[interface],
            sum_cleaned_blocks[interface]
        );
#endif
}



void cells_dump_all_packets( void *sock )
{
    cellhashrec *one_packet;
    unsigned iterations;
    u_int hashindex;

    for( hashindex=0; hashindex<NPKTBUCKETS; hashindex++ )
    {
        if( celltab[ hashindex ] )
            sock_printf( sock, "\nhash index %d head: 0x%.8X   tail: 0x%.8X\n",
                hashindex, celltab[hashindex], celltab_tails[hashindex] );
        else if( celltab_tails[ hashindex ] )
            sock_printf( sock, "\nhash index %d tail: 0x%.8X\n",
                hashindex, celltab_tails[hashindex] );

        for( one_packet=celltab[hashindex], iterations=0;
             one_packet && iterations<MAX_LIST_LENGTH;
             one_packet=one_packet->next, iterations++
        )
        {
            dump_packet( sock, one_packet );
        }

        if( iterations > 0 )
            sock_printf( sock, "\nhash index %d saw %d packets\n", hashindex, iterations );
    }
}



int cells_check_all_packets( void *sock )
{
    cellhashrec *one_packet;
    unsigned iterations;
    u_int hashindex;
    int result;

    for( hashindex=0; hashindex<NPKTBUCKETS; hashindex++ )
    {
        for( one_packet=celltab[hashindex], iterations=0;
             one_packet && iterations<MAX_LIST_LENGTH;
             one_packet=one_packet->next, iterations++
        )
        {
            result = check_packet( sock, one_packet );
            if( result )
            {
                sock_printf( sock, "error %d on %dth packet of hash index %d\n", result, iterations, hashindex );
                return result;
            }
        }

        if( iterations == MAX_LIST_LENGTH )
            sock_printf( sock, "infinite loop?  hash index %d saw more than %d packets\n", hashindex, iterations );
    }

    return 0;
}



#endif      // DO_FOREGROUND_STATS
