// this source file contains substitutes for the functions
// in FS2FLOWS.C that are called by TRACEFLO.CPP
// so we can perform our own analysis
//



// force these variables so this executable will always
// operate on an IP header trace, regardless how CAPTURE.EXE
// is compiled (as determined by CONSTANT.H)
//
// this is only safe to do if this is the only source file
// in this executable which cares about tracerec structure
//
#define CAPTURE_TO_DISK             FALSE
#define CARD_COPIES_ATM_PAYLOAD     TRUE



#include "traceflo.cpp"



// conditional compilation switches
//



// global variables
//
Signed32 gfc_wanted, vpi_wanted, vci_wanted;
Bit32 total_cells, unwanted_cells;



int flows_per_block( BufferBlock *block )
{
    Bit32 count, count2;
    tracerec *cell;
    Bit8 *pointer;
    int result;

    // for each record in the buffer...
    //
    for( count=0; count<block->used; count++ )
    {
        total_cells++;

        // shortcut for code that follows
        //
        cell = (tracerec *) (block->buffers + count*BUFFER_SIZE);

        // gotta swap the ATM header ourselves,
        // since we aren't going to call cells_reassemble()
        //
        cell->atm.whole = ntohl( cell->atm.whole );

        if( ( gfc_wanted!=-1 && cell->atm.bits.GFC!=gfc_wanted )
        ||  ( vpi_wanted!=-1 && cell->atm.bits.VPI!=vpi_wanted )
        ||  ( vci_wanted!=-1 && cell->atm.bits.VCI!=vci_wanted ) )
        {
            unwanted_cells++;
            continue;
        }

        dump_cell( stdout, cell );
    }

    return 0;
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

int flows_init( int argc, char *argv[], PentiumClock *keystroke_interval,
    PentiumClock *query_interval, PentiumClock *active_interval,
    int *want_facility_loopback, int *want_rx, int *want_tx )
{
    fprintf( stderr, "decodes the timestamp and ATM header, then does a hex dump\n" );
    fprintf( stderr, "of ATM payload, for desired card/VPI/VCI\n\n" );

    if( argc != 4 )
    {
        fprintf( stderr, "flows-args: GFC VPI VCI\n" );
        fprintf( stderr, "where -1 means 'any value'\n\n" );
        return 2;
    }

    gfc_wanted = strtol( argv[1], NULL, 0 );
    vpi_wanted = strtol( argv[2], NULL, 0 );
    vci_wanted = strtol( argv[3], NULL, 0 );

    fprintf( stdout, "you want GFC=%d VPI=%d VCI=%d\n", gfc_wanted, vpi_wanted, vci_wanted );

    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



int flows_cleanup( void )
{
    // show statistics
    //
    fprintf( stdout, "%lu cells received from file\n", total_cells );
    fprintf( stdout, "%lu cells didn't match GFC/VPI/VCI %d/%d/%d\n",
        unwanted_cells, gfc_wanted, vpi_wanted, vci_wanted );

    return 0;
}



int flows_is_query_wanted( void )
{
    return FALSE;
}


int flows_is_query_pending( void )
{
    return FALSE;
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

int flows_query( const char *extra_msg )
{
    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par
