#include <string.h>                             // stricmp()
#include <stdio.h>                              // fopen() fwrite() fclose()
#include <errno.h>                              // errno _sys_errlist
#ifdef __unix__
#define _sys_errlist sys_errlist
int sys_nerr;
char *sys_errlist[];
#endif
#include "../intel/mc146818.h"                  // class RealTimeClock



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *filename = argv[ 1 ];
    const char *want_overwrite_string = argv[ 2 ];
    int want_overwrite;
    Bit8 array[ RTC_USER_ADDRESS_COUNT ];
    FILE *handle;
    char *filemode;
    size_t result;

    if( argc<2 || argc>3 )
    {
        printf( "copies all %d of your BIOS's CMOS setup values to a disk file,\n", RTC_USER_ADDRESS_COUNT );
        printf( "or overwrites all your current values with such a file\n\n" );
        printf( "usage: %s file {overwrite-wanted}\n\n", program );
        printf( "note: you must spell out the word 'YES' as the 2nd argument\n"
                "      if you want to overwrite your current CMOS setup values\n" );
        return 1;
    }

    if( argc==3 && !stricmp(want_overwrite_string,"YES") )
    {
        want_overwrite = TRUE;
        filemode = "rb";
        printf( "overwriting current BIOS settings in realtime clock's CMOS RAM\n" );
    }
    else
    {
        want_overwrite = FALSE;
        filemode = "wb";
        printf( "saving current BIOS settings from realtime clock's CMOS RAM\n" );
    }

    handle = fopen( filename, filemode );
    if( !handle )
    {
        printf( "error %d '%s' opening file '%s' with mode '%s'\n", errno, _sys_errlist[errno], filename, filemode );
        return 2;
    }

    if( want_overwrite )
    {
        result = fread( array, 1, RTC_USER_ADDRESS_COUNT, handle );
        if( result != RTC_USER_ADDRESS_COUNT )
        {
            printf( "error %d '%s' reading %d elements to file '%s'\n", errno, _sys_errlist[errno], RTC_USER_ADDRESS_COUNT, filename );
            return 3;
        }

        RealTimeClock::put_all_binary( array );
    }
    else
    {
        RealTimeClock::get_all_binary( array );
        result = fwrite( array, 1, RTC_USER_ADDRESS_COUNT, handle );
        if( result != RTC_USER_ADDRESS_COUNT )
        {
            printf( "error %d '%s' writing %d elements to file '%s'\n", errno, _sys_errlist[errno], RTC_USER_ADDRESS_COUNT, filename );
            return 4;
        }
    }

    fclose( handle );
    return 0;
}
