#include <stdio.h>                      // printf()
#include <mem.h>                        // memset()
#include <dos.h>                        // FP_SEG() MK_FP()
#include <string.h>                     // stricmp()
#include "../scsi/aspi.h"               // AspiSrb aspi



int main( int argc, char *argv[] )
{
    int want_old;
    const char *program = argv[ 0 ];
    const char *old_or_new = argv[ 1 ];
    Bit8 *aspi_phys_to_real = (Bit8 *) MK_FP( FP_SEG( aspi ), 0x3F4D );
    Bit8 *aspi_signature = (Bit8 *) MK_FP( FP_SEG( aspi ), 0x34 );
    Bit8 expected_signature[] = "AIC-78XX/AIC-75XX ASPI Manager for DOS\r\n"
        "Version 1.32\r\n";

    Bit8 patching_opcodes[] =
    {
        // mov eax, es:[di+0Fh]
        //
        0x26, 0x66, 0x8B, 0x45, 0x0F,

        // NOP's until end of phys_to_real conversion
        //
        0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
        0x90, 0x90, 0x90, 0x90, 0x90, 0x90
    };

    Bit8 original_opcodes[] =
    {
        // movzx eax, es:word ptr [di+11h]
        //
        0x66, 0x26, 0x0f, 0xb7, 0x45, 0x11,

        // movzx edx, es:word ptr [di+0Fh]
        //
        0x66, 0x26, 0x0f, 0xb7, 0x55, 0x0f,

        // shl eax, 04
        //
        0x66, 0xc1, 0xe0, 0x04,

        // add eax, edx
        //
        0x66, 0x03, 0xc2
    };

#   if sizeof( original_opcodes ) != sizeof( patching_opcodes )
#   error "internal error: my patching opcodes aren't same size as original opcodes!\n"
#   endif

    if( memcmp( aspi_signature, expected_signature, sizeof(expected_signature)-1 ) )
    {
        fprintf( stderr, "error: ASPI driver signature at %p is '%s',"
            " but we expected it to be '%s'\naborting\n",
            aspi_signature, aspi_signature, expected_signature );
        return 2;
    }

    if( argc != 2 )
    {
        fprintf( stderr, "usage: %s old|new\n", program );
        return 3;
    }

    if( !stricmp( old_or_new, "old" ) )
        want_old = TRUE;
    else if( !stricmp( old_or_new, "new" ) )
        want_old = FALSE;
    else
    {
        fprintf( stderr, "you must type either 'old' or 'new'\nbut case doesn't matter\n" );
        return 4;
    }

    if( want_old )
    {
        memcpy( aspi_phys_to_real, original_opcodes, sizeof(original_opcodes) );
        printf( "restored DOS ASPI driver to transfer using realmode segment:offset addresses\n" );
    }
    else
    {
        memcpy( aspi_phys_to_real, patching_opcodes, sizeof(patching_opcodes) );
        printf( "patched DOS ASPI driver to transfer using physical addresses\n" );
    }

    return 0;
}
