#include <stdio.h>                  // printf()
#include <fcntl.h>                  // O_CREAT etc.   UNIX open()
#include <stdlib.h>                 // atoi()
#include <stdarg.h>                 // va_start() va_end() va_list
#include <string.h>                 // strcmp()
#include <ctype.h>                  // tolower()
#include <errno.h>                  // errno    DOS _sys_errlist
#include <sys/stat.h>               // S_IWRITE S_IREAD
#ifdef __unix__
#   if __alpha
#       include <unistd.h>          // read() write() close()
#   else
#       include <sys/unistd.h>      // read() write() close()
#   endif
#else
#   include <io.h>                  // open() close() read() write()
#   include <mem.h>                 // memset()
#   include <conio.h>               // kbhit()
#endif
#include "../intel/timestmp.h"
#include "../wattcp/tcp.h"
#include "../dpmi32/flat.h"         // flat_alloc()

void init_trace(int level);

#define noJOELS_SIZES  // Nevil's PC only has 32M of memory!
#define noUA_TEST                   // Use native unix socket read/write
                                    //    (rather than busy-wait in tcp_tick!)


#define FORCE_CACHING           FALSE
#define USE_NET_BUFFER          TRUE
#define SHOW_REAL_MEM_USAGE     FALSE
#define VERBOSE_TCP_FAILURES    TRUE
#define VERBOSE_FLAT_INIT       TRUE
#define FORCE_TCP_Q_SERVICE     FALSE



#define CONTROL_WAIT            100    // measured in seconds
#define FILENAME_SIZE           256
#define MODE_SIZE               100

#ifdef __unix__
#ifdef SUNOS
#define NET_BUFFER_SIZE         (16*1024 - 1)   // largest SunOS 4.1.3 can handle
#else
#ifdef SOLARIS
#define NET_BUFFER_SIZE         (64*1024)       // Solaris 2.5.1 can handle 128K but won't scale
//#define NET_BUFFER_SIZE         (10*1024*1024)  // largest Greg Miller has done on Solaris 2.6, but TCP_MAX_BUF kernel var must be upped
#else
#ifdef OSF
#define NET_BUFFER_SIZE         (128*1024)      // largest OSF/1 4.0 can handle
#else
#error unknown OS, cannot size net buffer!
#endif
#endif
#endif
#else   // defined( __unix__ )
#ifdef JOELS_SIZES
#define NET_BUFFER_SIZE         (2*1024*1024)
//#define NET_BUFFER_SIZE         (11*1024*1024)
//#define NET_BUFFER_SIZE         (63*1024)
//#define NET_BUFFER_SIZE         (64*1024)
//#define NET_BUFFER_SIZE         (100*1024)
//#define NET_BUFFER_SIZE         (200*1024)
//#define NET_BUFFER_SIZE         (300*1024)
#else  // Nevil's sizes
#define NET_BUFFER_SIZE         (1*1024*1024)
#endif
#endif  // defined( __unix__ )
#ifdef JOELS_SIZES
#define APP_BUFFER_SIZE         (10*1024*1024)  // Want an even multiple of 512 bytes for fast disk I/O
#else
#define APP_BUFFER_SIZE         (1*1024*1024)  // Want an even multiple of 512 bytes for fast disk I/O
#endif


#define TRUE                    1
#define FALSE                   0



Bit8 *app_buffer;
#if USE_NET_BUFFER
    Bit8 *net_buffer;
#endif



// wait for data to be available
// or socket to close
//
void wait_for_data( tcp_Socket *sock )
{
    while( tcp_tick( sock ) && !sock_dataready( sock ) )
        ;
}



#ifndef __unix__
// called during busy-waiting in the TCP/IP library
//
// only applicable to DOS; in UNIX you can always abort the program
// by killing the process
//
// will reset the socket if escape key is pressed
//
static tcp_Socket *my_yield_socket = NULL;
static PentiumClock my_yield_alarm( 0 );
#define YIELD_HOLDOFF_TIME  0.25     // seconds

void my_yield( void )
{
    char keystroke;
    PentiumClock now;

    now.set();
    if( now < my_yield_alarm )
        return;

    my_yield_alarm.put_seconds( YIELD_HOLDOFF_TIME );
    my_yield_alarm += now;

    if( kbhit() )
    {
        keystroke = getch();
        if( 27 == keystroke )
        {
            printf( "\nhalted by escape keypress\n" );
            sock_abort( my_yield_socket );
        }
    }
}
#endif



int cleanup( int retcode, tcp_Socket *sock, const char *format, ... )
{
    va_list argptr;

    va_start( argptr, format );
    vprintf( format, argptr );
    va_end( argptr );

    // signal that we want socket closed
    //
    sock_close( sock );

    // wait for socket to actually close
    //
    // does this involve 3 minute TIME_WAIT???
    //
    while( tcp_tick( sock ) )
#ifndef __unix__
        my_yield();
#else
        ;
#endif

    // return to OS
    //
    return retcode;
}



int sock_to_file( tcp_Socket *sock, const char *filename, long *bytes, PentiumClock *disk_io_time )
{
    Bit32 result;
    int result2;
    int file_handle;
    PentiumClock start, end;
#ifdef UA_TEST
    int rsock = * (int *)sock;
#endif

    // log to the console
    //
    printf( "copying TCP/IP stream to file '%s'\n", filename );

    // no bytes RX yet
    //
    *bytes = 0;

    // remember when this disk I/O started
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    start.set();

    // open up the file locally
    //
    file_handle = open( filename, O_CREAT|O_TRUNC|O_BINARY|O_WRONLY, S_IWRITE|S_IREAD );

    // get the disk I/O end time, calculate the delta,
    // and add it to the caller's total
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    end.set();
    *disk_io_time += end - start;

    // check the disk I/O result
    //
    if( file_handle == -1 )
    {
        return cleanup( 10, sock, "cannot create file '%s'\n", filename );
    }

    // change socket to binary mode so all bytes get copied verbatin
    //
    sock_mode( sock, TCP_MODE_BINARY );

    // while client hasn't closed socket...
    //
#ifndef UA_TEST
    result = 1;             // still have socket data to read
    while( result > 0 )
    {
#if FORCE_TCP_Q_SERVICE
        // joel debug:
        //
        // just service input queue ASAP for 20 seconds
        //
        start.set();
        end.put_seconds( 20 );
        end += start;
        do
        {
            result = tcp_tick( sock );
            start.set();
        } while( start<end && result );
#endif

        // read up to one buffer's worth of data
        //
        // note: cannot use sock_fastread() here because
        // we want our disk write to be an even multiple
        // of 512 bytes so it will be fast
        //
        result = sock_read( sock, app_buffer, APP_BUFFER_SIZE );

        // if we got anything...
        //
        if( result > 0 )
        {
            // fprintf( stderr, "got %d bytes from socket\n", result );

            // write it to disk file
            //
            start.set();
            result2 = write( file_handle, app_buffer, (unsigned)result );
            end.set();
            *disk_io_time += end - start;
            if( result2 != (unsigned)result )
            {
                close( file_handle );
                return cleanup( 12, sock, "could only write %d of %d bytes to file '%s'\n", result2, result, filename );
            }

            // keep track of how many bytes we have done
            //
            *bytes += result;
        }
    }
#else  // Nevil & Russell S, 31 Dec 97
   while ((result = read(rsock, app_buffer, APP_BUFFER_SIZE)) > 0) {
      write(file_handle, app_buffer, result);
      *bytes += result;
      }
#endif

    // close the disk file
    //
    start.set();
    close( file_handle );
    end.set();
    *disk_io_time += end - start;

    // exit to OS successfully
    //
    return 0;
}



int file_to_sock( tcp_Socket *sock, const char *filename, long *bytes, PentiumClock *disk_io_time )
{
    int result, result2;
    int file_handle;
    PentiumClock start, end;
#ifdef UA_TEST
    int rsock = * (int *)sock;
#endif

    // log to the console
    //
    printf( "copying file '%s' to TCP/IP stream\n", filename );

    // no bytes TX yet
    //
    *bytes = 0;

    // remember when this disk I/O started
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    start.set();

    // open up the file locally
    //
    file_handle = open( filename, O_BINARY|O_RDONLY );

    // get the disk I/O end time, calculate the delta,
    // and add it to the caller's total
    //
    // note: for clarity, will not be commented
    // on subsequent disk I/O
    //
    end.set();
    *disk_io_time += end - start;

    // check the disk I/O result
    //
    if( file_handle == -1 )
    {
        return cleanup( 20, sock, "error %d '%s' opening existing file '%s'\n", errno, _sys_errlist[errno], filename );
    }

    // change socket to binary mode so all bytes get copied verbatin
    //
    sock_mode( sock, TCP_MODE_BINARY );

    // while client hasn't closed socket
    // and we haven't reached end of file...
    //
#ifndef UA_TEST
    result = INT_MAX;           // still have file data to read
    while( result > 0 )
    {
        // read up to one buffer's worth of data
        //
        // note: error return code -1 is treated like EOF retcode 0
        //
        start.set();
        result = read( file_handle, app_buffer, APP_BUFFER_SIZE );
        end.set();
        *disk_io_time += end - start;

        // if we got anything...
        //
        if( result > 0 )
        {
            // write it to socket
            //
            result2 = sock_write( sock, app_buffer, result );
            if( result2 != result )
            {
                close( file_handle );
                return cleanup( 23, sock, "could only write %d of %d bytes to socket\n", result2, result );
            }

            // keep track of how many bytes we have done
            //
            *bytes += result;
        }
    }
#else  // Nevil & Russell S, 31 Dec 97
   while ((result = read(file_handle, app_buffer, APP_BUFFER_SIZE)) > 0) {
      write(rsock, app_buffer, result);
      *bytes += result;
      }
#endif

    // close the disk file
    //
    start.set();
    close( file_handle );
    end.set();
    *disk_io_time += end - start;

    // close the socket ourselves and
    // exit to OS successfully
    //
    return cleanup( 0, sock, "" );
}



int daemon( tcp_Socket *sock, unsigned port )
{
    int result;
    word count;
    Bit8 filename[ FILENAME_SIZE + 1 ];
    char mode_string[ MODE_SIZE + 1 ];
    int mode_is_upload;
    PentiumClock start, overall, disk, network;
    long bytes;

    // validate the socket
    //
    result = tcp_open( sock, port );

#if USE_NET_BUFFER
    // tell socket to use my buffer, not his, for receive
    //
    if( !sock_set_rx_buf( sock, net_buffer, NET_BUFFER_SIZE ) )
        return cleanup( 12, sock, "error setting socket to have %d byte receive buffer\n", NET_BUFFER_SIZE );
#endif

#ifndef __unix__
    // install a handler that gets called in TCP library's busy loops
    //
    my_yield_socket = sock;
    sock_yield( sock, my_yield );
#endif

    // listen for a connection from anyone
    //
    result = tcp_listen( sock, 0, 0, NULL, 0 );

    // wait for a connection to come in
    //
    sock_wait_established( sock, CONTROL_WAIT, NULL, &result );

    // remember when data transfer started
    //
    start.set();

    // change to ASCII mode because we don't want to be told
    // that any data is ready unless it is a complete line
    //
    sock_mode( sock, TCP_MODE_ASCII );

    wait_for_data( sock );

    // read the first line as command
    //
    count = sock_gets( sock, (Bit8 *)mode_string, MODE_SIZE );
    if( count == 0 )
    {
        return cleanup( 30, sock, "mode not provided!\n" );
    }
    mode_is_upload = strcmp( mode_string, "RETR" );

    wait_for_data( sock );

    // read the second line as a filename
    //
    count = sock_gets( sock, filename, FILENAME_SIZE );
    if( count == 0 )
    {
        return cleanup( 31, sock, "filename not provided!\n" );
    }

    disk.low = 0;
    disk.high = 0;
    if( mode_is_upload )
    {
        // suck up the last trailing newline
        //
//@@        wait_for_data( sock );
/* Above busy-waits forever if the first window of the data file
   being copied doesn't contain a \r !!!        Nevil, 16 Jan 98 */
        sock_dataready( sock );
        result = sock_to_file( sock, (char *)filename, &bytes, &disk );
    }
    else
    {
#if USE_NET_BUFFER
        // tell socket to use his buffer, not mine, for receive
        //
        if( !sock_set_rx_buf( sock, NULL, 0 ) )
            return cleanup( 12, sock, "error setting socket to use default receive buffer\n" );

        // tell socket to use my buffer, not his, for transmit
        //
        if( !sock_set_tx_buf( sock, net_buffer, NET_BUFFER_SIZE ) )
            return cleanup( 12, sock, "error setting socket to have %d byte transmit buffer\n", NET_BUFFER_SIZE );
#endif
        result = file_to_sock( sock, (char *)filename, &bytes, &disk );
    }

    // calculate how much time has elapsed
    //
    overall.set();
    overall -= start;
#ifndef __unix__
    if( !mode_is_upload )       // if we did the active close...
    {
        start.put_seconds( sock_get_2msl_timeout( sock ) );
        overall -= start;
    }
#endif
    network = overall - disk;

    // inform the user how much we got and how fast
    //
    printf( "%.3f S overall = %.3f disk + %.3f network\n%ld bytes, %.3f network KB/s\n",
       overall.get_seconds(), disk.get_seconds(), network.get_seconds(),
       bytes, ((double)bytes/1000.0)/network.get_seconds()
    );

    return result;

sock_err:
    printf( "got error %d from sock_wait_established\n", result );
    return 999;
}



int client( tcp_Socket *sock, unsigned port, char *hostname, char *mode_string, char *local_filename, char *remote_filename )
{
    int result;
    int mode_is_upload;
    longword hostaddr;
    char *command;
    PentiumClock start, overall, disk, network;
    long bytes;

    // resolve the host name to an address
    //
    hostaddr = resolve( hostname );
    if( !hostaddr )
    {
        printf( "unable to resolve host name '%s' to numerical IP address\n", hostname );
        return 40;
    }

    // activate the socket
    // so we can set the RX buffering on it
    //
    result = tcp_open( sock, 0 );

    // convert the first character of the mode string to lowercase
    //
    mode_string[ 0 ] = tolower( mode_string[ 0 ] );

    // convert the mode into a command
    //
    if( !strcmp( mode_string, "u" ) )
    {
        mode_is_upload = TRUE;
        command = "SAVE";
#if USE_NET_BUFFER
        // tell socket to use my buffer, not his, for transmit
        //
        if( !sock_set_tx_buf( sock, net_buffer, NET_BUFFER_SIZE ) )
            return cleanup( 12, sock, "error setting socket to have %d byte transmit buffer\n", NET_BUFFER_SIZE );
#endif

    }
    else if( !strcmp( mode_string, "d" ) )
    {
        mode_is_upload = FALSE;
        command = "RETR";
#if USE_NET_BUFFER
        // tell socket to use my buffer, not his, for receive
        //
        if( !sock_set_rx_buf( sock, net_buffer, NET_BUFFER_SIZE ) )
            return cleanup( 12, sock, "error setting socket to have %d byte receive buffer\n", NET_BUFFER_SIZE );
#endif
    }
    else
    {
        printf( "invalid file transfer mode '%s'\n", mode_string );
        return 41;
    }

    // initiate a connection
    //
    result = tcp_connect( sock, hostaddr, port, NULL );

#ifndef __unix__
    // install a handler that gets called in TCP library's busy loops
    //
    my_yield_socket = sock;
    sock_yield( sock, my_yield );
#endif

    // wait for our connection to finish
    //
    sock_wait_established( sock, CONTROL_WAIT, NULL, &result );

    // remember when data transfer started
    //
    start.set();

    // change to ASCII mode because we will be writing strings
    //
    sock_mode( sock, TCP_MODE_ASCII );

    // write the command as the first line
    //
    sock_puts( sock, (Bit8 *)command );

    // write the filename as the second line
    //
    sock_puts( sock, (Bit8 *)remote_filename );

    disk.low = 0;
    disk.high = 0;
    if( mode_is_upload )
        result = file_to_sock( sock, local_filename, &bytes, &disk );
    else
        result = sock_to_file( sock, local_filename, &bytes, &disk );

    // calculate how much time has elapsed
    //
    overall.set();
    overall -= start;
#ifndef __unix__
    if( mode_is_upload )        // if we did the active close...
    {
        start.put_seconds( sock_get_2msl_timeout( sock ) );
        overall -= start;
    }
#endif
    network = overall - disk;

    // inform the user how much we sent and how fast
    //
    printf( "%.3f S overall = %.3f disk + %.3f network\n%ld bytes, %.3f network kB/s\n",
       overall.get_seconds(), disk.get_seconds(), network.get_seconds(),
       bytes, ((double)bytes/1000.0)/network.get_seconds()
    );

    return result;

sock_err:
    printf( "got error %d from sock_wait_established\n", result );
    return 999;
}



int main( int argc, char *argv[] )
{
    const char *program = argv[0];
    unsigned control_port;
    int result;
    tcp_Socket control_socket;

    if( argc!=2 && argc!=5 && argc!=6 )
    {
        printf( "daemon:          %s tcp-port-number\n", program );
        printf( "client   upload: %s tcp-port-number hostname u local-filename {remote-filename}\n", program );
        printf( "client download: %s tcp-port-number hostname d local-filename {remote-filename}\n", program );
        return 1;
    }

    control_port = atoi( argv[ 1 ] );

    // initialize TCP/IP library
    //
    sock_init();
#if VERBOSE_TCP_FAILURES
    tcp_set_debug_state( 1 );
#endif
#if SHOW_REAL_MEM_USAGE
    system( "mem /d | find \"TCPCOPYD\"" );
#endif

//   tcp_set_debug_state(1);  /* Display flags for each packet */
//   tcp_set_debug_state(2);  /* Display full trace data */

//   init_trace(1);  /* Nevil's tcp packet trace */

    // initialize access to our own area of flat memory
    //
    flat_init( VERBOSE_FLAT_INIT );

    // allocate buffers
    //
    app_buffer = (Bit8 *) flat_alloc( APP_BUFFER_SIZE );
    if( !app_buffer )
    {
        printf( "error %d allocating %d bytes for application buffer\n", errno, APP_BUFFER_SIZE );
        return 50;
    }

#if USE_NET_BUFFER
    net_buffer = (Bit8 *) flat_alloc( NET_BUFFER_SIZE );
    if( !net_buffer )
    {
        printf( "error %d allocating %d bytes for network buffer\n", errno, NET_BUFFER_SIZE );
        return 51;
    }
#endif

#if FORCE_CACHING
    // touch all words of all buffers to force caching to a known state
    //
    // does this make timings more consistent?
    //
    memset( app_buffer, 0, APP_BUFFER_SIZE );
#if USE_NET_BUFFER
    memset( net_buffer, 0, NET_BUFFER_SIZE );
#endif
#endif

    // set socket to known state
    //
    memset( &control_socket, 0, sizeof(control_socket) );

    if( argc == 2 )
    {
        result = daemon( &control_socket, control_port );
    }
    else
    {
        result = client( &control_socket, control_port, argv[2], argv[3], argv[4], argc==6?argv[5]:argv[4] );
    }

    return result;
}



// written 7/17/96 by Joel Apisdorf
//
