/* 1125, Thu 8 Jan 98

   PCPKT.C: Packet driver interface for Waterloo TCP

   This version is derived from the original Waterloo one.
   Nevil Brownlee added a 'metering' capacbility for use by
      traffic measuring systems such as NeTraMet.
   Joel Apisdorf ported it to use flat 32-bit mode with
      Borland PowerPack for DOS.
   Nevil added the ability to handle high-performance
      packet drivers (at least for read lookahead),
      and to handle up to four ethernets simultaneously.
*/

// FRAGSUPPORT enables support for packet reassembly of fragmented packets
#define noFRAGSUPPORT
#define noPKT_DEBUG

#define JOELDEBUG              0

#define PKT_INIT_TALKS         0
#define GETBUF_TALKS           0
#define IP_BUF_TRACE           0
#define PKT_RELEASE_TALKS      1
#define PKT_ENTRY_TALKS        0
#define PKT_ENTRY_PARANOIA     0

#pragma inline

#include <dos.h>
#ifdef __BORLANDC__
#include <mem.h>
#endif
#include <string.h>
#include <stdlib.h>

#include "..\wattcp\copyrigh.h"
#include "..\wattcp\wattcp.h"
#include "..\wattcp\elib.h"
#include "..\wattcp\ethdev.h"

#include "..\types\types.h"
#include "..\dpmi32\intr.h"
#include "..\dpmi32\far.h"
#include "..\dpmi32\realmode.h"
#include "..\dpmi32\phys.h"

void hwt_init(void);  // Initialise timer routines (in ..\intel\timestmp.cpp)
void hwt_fin(void);  // Declared here to avoid compiling pcpkt as a C++ module

#define far
#include "..\wattcp\pktdrv.h"

#define MAXBUFS          200  /* Max nbr of Ethernet buffers */

#define TOT_BUF_MEM     (sizeof(struct ether) + MAXBUFS*sizeof(struct pbuf))
#define STACK_SIZE      4096

char au_monitor = 0;  /* AU: 1 to enable monitoring of sent packets */

#define INT_FIRST       0x60
#define INT_LAST        0x80
#define PD_DRIVER_INFO  0x01FF
#define PD_ACCESS       0x0200
#define PD_RELEASE      0x0300
#define PD_SEND         0x0400
#define PD_GET_ADDRESS  0x0600

#define PD_RCV_MODE     0x1400  /* AU Monitor */
#define PM_DEFAULT      3  /* Unicast + Broadcast (Nevil, 27 Dec 96) */
#define PM_PROMISCUOUS  6

#define PD_GET_PARAMS   0x0A00  /* 1.10 pkt driver spec.  Nevil, 15 Jul 96 */
struct pd_param {
   unsigned char major_rev;
   unsigned char minor_rev;
   unsigned char length;
   unsigned char addr_len;
   unsigned short mtu;
   unsigned short multicast_aval;
   unsigned short rcv_bufs;
   unsigned shortxmt_bufs;
   unsigned short int_num;
   };

#define  CARRY          1    /* Carry bit in flags register */

Bit32 real_mem_handle;  /* For accessing bottom 1M of real memory */
Bit8 *real_mem;

longword pkt_flat2phys(void *pointer)
{
   return (Bit8 *)pointer - real_mem;
   }

Bit16 buffer_selector;  /* Real memory buffer space for IP buffers */
Bit16 buffer_segment;
Bit8 *buffer_base;

/* Interface to packet drivers */
struct RealModeRegsC pkt_entry_real_regs[ MAX_PD ];
Bit32 pkt_entry_real_ptr[ MAX_PD ];
Bit8 pkt_entry_stack[ MAX_PD ][ STACK_SIZE ];
Bit48 pkt_entry( Bit48 real_regs_pointer, Bit48 real_stack_pointer );

#define IP_DF       0x0040      // Don't fragment bit set for FRAG Flags
byte *fragment(in_Header *ip);

typedef byte (*pbuf_ptr)[sizeof(struct pbuf)];  /* Packet buffers */
pbuf_ptr pktbuf;    /* First byte is busy flag, 2nd spare */
struct ether *outbuf;

char *pkt_line = "PKT DRVR";
word pkt_ip_type  = 0x0008;  /* These are Intelled values */
word pkt_arp_type = 0x0608;

struct en_info en_table[MAX_PD] =  /* Default: one pkt driver, IP only */
   {{EN_IP,0,0}, {0,0,0}, {0,0,0}, {0,0,0}};

/* en_table holds info for each internet interface.
   IP may only be active on ONE of these, using the
   variables below */

word _pktipofs = 0;  /* Offset from header to start of pkt */
word pkt_interrupt;
word pkt_arp_handle;
byte eth_addr[6];

long rx_packets = 0, rx_bytes = 0,  /* AU monitor */
   tx_packets = 0, tx_bytes = 0,
   found_packets = 0,
   badpackets = 0, nobufpackets = 0,
   send_fails = 0,  // Nevil, 10 Jan 97
   low_send_fails = 0;  // Joel, 22 Jan 97

struct bufq freeq;
struct bufq readyq;

/* Add a pbuf to the tail of a queue */
void enqueue(struct bufq *bq, struct pbuf *pb)
{
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      if (bq->len == 0) bq->head = pb;
      else bq->tail->next = pb;
      bq->tail = pb;
      ++bq->len;
   asm popf;   /* Restore flags */
   }

/* Put a pbuf into a queue, maintained in ascending seq order */
void seq_enqueue(struct bufq *bq, struct pbuf *np)
{
   struct pbuf *p, *prev_p;
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      if (bq->len == 0) {  /* Empty queue */
         bq->head = bq->tail = np;  np->next = NULL;
         }
      else {
         p = bq->head;
         if (np->seq < p->seq) {  /* Add to head */
            bq->head = np;  np->next = p;
            }
         else for (;;) {
            prev_p = p;  p = p->next;
            if (p == NULL) {  /* Add to tail */
               prev_p->next = np;  np->next = NULL;
               bq->tail = np;
               break;
               }
            if (np->seq < p->seq) {  /* Add into queue */
               prev_p->next = np;  np->next = p;
               break;
               }
            }
         }
      ++bq->len;
   asm popf;   /* Restore flags */
   }

/* Take the head pbuf from a queue */
struct pbuf *dequeue(struct bufq *bq)
{
   struct pbuf *rb;
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      if (bq->len == 0) rb = NULL;
      else {
         rb = bq->head;
         bq->head = rb->next;
         --bq->len;
         }
   asm popf;   /* Restore flags */
   return rb;
   }

void pd_error(unsigned int err_nbr, int sw_int, char *msg)
{
   printf("PD ERROR %d, interrupt %d: %s\n",
      err_nbr >> 8, sw_int, msg);
   }

int pkt_rcv_mode(int pkt_int, int handle, int mode)
{
   union REGS regs;
   regs.w.ax = PD_RCV_MODE;
   regs.w.bx = handle;
   regs.w.cx = mode;
   regs.w.flags = 0;
   int386(pkt_int, &regs, &regs);
   if (regs.x.flags & CARRY) {
      pd_error(regs.w.dx, pkt_int, "setting receive mode\n");
      return 1;
      }
   return 0;
   }

int check_pkt_driver(int *pi,  /* Returns 0 if OK, 1 otherwise */
   Bit32 pkt_entry_real_ptr_parm)
{
   union REGS regs;
   struct RealModeRegs real_regs;
   Bit16 result;
   Bit32 ptr;
   int pd_type, class;  /* Packet driver type and class */
   int pd_ok, pint;

   pd_ok = 0;
   if (*pi == 0) {
      for (pint = INT_FIRST; pint <= INT_LAST; ++pint) {
     ptr = real_to_physical(getvect_real(pint));
     if (!memcmp(real_mem+ptr+3, pkt_line, strlen(pkt_line))) {
            *pi = pint;
            pd_ok = 1;
        break;
            }
     }
      }
   else {
      ptr = real_to_physical(getvect_real(pint = *pi));
      if (!memcmp(real_mem+ptr+3, pkt_line, strlen(pkt_line)))
         pd_ok = 1;
      }
   if (!pd_ok) {
      pd_error(0, *pi, "NO PACKET DRIVER FOUND\n");
      return 1;
      }

   /* We need to point at IP type in DOS memory,
    * so store it temporarily at the beginning of
    * the packet buffer space we allocated in DOS
    * memory, and copy it back out afterwards
    *
    * This assumes that packet driver will not continue
    * to use the pointer after this call
    *
    * There is no need to copy it back out afterwards,
    * since the driver shouldn't be changing it
    */
   *(word *)outbuf = pkt_ip_type;

   /* Lets find out about the driver */
   regs.w.ax = PD_DRIVER_INFO;
   regs.w.flags = 0;
   int386(pint, &regs, &regs);

   /* Handle old versions, assume a class and just keep trying */
   if (regs.w.flags & CARRY) {  /* Carry set -> error */
      for (class = 0; class != 2; ++class) {
     _pktdevclass = (class) ? PD_SLIP : PD_ETHER;
         for (pd_type = 1; pd_type != 128; ++pd_type) {
        real_regs.reserved = 0;
        real_regs.ss = 0;
        real_regs.sp = 0;
        real_regs.flags = 0;

        real_regs.eax.low16 = PD_ACCESS | _pktdevclass;
        real_regs.ebx.low16 = pd_type;
        real_regs.edx.low16 = 0;  /* Interface number */
        real_regs.ecx.low16 =
               (_pktdevclass == PD_SLIP) ? 0 : sizeof pkt_ip_type;
        real_regs.ds = buffer_segment;
        real_regs.esi.low16 = (Bit8 *)outbuf - buffer_base;
        real_regs.es = pkt_entry_real_ptr_parm >> 16;
        real_regs.edi.low16 = pkt_entry_real_ptr_parm & 0xFFFF;

        result = simulate_real_interrupt(pint, &real_regs);
        if (result != 0) {
           printf("check_pkt_driver() saw DPMI error 0x%4.4X simulating real mode"
                      " interrupt 0x%2.2X to IP access packet driver\n", result, pkt_interrupt);
           return result;
               }

        if (!(real_regs.flags & CARRY))  /* Stop on the first success */
           break;
            }

     /* We have found a working type, so kill it */
     regs.w.bx = regs.w.ax;      /* Handle */
     regs.w.ax = PD_RELEASE;
     regs.w.flags = 0;
     int386(pint, &regs, &regs);
     }

      if (pd_type == 128) {
         pd_error(0, pint, "couldn't initialise driver\n");
     return 1;
         }
      }
   else {
      /* We could get pd_type from regs.w.dx
       * if we had any use for it
       */
      switch (_pktdevclass = (regs.w.cx >> 8)) {
      case PD_ETHER:
     _pktipofs = 14;
      case PD_SLIP:
     break;
      default:
         pd_error(0, pint, "only Ethernet or SLIP drivers allowed\n");
     return 1;
     }
      }
   return 0;
   }

int read_en_address(int if_nbr)
{
   int pkt_int;
   struct RealModeRegs real_regs;
   Bit16 result;
   /*
    * It would be VERY bad to use a packet buffer for
    * temporary storage of the Ethernet address here,
    * because as soon as we did the access call, received
    * packets could start using that area of memory!
    */
   pkt_int = en_table[if_nbr].int_nbr;
   real_regs.reserved = 0;
   real_regs.ss = 0;
   real_regs.sp = 0;
   real_regs.flags = 0;
   real_regs.eax.low16 = PD_GET_ADDRESS;
   real_regs.ebx.low16 = en_table[if_nbr].handle;
   real_regs.es = buffer_segment;
   real_regs.edi.low16 = (Bit8 *)outbuf - buffer_base;
   real_regs.ecx.low16 = sizeof(eth_addr);
   result = simulate_real_interrupt(pkt_int, &real_regs);
   if (result != 0) {
      printf("read_en_address() saw DPMI error 0x%4.4X simulating real mode"
     " interrupt 0x%2.2X to get ethernet address\n",
         result, pkt_int);
      return result;
      }
   if (real_regs.flags & CARRY) {
      pd_error(real_regs.edx.low16, pkt_int, "reading Ethernet address\n");
      return 1;
      }
   memcpy(eth_addr, outbuf, sizeof(eth_addr));
#if PKT_INIT_TALKS
   printf("Int %d: Eth addr %02X:%02X:%02X:%02X:%02X:%02X\n", pkt_int,
      eth_addr[0], eth_addr[1], eth_addr[2],
      eth_addr[3], eth_addr[4], eth_addr[5]);
#endif
   return 0;
   }

int en_init(int if_nbr)  /* Returns 0 if OK, non-zero otherwise */
{
   struct RealModeRegs real_regs, real_regs2;
   Bit16 result;
   union REGS regs;
   int pkt_int, if_type;
   struct pd_param *pd_pp;  /* Pkt driver parameters */

   if (if_nbr >= MAX_PD)
      return 2;  /* Fail */

   if (check_pkt_driver(&en_table[if_nbr].int_nbr, pkt_entry_real_ptr[if_nbr]))
      return 1;  /* Fail */

   pkt_int = en_table[if_nbr].int_nbr;  /* May be set by check_pkt (above) */
   if_type = en_table[if_nbr].type;
   if (if_type & EN_IP)  /* IP on this interface */
      pkt_interrupt = pkt_int;
#ifdef PKT_DEBUG
      printf("   en_init(%d), pkt_int=%d, type=%d, pkt_interrupt=%d\n",
         if_nbr,pkt_int,if_type, pkt_interrupt);
#endif

   if (if_type & EN_HP) { /* Try to use high-performance driver */
      real_regs.reserved = 0;  /* Get driver parameters */
      real_regs.ss = 0;
      real_regs.sp = 0;
      real_regs.flags = 0;
      real_regs.eax.low16 = PD_GET_PARAMS;
      result = simulate_real_interrupt(pkt_int, &real_regs);
      if (result != 0) {
     printf("en_init() saw DPMI error 0x%4.4X simulating real mode"
        " interrupt 0x%2.2X to get driver parameters\n",
            result, pkt_int);
     return result;
     }
      if ((real_regs.flags & CARRY) == 0) {  /* Success */
         pd_pp = (struct pd_param *)(  /* Make flat-mode protected pointer */
            real_mem + real_to_physical(
            real_regs.es << 16 | real_regs.edi.low16));
#if PKT_INIT_TALKS
         printf("Got parameters, int %d:  PD Spec %d.%d\n",
            pkt_int, pd_pp->major_rev,pd_pp->minor_rev);
#endif
     if (pd_pp->minor_rev < 10) if_type ^= EN_HP;
     }
      else if_type ^= EN_HP;
      }

   /* Tell the packet driver our callback address and desired ethertype */
   real_regs.reserved = 0;
   real_regs.ss = 0;
   real_regs.sp = 0;
   real_regs.flags = 0;

   real_regs.eax.low16 = PD_ACCESS | _pktdevclass;
   real_regs.ebx.low16 = 0xFFFF;  /* Any type */
   real_regs.edx.low16 = 0;       /* Interface number */

   real_regs.es = pkt_entry_real_ptr[if_nbr] >> 16;
   real_regs.edi.low16 = pkt_entry_real_ptr[if_nbr] & 0xFFFF;

   if (if_type & EN_MT) {  /* Metering required */
      real_regs.ecx.low16 = 0;  /* Typelen */
      *(word *)outbuf = 0;
      real_regs.ds = buffer_segment;
      real_regs.esi.low16 = (Bit8 *)outbuf - buffer_base;
      result = simulate_real_interrupt(pkt_int, &real_regs);
      if (result != 0) {
     printf("en_init() saw DPMI error 0x%4.4X simulating real mode"
            " interrupt 0x%2.2X to access packet driver\n",
            result, pkt_int);
     return result;
     }
      if (real_regs.flags & CARRY) {
         pd_error(real_regs.edx.low16, pkt_int,
            "accessing packet driver\n");
     return 1;
     }

      en_table[if_nbr].handle = real_regs.eax.low16;
      if (if_type & EN_IP) {  /* IP on this interface */
         au_monitor = 1;
     read_en_address(if_nbr);
         }
      pkt_rcv_mode(pkt_int, en_table[if_nbr].handle, PM_PROMISCUOUS);
      return 0;  /* OK */
      }

   /* No metering: assume IP is required */
   real_regs.ecx.low16 = (_pktdevclass == PD_SLIP) ? 0 : sizeof pkt_ip_type;
   *(word *)outbuf = pkt_ip_type;
   real_regs.ds = buffer_segment;
   real_regs.esi.low16 = (Bit8 *)outbuf - buffer_base;

   real_regs2 = real_regs;

   result = simulate_real_interrupt(pkt_interrupt, &real_regs);
   if (result != 0) {
      printf("en_init() saw DPMI error 0x%4.4X simulating real mode"
         " interrupt %d to IP access packet driver\n",
         result, pkt_interrupt);
      return result;
      }
   if (real_regs.flags & CARRY) {
      pd_error(real_regs.edx.low16, pkt_int,
         "accessing packet driver for IP\n");
      return 1;
      }
   en_table[if_nbr].handle = real_regs.eax.low16;
   read_en_address(if_nbr);
#if PKT_INIT_TALKS
   printf("      IP handle  = 0x%04X\n", en_table[if_nbr].handle);
#endif
   if (_pktdevclass != PD_SLIP) {  /* Get ARP handle */
      *(word *)outbuf = pkt_arp_type;
      real_regs = real_regs2;
      result = simulate_real_interrupt(pkt_interrupt, &real_regs);
      if (result != 0) {
     printf("en_init DPMI error 0x%4.4X simulating real mode"
            " interrupt 0x%2.2X to ARP access packet driver\n",
            result, pkt_interrupt);
            return result;
     }
      if (real_regs.flags & CARRY) {
     regs.w.ax = PD_RELEASE;  /* Release IP handle */
     regs.w.bx = en_table[if_nbr].handle;
     regs.w.flags = 0;
     int386(pkt_interrupt, &regs, &regs);
         pd_error(regs.w.dx, pkt_int, "accessing driver for ARP\n");
     return 1;
     }
      pkt_arp_handle = real_regs.eax.low16;
      }
#if PKT_INIT_TALKS
   printf("      ARP handle = 0x%04X\n", pkt_arp_handle);
#endif
   return 0;  /* OK */
   }

int pkt_init(void)
{
   Bit16 result;
   int init_result;
   Bit32 length_available;
   int j;

   /* Create a mapping for all 1 MB of real mode memory space
    *
    * This is used to search for signature near the beginning of
    * real mode interrupt handlers, and to access our DOS
    * memory packet buffers
    */
   real_mem = map_physical(0, 1024*1024, &real_mem_handle);
   if (!real_mem) {
      printf("error mapping real mode memory\n");
      return 1;
      }

   /* Allocate enough DOS memory for all
    * our input and output packet buffers
    */
   result = alloc_dos_mem(TOT_BUF_MEM,
      &length_available, &buffer_selector, &buffer_segment);
   if (result != 0) {
      printf("error 0x%4.4X allocating %d bytes of DOS memory; only %d available\n",
         result, TOT_BUF_MEM, length_available);
      unmap_physical(1024*1024, real_mem_handle);
      return 1;
      }
   buffer_base = real_mem + real_to_physical(buffer_segment << 16);

   outbuf = (struct ether *)buffer_base;
   pktbuf = (pbuf_ptr)(outbuf + 1);
   pkt_buf_wipe();  /* Initialise ethernet buffer pool */

   /* Create real mode addresses that the packet drivers
    * can call to get to our code
    */
   for (j = 0; j != MAX_PD; ++j) {
      if (en_table[j].type) {
         /* Make real mode addresses which packet
               drivers can call to get to our code */
         result = alloc_real_callback(pkt_entry,
            pkt_entry_stack[j]+STACK_SIZE,
            &pkt_entry_real_regs[j], &pkt_entry_real_ptr[j]);
         if (result != 0) {
            printf("Can't alloc real callback %d\n", j);
            free_dos_mem(buffer_selector);
            unmap_physical(1024*1024, real_mem_handle);
            return 1;
            }
         }
      }
   /* Initialise timer routines (before we start pkt drivers) */
   /* hwt_init();  Now initialised in timestmp.cpp.  Nevil, 8 Jun 98 */
   for (result = j = 0; j != MAX_PD; ++j) {
      if (en_table[j].type) {
         if ((result = en_init(j)) != 0)  /* Initialise packet drivers */
            break;
         }
      }
   return result;  /* OK */
   }

void en_release(int if_nbr)
{
   union REGS regs;
   Bit16 result;

   int pkt_int = en_table[if_nbr].int_nbr;
   int handle  = en_table[if_nbr].handle;
   int if_type = en_table[if_nbr].type;

   if (if_nbr >= MAX_PD)
      return;

   if (if_type & EN_MT) {  /* We were metering */
      pkt_rcv_mode(pkt_int, handle, PM_DEFAULT);
      }
   else if (if_type & EN_IP) {  /* IP without metering */
      if (_pktdevclass != PD_SLIP) {
         regs.w.ax = PD_RELEASE;
         regs.w.bx = pkt_arp_handle;
         regs.w.flags = 0;
         int386(pkt_interrupt, &regs, &regs);
         if (regs.w.flags & CARRY)
         pd_error(regs.w.dx, pkt_int, "releasing driver for ARP\n");
         }
      }
   regs.w.ax = PD_RELEASE;
   regs.w.bx = handle;
   regs.w.flags = 0;
   int386(pkt_int, &regs, &regs);
   if (regs.w.flags & CARRY)
      pd_error(regs.w.dx, pkt_int, "releasing driver for IP\n");

   result = free_real_callback(pkt_entry_real_ptr[if_nbr]);
   if (result != 0)
      printf("ERROR 0x%4.4X freeing real mode callback for i/f %d\n",
      result, if_nbr);
   }

void pkt_release()
{
   int j;
   Bit16 result;

   for (j = 0; j != MAX_PD; ++j) {
      if (en_table[j].int_nbr)
      en_release(j);
      }

   hwt_fin();  // Finished with clock routines

   result = free_dos_mem(buffer_selector);
   if (result != 0)
      printf("ERROR 0x%4.4X freeing DOS memory\n", result);

   unmap_physical(1024*1024, real_mem_handle);

#if PKT_RELEASE_TALKS
   printf("ethernet bytes/packets %ld/%ld TX, %ld/%ld RX\n",
      tx_bytes,tx_packets, rx_bytes,rx_packets);
   printf("%ld ethernet packets found for application\n", found_packets);
   printf("%ld ethernet packets dropped for 0 or >%d length\n",
      badpackets, ETH_MTU);
   printf("%ld ethernet packets dropped because no buffer available\n",
      nobufpackets);
   printf("%ld packet sends failed (and were retried)\n", send_fails);
   printf("%ld low-level packet sends failed\n", low_send_fails);
#endif
   }

unsigned char *get_pktbuf(Bit16 length_wanted)
{
   struct pbuf *bp;

   // Make sure the packet driver isn't asking for more than one buffer
   if (length_wanted > ETH_MTU || length_wanted == 0) {
      ++badpackets;
#if GETBUF_TALKS
      printf("get_pktbuf() failed; pkt driver wanted %u bytes; ethernet MTU is %u\n",
         length_wanted, ETH_MTU );
#endif
      return (Bit8 *)NULL;
      }

   if ((bp = dequeue(&freeq)) != NULL) {
      bp->state = 1;  /* Waiting for 2nd interrupt */
      return (Bit8 *)&bp->data;
      }

#if GETBUF_TALKS
      printf("get_pktbuf() failed; no buffers free\n");
#endif
   nobufpackets++;
   return (Bit8 *)NULL;
   }

// WARNING: don't mess with the high bits of 32-bit registers
// because the real mode interrupt handler didn't know to save
// them, so you will be screwing up your foreground code's
// registers!
//
// By the way, it is safe to call printf() from here, but
// stopping at a debugger breakpoint causes a TRAP E

Bit48 pkt_entry(Bit48 real_regs_pointer, Bit48 real_stack_pointer)
{
   Bit8 *lah_buffer, *buffer;
   Bit32 return_address;
   struct RealModeRegs *my_regs = (struct RealModeRegs *) real_regs_pointer.offset;
   struct RealModeRegsC *my_c_regs = struct_from_field( RealModeRegsC, regs, my_regs );
   Bit16 interface = my_c_regs - pkt_entry_real_regs;

#if PKT_ENTRY_TALKS
   printf( "pkt_entry( real_regs=0x%.4X:%.8X real_stack=0x%.4X:%.8X )\n",
      real_regs_pointer.selector, real_regs_pointer.offset,
      real_stack_pointer.selector, real_stack_pointer.offset );
   printf( "note: _DS=0x%.4X &pkt_entry_real_regs[0]=0x%.8X\n", _DS, &pkt_entry_real_regs[0] );
   printf( "calculated c_regs=0x%.8X interface=%u\n", my_c_regs, interface );
#endif

#if PKT_ENTRY_PARANOIA
   // Make sure that the real mode registers we got
   // are actually our static variable, or we will
   // need to set up addressing to them from flat mode
   // by calling map_physical()
   //
   if (my_regs < &pkt_entry_real_regs[0].regs
   ||  my_regs > &pkt_entry_real_regs[MAX_PD-1].regs ) {
      printf("pkt_entry() error: we got real mode registers at 0x%4.4X:%8.8X,"
         " but static pkt_entry_real_regs array is at 0x%4.4X:%8.8X length %d,"
         " so we do not have addressibility from flat C code!\n",
         real_regs_pointer.selector, real_regs_pointer.offset,
         _DS, (Bit32)pkt_entry_real_regs, sizeof(pkt_entry_real_regs));
      return real_regs_pointer;
      }
#endif

   // Pop the return address on the real mode stack
   // to the image of the real mode CS and IP registers
   //
   return_address = far_read32( real_stack_pointer );
   my_regs->ip = return_address & 0xFFFF;
   my_regs->cs = return_address >> 16;
   my_regs->sp += 4;

   // If this is the first call...
   //
#define opcode  (my_regs->eax.low16)
#define length  (my_regs->ecx.low16)
#define lah_len (my_regs->edx.low16)
#if PKT_ENTRY_TALKS
   printf( "pkt_entry() sees opcode=%d length=%d lah_len=%d\n", opcode, length, lah_len );
#endif
   if (opcode == 0) {
      // update counters
      //
      ++rx_packets;
      rx_bytes += length;

      // We need to allocate some storage.
      //
      // We could allow the size to be adjusted (as per pkt driver
      // 1.10 spec), but this could cause havoc with earlier drivers.
      // Instead we make sure we return the length as it was given
      // to us.
      //
      // The user's rcv_call1 can still return a null buffer pointer,
      // indicating that the packet is to be discarded.
      //
      lah_buffer = real_mem + real_to_physical(
         my_regs->ds << 16 |
         my_regs->esi.low16);
      buffer = pkt_rcv_call1(length, lah_len,lah_buffer, interface);

      // Convert the flat pointer to real mode
      //
      if (buffer<buffer_base || buffer+length>buffer_base+TOT_BUF_MEM) {
         if (buffer) printf(
            "pkt_entry() internal error: pkt_rcv_call1(length=%d) returned 0x%8.8X"
            " which is outside range 0x%8.8lX to 0x%8.8lX which is visible from real mode\n",
            length, buffer, buffer_base, buffer_base+TOT_BUF_MEM);

         my_regs->es = 0;
         my_regs->edi.low16 = 0;
         }
      else {
         // normalize the pointer so most of it is in the segment,
         // so the ethernet card's 16-bit packet driver doesn't
         // wrap around at a 64K byte boundary, corrupting the
         // beginning of the segment it is on (some other packet's
         // data and possibly the linked list we manage them with!)
         // and causing us to see garbage when we (properly) look
         // beyond the 64K line with our 32-bit code
         //
         // never would have found this if TCP wasn't checksumming
         // the packet beyond the IP header      Joel    17 Jan 97
         //
         my_regs->es = (buffer - real_mem) >> 4;
         my_regs->edi.low16 = (buffer - real_mem) & 15;
         }
      }
   else if (opcode == 1) {

      // Convert the real-mode pointer to one
      // we can use in flat protected mode
      //
      buffer = real_mem + real_to_physical(
         my_regs->ds << 16 |
         my_regs->esi.low16);
      if (buffer < buffer_base || buffer+length > buffer_base+TOT_BUF_MEM) {
         printf("pkt_entry(ptr=0x%4.4X:%4.4X len=%d) internal error: not part of packet buffer pool"
            " ptr=0x%4.4X:0000 len=%d allocated by WAT TCP/IP flat protmode library\n",
            my_regs->ds,
            my_regs->esi.low16,
            length, buffer_segment, TOT_BUF_MEM);
         }
      else {
         // Tag the buffer as filled
         //
         pkt_rcv_call2(length, buffer, interface);
         }
      }
   else
      printf("pkt_entry() error: unrecognized opcode AX=0x%4.4X\n", opcode);

   return real_regs_pointer;
#undef opcode
#undef length
#undef lah_len
   }

int pkt_send(char *buffer, int length)
{
   struct RealModeRegs real_regs;
   int retries;
   unsigned char au_save;
   Bit16 result;

   if (buffer < buffer_base || buffer+length > buffer_base+TOT_BUF_MEM) {
      printf( "pkt_send( 0x%8.8lX, len %d ) internal error: outside range"
         " 0x%8.8lX to 0x%8.8lX which is visible from real mode\n",
         buffer, length, buffer_base, buffer_base+TOT_BUF_MEM);
      return 1;
      }

   real_regs.reserved = 0;
   real_regs.ss = 0;
   real_regs.sp = 0;
   real_regs.flags = 0;

   retries = 500;  // Joel, Feb 97
   // Using a big number of retries means we busy-wait here,
   // hoping the ethernet card can send the packet real soon now
   while (retries--) {
      real_regs.eax.low16 = PD_SEND;
      real_regs.ds = buffer_segment;
      real_regs.esi.low16 = buffer - buffer_base;
      real_regs.ecx.low16 = length;
      result = simulate_real_interrupt(pkt_interrupt, &real_regs);
      if (result != 0) {
         printf( "pkt_send() saw DPMI error 0x%4.4X simulating real mode"
            " interrupt 0x%2.2X\n", result, pkt_interrupt );
         return 1;
         }

      if (real_regs.flags & CARRY) {
         ++low_send_fails;  // Joel, 22 Jan 97
         continue;
         }

      ++tx_packets;  // Update counters
      tx_bytes += length;
      if (au_monitor)  /* Monitor transmitted packets as well as received */
         au_send_mon(length, (unsigned char *)buffer);
      return 0;
      }

   ++send_fails;  // Nevil, 10 Jan 97
   return 1;  /* Failed to send */
   }

void pkt_buf_wipe()  /* Zero out all buffers */
{
   int j;
   struct pbuf *bp;
   freeq.len = readyq.len = 0;
   for (j = 0; j != MAXBUFS; ++j) {  // Link buffers into free queue
      bp = (struct pbuf *)&pktbuf[j];
#if IP_BUF_TRACE
      bp->unused = j;  // 'unused' byte = buffer number
#endif
      bp->state = 0;
      enqueue(&freeq, bp);
      }
   }

void pkt_buf_release(char *ptr)  /* Return one IP buffer to the pool */
{
   struct pbuf *bp;
   bp = (struct pbuf *)(ptr - _pktipofs - BUF_HDR_LEN);
   bp->state = 0;
#if IP_BUF_TRACE
   printf(" {-%u}", bp->unused);
#endif
   enqueue(&freeq, bp);
   }

void *pkt_received(struct bufq *bq)
{
   struct pbuf *bp;
   bp = dequeue(bq);
   if (bp == NULL) return NULL;  /* No packets ready */

#if JOELDEBUG
   printf("pkt_received: chose buffer %d\n", bp->unused);
#endif
   ++found_packets;
   bp->state = 3;  /* In use by application */
#if IP_BUF_TRACE
   printf(" {+%u}", bp->unused);
#endif
   return &bp->data;
   }

void *_pkt_eth_init()
{
   if (pkt_init()) {
      outs("Program halted\n");
      exit(1);
      }
   return eth_addr;
   }
