        TITLE   PCREMOT2 - Unattended Computer Access - 1991 by Terry Lahman
        PAGE    78,132
;======================================================================
;
;  PCREMOT2 - An unattended computer access utility.  Allows access to an
;  unattended computer from a manned computer.  The unattended computer
;  executes the TSR portion of the program and waits for the manned
;  computer to call.  For use with text only programs.  Like using the
;  phone lines and modems as a long extension cord for your keyboard
;  and monitor.
;
;       Usage:  PCREMOT2 [ /M ] [ /# ] [ /B# ] [ /D ] [ /S ] [ /N ] [ /U ] [ /Pstring ] [ /Ystring ] [ /? ]
;
;               [ /Pstring ]  = set password
;               [ /Ystring ]  = set xterm window title and icon name
;               [ /M  ]  = "Manned" mode (used on remote system)
;               [ /#  ]  = 0 - 7, Communication port used
;               [ /B# ]  = Baud rate 0=1200, 1=2400, 2=4800, 3=9600
;                                    4=19200, 5=38400, 6=57600, 7=115200
;               [ /D  ]  = "Desnow" flag (used on CGA monitors)
;               [ /S  ]  = Smiley face displayed (used on host system)
;               [ /N  ]  = Null modem cable connecting 2 systems
;               [ /U  ]  = Uninstall (used on host computer to remove from memory) "
;               [ /?  ]  = Display help screen
;
;======================================================================
;  modified 7/19/96 by Joel Apisdorf of MCI
;  to support 50 line screen mode and remote reboot
;
;  modified 7/24/96 to support 16550A UART and 57600/115200 bauds
;  (figured out 8/5/96 I needed to increase TX buffering from 256 to
;  1024 bytes, and RX buffering from 256 to 512 bytes, to avoid losses)
;  (figured out 8/5/96 I needed to disable FIFO when shelling to DOS
;  so programs like ZCOPY which only understand an 8250 do not lock up
;  either machine)
;
;  modified 7/29/96 to disable password checking and
;  automatic baud rate detection
;
;  modified 8/2/96 to make serial port IRQ the highest priority
;  so remote control still works when somebody else's handler
;  forgets to issue EOI to master 8259 PIC
;
;  modified 8/5/96 to propagate cursor shape along with its position
;  and save/restore/restore cursor shape on manned entry/exit/shell
;
;  modified 8/5/96 to tell host that alt key is not down before
;  shelling to DOS so host user does not have trouble using KB
;  while shell is going on
;
;  modified 8/5/96 to clear serial input buffer when we get back
;  from shelling to DOS so misc screen updates received during shell
;  do not take effect after the full screen update that we force
;
;  modified 8/6/96 to convert characters which the link will not carry
;  (possibly 0, definitely telnet escape 29 and its alias 157, definitely
;  255) into a 2-character sequence that it will, and decode such escapes
;  upon receipt
;
;  modified 8/6/96 so EXIT_CODE was used in all places where user can
;  quit the program, not ESCAPE in some places
;
;  modified 8/6/96 CONNECT_MANNED to check for EXIT_CODE on keyboard
;  where it had tight loops before
;
;  modified 8/6/96 CONNECT_MANNED to go back to looking for NUL if it
;  gets an unrecognized character and has never gotten valid host commport
;  so host starting up after manned is less likely to get stuck
;
;  modified 8/14/96 to also escape XON and XOFF for modem links
;
;  added code 8/26/96 to reset machine by telling the keyboard controller
;  to pulse the CPU reset line, rather than telling the Intel Triton
;  chipset to simulate pressing the reset button, in hopes that serial
;  ports which have AUTO I/O port in AMI BIOS setup (rather than fixed
;  2F8h I/O port) would work afterwards.  It didn't help.
;
;  modified 8/29/96 to display a message and exit when BIOS says I/O port
;  for COM1/COM2 is zero, so user has better feedback when COM ports are
;  disabled by BIOS setup
;
;  modified 6/5/97 to do a simple dumb terminal emulator when manned,
;  until the user presses the clear key, so you don't need to run a
;  separate program to hop through the network and enter passwords,
;  then exit it, and run this one, hoping you either didn't drop DTR
;  or the modem didn't care!   Also helps when your terminal emulator
;  cannot do higher speeds
;
;  modified 6/9/97 to send a special code when screen has scrolled
;  by N lines, rather than resending the whole screen, so more than
;  just the top few lines will update on controlling PC when lots of
;  scrolling is going on
;
;  modified 6/10/97 to calculate output buffer fullness in unattended
;  mode, rather than iterating until desired bytes found in buffer,
;  for speed when output buffer is large
;
;  modified 2/4/98 to ignore blocks beyond the end of the screen in
;  manned mode, to avoid corrupting program memory when link is noisy
;
;  modified 2/5/98 to stop comparing the commandline argument character
;  when it matches "/1" or "/2" or "/3", not just when "/4", so explicit
;  requests for those other ports can be in the middle of the commandline
;  without confusing the parser
;
;  modified 2/5/98 to not increment commandline argument pointer and
;  decrement argument byte count, after parsing the baud rate argument,
;  so it doesn't attempt to parse and extra 256 garbage chars when baud
;  rate is the last argument
;
;  modified 2/5/98 to only test the least significant bit of SMILE_FLAG,
;  NULL_MODEM, DESNOW_FLAG, and MANNED_FLAG, and to flip that bit
;  (instead of incrementing the byte) during parsing, so they can be
;  toggled by commandline argument regardless of their default state
;
;  modified 2/6/98 to detect scrolls that start after the first block,
;  and added a new scroll packet 0xE1 to include a destination address
;  in addition to source and length fields of old 0xE0 scroll packet
;
;  got rid of redundant "BYTE PTR" in front of references to variables
;  declared with sizes, to help comment columns align
;
;  2/10/98 added a compile-time-selected mode for unattended PC to send
;  and receive ANSI escape sequences instead of proprietary packets, so
;  the PC can be controlled from a dumb terminal, or better yet, a UNIX
;  host running xterm (although X11R6.3 xterm doesn't understand the color
;  subparameters 30-37 and 40-47 of the set graphic rendition escape, and
;  its font probably doesn't match PC codepage 850 or 437 which has umlauts
;  and linedrawing characters)
;
;  don't know how to allow scrolling that starts after the first line,
;  so that must be disabled for now
;
;  cannot escape any characters, since the manned host will not understand
;
;  moved clearing of video shadow for unattended PC, from initialization-
;  time to connection-time, so the common routing CLEAR_VIDEO could be
;  called, and moved the code to clear the dumb terminal's screen into
;  that subroutine so it could be common with processing of the clear key
;
;  2/12/98 found out that color_xterm from Jacob L. Mandelson implements
;  the above SGR escape color subparameters 30-37 and 40-47
;
;  2/15/98 optimized SEND_CURSOR_ESCAPE to not bother to send the semicolon
;  and the digits for the column, when the column would be 1, since the
;  dumb terminal will assume that for us
;
;  2/16/98 implemented SEND_SCROLL for dumb ANSI terminals by moving cursor
;  to the destination line of scroll and deleting length lines (requires
;  scroll_clears_lines to be 1 to match dumb terminal behavior)
;
;  2/16/98 updated scroll_clears_lines code for unattended PC to work when
;  dest is not first line
;
;  2/16/98 needed to also set color attributes on dumb ANSI terminal
;  when clearing video
;
;  2/16/98 added code to keep track of row and column for block as it moves
;  if manned is dumb ANSI, to get rid of most of the multiplying and dividing
;  to convert bytes to rows and column
;
;  2/16/98 added code to only treat whole lines as scrolls, not portions,
;  since we chose to only tell the dumb ANSI terminal to insert/delete
;  whole lines; it turned out to be most efficient to go back to the
;  beginning of the line and transfer blocks from our shadow rather than
;  copying them out of video RAM again
;
;  2/17/98 treated 255 as unprintable in TRANSFER_BLOCK because telnet doesn't
;  like to receive it
;
;  optimized SEND_CURSOR_ESCAPE to not bother to send either row, semicolon,
;  or column, when both row and column would be 1, since the dumb terminal
;  will assume that for us
;
;  got rid of redundant vidcopy/compare cycle after transferring a miscompared
;  block, by changing the code at the end of UA_120 to move forward a block
;  and maybe fall through to UA_EXIT, instead of always jumping to UA_30
;  (required factoring out the change-the-block code that used to be between
;  UA_112 and UA_115, into a new subroutine UA_CHANGE_VID_PTR)
;
;  changed code that recognizes less-than-a-line scrolls to go back and try
;  a larger scroll, instead of transferring the block, because it was wasting
;  bandwidth during definite scrolls of mostly-similar lines, i.e. PKUNZIP
;  with long directory paths
;
;  remapped the PC linedrawing characters and boxes to DEC equivalents
;
;  implemented a commandline switch /Y which sets the (new feature) string
;  sent to xterm during connection to change title bar and icon text (also
;  changed to "disconnected" after exit code received)
;
;  2/26/98 fixed bugs: lines on PC with dot caused bell and shortened line
;  on xterm, and xterm scrolled when PC put cursor offscreen
;
;  mapped diamond to the proper linedrawing character of xterm (DEC VT100)
;
;  2/27/98 mapped PC reverse/normal dot/circle to xterm small zero, and
;  mapped up/down/left/right triangle/arrow to ^/v/</>
;
;  added compile-time switch to force background color to be different from
;  foreground, since xterm doesn't use a brighter color for the bold attribute
;  (it makes the strokes of the font thicker horizontally) and doesn't
;  implement the blink attribute at all
;
;  3/3/98 removed redundant push/usage/pop of SI register in PUT_KEY_DATA
;
;  changed PUT_KEY_DATA to set keycode byte when ASCII is CR or TAB
;
;  moved KEYCODE_ENTER and KEYCODE_TAB from KBSTUFF.ASM to this module
;
;  created new subroutine SEND_ATTRIBUTE_ESCAPE to simplify TRANSFER_BLOCK
;
;  changed SEND_SCROLL to define a scroll region, move cursor to last line
;  of it, send many linefeeds, and remove the scroll region, instead of
;  moving the cursor to the top of the region, deleting lines, moving cursor
;  to bottom, and inserting lines, all so that lines below those scrolled
;  (such as T editor status line) don't flash up and then down again
;
;  3/4/98 fixed password checking to not lock up machine until password OK
;
;  new code to wink DTR, RTS, and OUT2 when exiting or bad password
;
;  changed CHECK_CARRIER to look at DTR regardless of NULL_MODEM flag
;
;  changed IP_95 and CLEAR_VIDEO to call LOAD_ZSTRING instead of doing
;  basically the same thing themselves inline
;
;  removed call to CLEAR_OUTBUF and assignment of 0 to COMM_INT_STA
;  which used to be at IP_95, since output buffer hasn't finished draining
;  until several timer ticks later, when CU_HANGUP_DELAY runs
;
;  implemented a commandline switch /P which sets the password
;
;  3/6/98 added a compile-time switch respect_control_lines to control
;  whether CHECK_CARRIER will ignore control lines for null_modem connections,
;  so imperfect cables can be used
;
;  4/29/98 removed code in INPUT_PROCESSING which buffered serial input
;  until BIOS KB buffer has room, so reboot and other special codes can
;  always be recognized, and changed PUT_KEY_DATA to send a BELL character
;  if KB buffer is full, since such input will now be ignored
;
;  4/29/98 changed INPUT_PROCESSING to allow all 4 shifts for tab like delete,
;  and took out the special-case code for tab in PUT_KEY_DATA, so shift-tab
;  outputs the correct keycode now
;
;======================================================================
CSEG    SEGMENT PARA PUBLIC 'CODE'
        ASSUME  CS:CSEG,DS:NOTHING,ES:NOTHING,SS:NOTHING

; compile-time switches
;
debug_unattended        equ     0               ; pretend output buffer never full, always connected

respect_control_lines   equ     1
ask_for_password        equ     0
force_fg_ne_bg          equ     1
slash_delimits_strings  equ     1               ; otherwise whitespace is enough
allow_pc_chars          equ     0
manned_is_dumb_ansi     equ     1
detect_scrolling        equ     1
scroll_clears_lines     equ     1
escape_bad_chars        equ     0
baud_rate_detection     equ     0
simple_dumb_terminal    equ     1
enable_uart_fifos       equ     1
force_irq_priorities    equ     1

mask_all_when_reboot    equ     0
seoi_all_when_reboot    equ     0
disable_com_when_reboot equ     1
disable_kb_when_reboot  equ     0
sysflag_0_when_reboot   equ     0
ctrl_alt_del_reboot     equ     0               ; OK for AUTO COM2, doesn't reset PCI cards
kb_controller_reboot    equ     0               ; OK for AUTO COM2, doesn't reset PCI cards
triton_reboot_hard      equ     1               ; hard bad for AUTO COM2, resets PCI cards OK
triton_reboot_soft      equ     0               ; soft OK for AUTO COM2, doesn't reset PCI cards

lockup_when_int_comm    equ     0
beep_when_int_comm      equ     0
beep_when_rx_overruns   equ     0               ; should be zero for production host
beep_when_tx_overruns   equ     0               ; should be zero for production host
reinit_when_tx_overruns equ     0
revector_when_tx_overruns equ     0
reinit_when_timer       equ     0

; control characters that work in either environment
;
BELL            EQU     7D
BACKSPACE       EQU     8D
TAB             EQU     9D
LF              EQU     10D
VT              EQU     11D
FORMFEED        EQU     12D
CR              EQU     13D
SHIFTOUT        EQU     14D
SHIFTIN         EQU     15D
XON             EQU     17D
XOFF            EQU     19D
ESC_KEY         EQU     27D
TELNET_ESCAPE   EQU     29D
SPACE           EQU     32D
EXCLAIM         EQU     33D
QUOTE           EQU     39D
BACKQUOTE       EQU     96D
DELETE          EQU     127D
HIGH_SPACE      EQU     (SPACE + 128D)
HIGH_DELETE     EQU     (DELETE + 128D)

; graphic characters that are only in the VT100/XTERM unshifted font
;
XTERM_UPSIDE_EXCLAIM    EQU     161D
XTERM_MONEY_CENTS       EQU     162D
XTERM_MONEY_POUNDS      EQU     163D
XTERM_MONEY_YEN         EQU     165D
XTERM_SECTION_SIGN      EQU     167D
XTERM_COPYRIGHT_SIGN    EQU     169D
XTERM_MATH_MUCHLESS     EQU     171D
XTERM_MATH_NEGATION     EQU     172D
XTERM_SMALL_ZERO        EQU     176D
XTERM_SMALL_TWO         EQU     178D
XTERM_SMALL_THREE       EQU     179D
XTERM_PARAGRAPH_SIGN    EQU     182D
XTERM_TINY_DOT          EQU     183D
XTERM_SMALL_ONE         EQU     185D
XTERM_MATH_MUCHGREATER  EQU     187D
XTERM_UPSIDE_QUESTION   EQU     191D
XTERM_GREEK_PHI         EQU     248D
XTERM_LINEDRAW_DIAMOND  EQU     BACKQUOTE

; graphic characters that are only in the PC textmode font
;
SMILE                   EQU     1D                      ;''
REV_SMILE               EQU     2D                      ;  ''
HEART                   EQU     3D                      ;''
DIAMOND                 EQU     4D                      ;  ''
CLUB                    EQU     5D                      ;''
SPADE                   EQU     6D                      ;  ''
DOT                     EQU     7D                      ;''
REV_DOT                 EQU     8D                      ;  ''
CIRCLE                  EQU     9D                      ;
REV_CIRCLE              EQU     10D                     ;
MALE_SIGN               EQU     11D                     ;''
FEMALE_SIGN             EQU     12D                     ;  ''
HALF_NOTE               EQU     13D                     ;'
'
FULL_NOTE               EQU     14D                     ;  ''
BIG_STAR                EQU     15D                     ;''
RIGHT_TRIANGLE          EQU     16D                     ;  ''
LEFT_TRIANGLE           EQU     17D                     ;''
UPDOWN_ARROWS           EQU     18D                     ;  ''
DOUBLE_EXCLAIM          EQU     19D                     ;''
PARAGRAPH_SIGN          EQU     20D                     ;  ''
SECTION_SIGN            EQU     21D                     ;''
DOWN_PART_BOX           EQU     22D                     ;  ''
UL_UPDOWN_ARROWS        EQU     23D                     ;''
UP_ARROW                EQU     24D                     ;  ''
DOWN_ARROW              EQU     25D                     ;''
RIGHT_ARROW             EQU     26D                     ;  ''
LEFT_ARROW              EQU     27D                     ;''
MATH_UNKNOWN1           EQU     28D     ;ISN'T NEGATION ;  ''
LEFTRIGHT_ARROWS        EQU     29D                     ;''
UP_TRIANGLE             EQU     30D                     ;  ''
DOWN_TRIANGLE           EQU     31D                     ;''

MONEY_CENTS             EQU     155D                    ;   ''
MONEY_POUNDS            EQU     156D                    ; ''
MONEY_YEN               EQU     157D                    ;   ''
MONEY_P_SUB_LC_T        EQU     158D   ;WHAT IS THIS??? ; ''
MONEY_FRANCS            EQU     159D   ;I think...      ;   ''
UPSIDE_QUESTION         EQU     168D   ;has a name...   ; ''
MATH_UNKNOWN2           EQU     169D       ;ISN'T NEG   ;   ''
MATH_NEGATION           EQU     170D                    ; ''
ONE_HALF                EQU     171D                    ;   ''
ONE_QUARTER             EQU     172D                    ; ''
UPSIDE_EXCLAIM          EQU     173D   ;has a name...   ;   ''
MATH_MUCHLESS           EQU     174D                    ; ''
MATH_MUCHGREATER        EQU     175D                    ;   ''
LIGHT_GRAY_BOX          EQU     176D                    ; ''
MED_GRAY_BOX            EQU     177D                    ;   ''
DARK_GRAY_BOX           EQU     178D                    ; ''
FULL_BOX                EQU     219D                    ;   ''
DOWN_FULL_BOX           EQU     220D                    ; ''
LEFT_FULL_BOX           EQU     221D                    ;   ''
RIGHT_FULL_BOX          EQU     222D                    ; ''
UP_FULL_BOX             EQU     223D                    ;   ''
GREEK_ALPHA             EQU     224D                    ; ''
GREEK_BETA              EQU     225D                    ;   ''
GREEK_TAU               EQU     226D    ;csong says...  ; ''
GREEK_PI                EQU     227D                    ;   ''
GREEK_SIGMA             EQU     228D                    ; ''
GREEK_LC_SIGMA          EQU     229D                    ;   ''
GREEK_MU                EQU     230D                    ; ''
GREEK_GAMMA             EQU     231D                    ;   ''
GREEK_PSI               EQU     232D    ;OR MAYBE PHI?? ; ''
                     ;I am told theta looks more like an O with / through it
GREEK_THETA             EQU     233D                    ;   ''
GREEK_OMEGA             EQU     234D                    ; ''
GREEK_DELTA             EQU     235D                    ;   ''
INFINITY                EQU     236D                    ; ''
GREEK_PHI               EQU     237D                    ;   ''
SMALL_ZERO              EQU     248D                    ; ''
TINY_DOT                EQU     250D                    ;   ''
MATH_SQUAREROOT         EQU     251D                    ; ''
SMALL_TWO               EQU     253D                    ;   ''
CENTER_PART_BOX         EQU     254D                    ; ''

; scancodes for a few important, language-independent, keys
;
KEYCODE_ESCAPE  equ     01h
KEYCODE_ENTER   equ     1Ch                     ;on numeric keypad would be 0E0h
KEYCODE_TAB     equ     0Fh
KEYCODE_SPACE   equ     29h
KEYCODE_C       equ     2Eh
KEYCODE_D       equ     20h

; primary manifest constants
;
BAD_OFFSET      EQU     50D                     ;Subtracted from char when escaped
BAD_ESCAPE      EQU     253D                    ;Prefixes a character when escaped
CHARS_PER_LINE  EQU     80D
LINES_PER_SCR   EQU     50D
STACK_SIZE      EQU     256D
IN_BUFF_SIZE    EQU     512D                    ;Size for input buffer
OUT_BUFF_SIZE   EQU     2048D                   ;Size for output buffer
BLOCK_SIZE      EQU     16D                     ;Block transfer size in words
UNKNOWN_ATTR    EQU     255D                    ;high-intensity blinking white-on-white
UNKNOWN_GRAPHIC EQU     2                       ;neither 0 nor 1

; derived values
;
BLOCK_SIZEX2    EQU     (BLOCK_SIZE * 2)        ;Block transfer size in bytes
BYTES_PER_LINE  EQU     (CHARS_PER_LINE * 2)
LAST_SCR_LINE   EQU     (LINES_PER_SCR - 1)
CHARS_PER_SCR   EQU     (LINES_PER_SCR * CHARS_PER_LINE)
BYTES_PER_SCR   EQU     (CHARS_PER_SCR * 2)
MAX_BLOCK_NUM   EQU     (CHARS_PER_SCR / BLOCK_SIZE)
UNKNOWN_POS     EQU     ((LINES_PER_SCR * 256) + CHARS_PER_LINE)

;----------------------------------------------------------------------
;  Start of code
;----------------------------------------------------------------------
                ORG     100H                    ;Starting offset for .com
START:          JMP     INITIALIZE              ;Jump over resident code
                DW      OFFSET CONFIG - $       ;Offset to config parameters

;----------------------------------------------------------------------
;  Data storage
;----------------------------------------------------------------------

COPYRIGHT       DB      "PCREMOT2 Version 1.0.40 (c) 1989, 1991 Ziff Communications Co."
                DB      CR,LF,"        PC Magazine ",254," by Terry Lahman & Kevin Sims"
                DB      CR,LF,"last modified 4/29/98 by Joel Apisdorf"
                DB      CR,LF,"$",26
IDT_LIMIT       DW      256D * 4
IDT_BASE        DD      0
RX_OVERRUNS     DW      0
TX_OVERRUNS     DW      0
CONNECT_FLAG    DB      0                       ;0=not connected
                                                ;1=connected to manned system
AUTOBAUD_FLAG   DB      0                       ;nonzero means saw enough carriage returns
PASSWORD_BAD    DB      ?                       ;0=No, don't send exit code
                                                ;1=Yes, 3 bad, so send code
PASSWORD_TRIES  DB      0                       ;how many password tries so far?
PASSWORD_MAX    EQU     100                     ;max password size, must be <256
        if ask_for_password
PASSWORD_FLAG   DB      0                       ;password length + 1
PASSWORD_MSG    DB      CR,LF,"Enter password:",0
PASSWORD_MSG2   DB      CR,LF,"Get lost!",CR,LF,LF,0
PASSWORD_BUFFER DB      PASSWORD_MAX DUP( 0 )
                DB      0                       ;room for terminating NUL
        endif
ACT_FLAG        DB      0                       ;0=unattended prog not active
                                                ;1=unattended program is active
                                                ; do not run again
OLDINT_8        DW      0,0                     ;Old timer vector
OLDINT_21       DW      0,0                     ;Old DOS vector
OLDINT_COMM     DW      0,0                     ;Old communications vector
INSTALLED_SEG   DW      0                       ;Segment of resident program
SET_VEC         DW      2500H                   ;DOS set vector call
SLEEP_FLAG      DB      0                       ;0=pcremote is awake
                                                ;1=pcremote is asleep
                                                ;needs to sleep during init.
CURRENT_SPEED   DB      0                       ;Currently selected speed
SPEED_COUNT     DB      0                       ;Change speed once a second
COMM_PORT       DW      0                       ;Comm port address
COMM_INT_STA    DB      0                       ;Comm port interrupt status
                                                ;0=not transmitting data
                                                ;1=transmitting data
BAD_CHARS       DW      0                       ;Address of table of chars needing escaping
VIDEO_COPY      DW      0                       ;Address of video ram copy
VIDEO_SEGMENT   DW      0                       ;Segment register for vid ram
STACK_TOP       DW      INITIALIZE+STACK_SIZE   ;Top of stack for unattended
OLD_SS          DW      0                       ;Old stack segment
OLD_SP          DW      0                       ;Old stack pointer
TEMP_REG        DW      0                       ;Temp storage for register
CURSOR_POSITION DW      0                       ;Old cursor position (column in lsb, both zero-based)
CURSOR_SHAPE    DW      0                       ;Old cursor shape
        if manned_is_dumb_ansi
PREV_POSITION   DW      UNKNOWN_POS             ;where is cursor now on dumb terminal?
PREV_ATTRIBUTE  DB      UNKNOWN_ATTR            ;color/blink/bold of last printable char output
PREV_GRAPHIC    DB      UNKNOWN_GRAPHIC         ;was previous character graphic? 0=no 1=yes 2=unknown

COLOR_XLATE     DB      0, 4, 2, 6, 1, 5, 3, 7  ;array to convert 8 PC colors to ANSI

LINEDRAW_XLATE:
                ; DEC linedrawing characters for PC characters
                ; in the 176-223 range (gray full blocks,
                ; single and double and mixed linedrawing,
                ; dark half blocks)
                ;
                ; note that the DEC characters are only single linedrawing
                ; and half gray, so this can only be a one-way mapping
                ;
                ; we must output ESC_KEY ( 0 during initialization
                ; to designate the DEC special characters and linedrawing
                ; as font G0, and we must output SHIFTOUT each time we want
                ; them to appear as BACKQUOTE, a, and j through x
                ; and we must output SHIFTIN to get normal lowercase back
                ;
                ; 176                                             
                ;
                ; 183                                             
                ;
                ; 190                                             
                ;
                ; 197                                             
                ;
                ; 204                                             
                ;
                ; 211                                             
                ;
                ; 218                                      
                ;
                DB      'a',    'a',    'a',    'x',    'u',    'u',    'u'
                DB      'k',    'k',    'u',    'x',    'k',    'j',    'j'
                DB      'j',    'k',    'm',    'v',    'w',    't',    'q'
                DB      'n',    't',    't',    'm',    'l',    'v',    'w'
                DB      't',    'q',    'n',    'v',    'v',    'w',    'w'
                DB      'm',    'm',    'l',    'l',    'n',    'n',    'j'
                DB      'l',    'a',    'q',    'x',    'x',    'q'

                ; XTERM equivalents of PC characters, used when the XTERM
                ; doesn't require using the linedrawing character set
                ;
                ; initialized by INIT_XLATE
                ;
                ; these are pointers, not the tables themselves
                ;
FONT_XLATE          DW      ?
FONT_XLATE_ATTR     DW      ?

                ; attribute bits coded into the above translation table
                ;
USE_NOTHING         EQU     0
USE_GRAPHIC_FONT    EQU     1
USE_REVERSE_VIDEO   EQU     2

ARROW_XLATE:
                ; PC keycodes for the arrow keys
                ;
                ; ANSI escapes are ESC [ A/B/C/D
                ;
                ;       up          down        right       left
                DB      224,72,     224,80,     224,77,     224,75  ; unshifted
                DB      224,72,     224,80,     224,77,     224,75  ; SHIFT
                DB      224,141,    224,145,    224,116,    224,115 ; CTRL
                DB      224,152,    224,154,    224,157,    224,155 ; ALT

MOTION_XLATE:
                ; PC keycodes for the keys insert, home, end, pageup and pagedown
                ;
                ; ANSI escapes are ESC [ up-to-3-digits tilde
                ; (same sort of sequence as function keys)
                ; where the number is between 2 and 6 inclusive
                ;

                ; normal motion keys:
                ;       insert      home        end         pageup      pagedown
                ;
                DB      224,82,     224,71,     224,79,     224,73,     224,81

                ; SHIFT motion keys
                ;
                DB      224,82,     224,71,     224,79,     224,73,     224,81

                ; CTRL motion keys
                ;
                DB      224,146,    224,119,    224,117,    224,132,    224,118

                ; ALT motion keys
                ;
                DB      224,162,    224,151,    224,159,    224,153,    224,161

DELETE_XLATE:
                ; PC keycodes for the delete key by itself,
                ; since it outputs a single value instead of
                ; an escape sequence
                ;
                ;       normal      SHIFT       CTRL        ALT
                ;
                DB      224,83,     224,83,     224,147,    224,163

TAB_XLATE:
                ; PC keycodes for the tab key by itself,
                ; since it outputs a single value instead of
                ; an escape sequence
                ;
                ;       normal      SHIFT       CTRL        ALT
                ;
                DB      9,15,       0,15,       0,148,      0,165

FUNC_XLATE:
                ; PC keycodes for the function keys
                ;
                ; ANSI escapes are ESC [ up-to-3-digits tilde
                ; (same sort of sequence as function keys)
                ; where the number is between 11 and 24 inclusive
                ;

                ; normal function keys:
                ;       F1          F2          F3          F4          F5
                ;       empty
                ;       F6          F7          F8          F9          F10
                ;       empty
                ;       F11         F12
                ;
                DB      0,59,       0,60,       0,61,       0,62,       0,63
                DB      ?,?
                DB      0,64,       0,65,       0,66,       0,67,       0,68
                DB      ?,?
                DB      0,133,      0,134

                ; SHIFT function keys
                ;
                DB      0,84,       0,85,       0,86,       0,87,       0,88
                DB      ?,?
                DB      0,89,       0,90,       0,91,       0,92,       0,93
                DB      ?,?
                DB      0,135,      0,136   ; SHIFT

                ; CTRL function keys
                ;
                DB      0,94,       0,95,       0,96,       0,97,       0,98
                DB      ?,?
                DB      0,99,       0,100,      0,101,      0,102,      0,103
                DB      ?,?
                DB      0,137,      0,138

                ; ALT function keys
                ;
                DB      0,104,      0,105,      0,106,      0,107,      0,108
                DB      ?,?
                DB      0,109,      0,110,      0,111,      0,112,      0,113
                DB      ?,?
                DB      0,139,      0,140

ALPHA_XLATE:
                ; PC keycodes for ALT ASCII keys
                ;
                ; normal, SHIFT, and CTRL from the manned host
                ; are already OK to put in PC keyboard buffer
                ; (except for CTRL-BREAK and CTRL-PRTSC and CTRL-TAB,
                ; which can easily be special-cased)
                ; so they are not looked up in this table
                ;
                ; don't forget to convert lowercase to uppercase
                ; by masking 0x20 off (AND'ing with 0xDF)
                ;

                ;       A       B       C       D       E       F       G
                ;       H       I       J       K       L       M       N
                ;       O       P       Q       R       S       T       U
                ;       V       W       X       Y       Z       [       \
                ;       ]
                DB      0,30,   0,48,   0,46,   0,32,   0,18,   0,33,   0,34
                DB      0,35,   0,23,   0,36,   0,37,   0,38,   0,50,   0,49
                DB      0,24,   0,25,   0,16,   0,19,   0,31,   0,20,   0,22
                DB      0,47,   0,17,   0,45,   0,21,   0,44,   0,26,   0,43
                DB      0,27

DIGIT_XLATE:
                ; just the ALT keycodes for digits
                ;
                ;       0       1       2       3       4       5       6
                ;       7       8       9
                ;
                DB      0,129,  0,120,  0,121,  0,122,  0,123,  0,124,  0,125
                DB      0,126,  0,127,  0,128

ENTER_XLATE:    DB      0,28                    ; just ALT enter keycode
DASH_XLATE:     DB      0,130                   ; just ALT dash (hyphen) keycode
COMMA_XLATE:    DB      0,51                    ; just ALT comma keycode
PERIOD_XLATE:   DB      0,52                    ; just ALT comma keycode
SLASH_XLATE:    DB      0,53                    ; just ALT / keycode
QUOTE_XLATE:    DB      0,41                    ; just ALT quotation mark keycode
EQUALS_XLATE:   DB      0,131                   ; just ALT equals-sign keycode
SEMICOLON_XLATE: DB     0,39                    ; just ALT semicolon keycode
BACKQUOTE_XLATE: DB     0,40                    ; just ALT backwards quotation mark keycode

        endif
VID_RAM_OFFSET  DW      0                       ;Current compare offset
ASCII_FLAG      DB      0                       ;0=no data is pending
        if manned_is_dumb_ansi
                                                ;1=received ESC_KEY
                                                ;2=received [ or ^G
                                                ;3=got 1 ASCII digit
                                                ;4=got 2 ASCII digits
                                                ;5=got 3 ASCII digits
                                                ;6=got >3 ASCII digits (too many) or semicolon
        else
                                                ;1=received a FE
                                                ;2=received ASCII char
        endif
        if manned_is_dumb_ansi
SHIFT_FLAG      DB      0                       ;0=next char as sent
                                                ;1=next char should be SHIFT'ed
                                                ;2=next char should be CTRL'ed
                                                ;3=next char should be ALT'ed
        else
SHIFT_FLAG      DB      0                       ;0=no data is pending
                                                ;1=received a FD
        endif
KEY_ONE         DB      0                       ;ASCII char received
KEY_TWO         DB      0                       ;Scan code received
        if manned_is_dumb_ansi
KEY_DIGITS      DB      ?,?,?,?                 ;ASCII digits for function key
        endif
BAUD_RATE       DB      0,60H                   ;1200 BAUD divisor MSB LSB
                DB      0,30H                   ;2400 BAUD divisor MSB LSB
                DB      0,18H                   ;4800 BAUD divisor MSB LSB
                DB      0,0CH                   ;9600 BAUD divisor MSB LSB
                DB      0,06H                   ;19200 BAUD divisor MSB LSB
                DB      0,03H                   ;38400 BAUD divisor MSB LSB
                DB      0,02H                   ;57600 BAUD divisor MSB LSB
                DB      0,01H                   ;115200 BAUD divisor MSB LSB
IN_BUFF_HEAD    DW      ?                       ;Pointer to input buffer head
IN_BUFF_TAIL    DW      ?                       ;Pointer to input buffer tail
IN_BUFF_BEGIN   DW      ?                       ;Pointer to input buffer begin
IN_BUFF_END     DW      ?                       ;Pointer to input buffer end
OUT_BUFF_HEAD   DW      ?                       ;Pointer to output buffer head
OUT_BUFF_TAIL   DW      ?                       ;Pointer to output buffer tail
OUT_BUFF_BEGIN  DW      ?                       ;Pointer to output buffer begin
OUT_BUFF_END    DW      ?                       ;Pointer to output buffer end
BLOCK_COUNT     DB      ?                       ;Block number being processed
TEMP_VIDEO_PTR  DW      ?                       ;Pointer to temp video storage
CR_COUNT        DB      ?                       ;Number of CRs for speed sync
MODEM_ATTENTION DB      CR,"AT",CR,0
MODEM_SETUP1    DB      "ATE0 S12=40 Q0 V0 X1 S0=0",CR,0 ;Manned
MODEM_SETUP2    DB      "ATE0 S12=40 Q1S0=2",CR,0 ;Unattended
MODEM_SETUP3    DB      "AT&C1",CR,0            ;Enable DCD on 2400 baud modem
MODEM_SETUP4    DB      "ATW0X4&Q5&K3L0",CR,0   ;for 9600 baud modems
MODEM_HANGUP    DB      "ATH0",CR,0
MODEM_ESCAPE    DB      "+++",0
        if detect_scrolling
SCROLL_DEST     DW      ?                       ;where on screen to copy to?
SCROLL_OFFSET   DW      0                       ;how many bytes have scrolled?
SCROLL_DEST_ROW DB      ?                       ;what row is SCROLL_DEST?
SCROLL_DEST_BLOCK  DB   0                       ;in case we don't match the whole line
SCROLL_OFFSET_ROW  DB   0                       ;how many rows is SCROLL_OFFSET?
SCROLL_FOUND    DB      0                       ;did we try to detect scrolling?
BLOCK_COLUMN    DB      0                       ;horizontal pos of current block
BLOCK_ROW       DB      0                       ;vertical pos of current block
MAX_SCROLL      EQU     (BYTES_PER_SCR-BYTES_PER_LINE)  ;most scrolled bytes we can detect
;MAX_SCROLL      EQU     (2*BYTES_PER_LINE)     ;most scrolled bytes we can detect
;MAX_SCROLL      EQU     (BYTES_PER_LINE)        ;most scrolled bytes we can detect
;MAX_SCROLL      EQU     0                      ;most scrolled bytes we can detect
        endif
CONFIG:         ;The following parameters must remain in order to support
                ; the setup program.
MANNED_FLAG     DB      0                       ;0=Operate in unattended mode
                                                ;1=Operate in manned mode
COMM_FLAG       DB      1                       ;0=Use comm1
                                                ;1=Use comm2
                                                ;2=Use comm3
                                                ;3=Use comm4
COMM_PORT3      DW      03E8H                   ;Port address for COMM 3
COMM_PORT4      DW      02E8H                   ;Port address for COMM 4
COMM3_INT       DB      4                       ;Interrupt for com1 and 3
COMM4_INT       DB      3                       ;Interrupt for com2 and 4
SPEED_FLAG      DB      5                       ;0=Use 1200 baud
                                                ;1=Use 2400 baud
                                                ;2=Use 4800 baud
                                                ;3=Use 9600 baud
                                                ;4=Use 19200 baud
                                                ;5=Use 38400 baud
                                                ;6=Use 57600 baud
                                                ;7=Use 115200 baud
DESNOW_FLAG     DB      0                       ;0=Do not use desnow code
                                                ;1=Use desnow code
NULL_MODEM      DB      1                       ;0=Using a modem
                                                ;1=Using a null modem cable
SMILE_FLAG      DB      1                       ;0=Disable corner smile face
                                                ;1=Enable corner smile face
        if manned_is_dumb_ansi
CLEAR_CODE      DW      6C07h                   ;^G l, code to clear and resend screen
EXIT_CODE       DW      7807h                   ;^G x, code to exit program
REBOOT_CODE     DW      7207h                   ;^G r, code to reboot remote machine
        else
CLEAR_CODE      DW      2E00h                   ;alt-c, code to clear and resend screen
EXIT_CODE       DW      2D00h                   ;alt-x, code to exit program
REBOOT_CODE     DW      1300h                   ;alt-r, code to reboot remote machine
        endif
SHELL_CODE      DW      1F00h                   ;alt-s, shell to DOS code
TRANSFER_CODE   DW      1400h                   ;alt-t, file transfer code
PASSWORD        DB      "coral"
PASSWORD_SIZE   EQU     $ - PASSWORD
PASSWORD_PAD    DB      PASSWORD_MAX-PASSWORD_SIZE DUP(?)
PASSWORD_END    DB      0                       ;room for the terminating NUL
MODEM_SETUP5    DB      30 DUP(0);              ;Extra modem setup string
TONE_DIAL       DB      "ATDT",0                ;Tone dial command
        if manned_is_dumb_ansi
SHIFT_CODE      DW      7307h                   ;^G s, SHIFT next character
CTRL_CODE       DW      6307h                   ;^G c, CTRL  next character
ALT_CODE        DW      6107h                   ;^G a, ALT   next character
SYSTEM_NAME     DB      "OCxMON.unknown"
SYSTEM_NAME_MAX EQU     255
SYSTEM_NAME_PAD DB      (SYSTEM_NAME_MAX - ($ - SYSTEM_NAME)) DUP( 0 )
SYSTEM_NAME_END DB      0                       ;room for the terminating NUL
OFFLINE_NAME    DB      "disconnected",0
        endif

;======================================================================
;  Interrupt handlers.  Interrupt 8 is used in unattended mode only.
;  The communications interrupt is used in both modes.
;======================================================================
;----------------------------------------------------------------------
;  Interrupt 8 handling routine.  If program is already active do not run.
;  Run connect unattended if not connected, otherwise run unattended.
;----------------------------------------------------------------------
INT8            PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    DS                      ;Save data segment
                PUSH    CS                      ;Set data segment to code seg
                POP     DS
                PUSHF                           ;Call old int 8
                CALL    DWORD PTR OLDINT_8
I8_10:
                CLI                             ;Disable interrupts
                CMP     SLEEP_FLAG,1            ;If another program has control
                JZ      I8_EXIT                 ;then exit
                CMP     ACT_FLAG,0              ;Check for program active
                JNZ     I8_EXIT                 ;Exit if progam is active
                INC     ACT_FLAG                ;Set program active flag
                STI                             ;Enable interrupts
                CALL    SET_STACK               ;Create stack & save registers
        if reinit_when_timer
                CALL    INIT_SERIAL
        endif
        if debug_unattended
        else
                CMP     CONNECT_FLAG,0          ;Connected to manned?
                JNZ     I8_20                   ;Yes, then run unattended
                CALL    CONNECT_UNATTENDED      ;No, check for connection
                JMP     I8_30
        endif
I8_20:
                CALL    UNATTENDED              ;Run unattended
I8_30:
                CALL    RESET_STACK             ;Restore registers & stack
                DEC     ACT_FLAG                ;Clear program active flag
I8_EXIT:
                STI                             ;Enable interrupts
                POP     DS                      ;Restore data segment
                IRET
INT8            ENDP

;----------------------------------------------------------------------
;  Interrupt 21 handling routine.  If another program, zcopy, changes
;  the comm interrupt, go to sleep by setting the sleep flag.  When the
;  comm interrupt vector returns, WAKE UP!
;----------------------------------------------------------------------
INT21           PROC    FAR
                PUSHF                           ;save registers
                PUSH    AX
                PUSH    BX
                CMP     AX,WORD PTR CS:SET_VEC  ;setting comm int. vector?
                JZ      I21_20                  ;yes
I21_10:
                POP     BX                      ;no, restore registers
                POP     AX
                POPF
                JMP     DWORD PTR CS:OLDINT_21  ;execute original DOS int
I21_20:
                MOV     AX,CS                   ;is new code segment
                MOV     BX,DS                   ; the same as pcremote
                CMP     AX,BX
                JNZ     I21_30                  ;no, then go to sleep
                CMP     DX,OFFSET CS:INT_COMM   ;is it my comm interrupt vec?
                JZ      I21_40                  ;yes, wake up
I21_30:
                MOV     CS:SLEEP_FLAG,1         ;I'm getting sleeeeepy
                JMP     I21_10
I21_40:
                POP     BX                      ;restore registers
                POP     AX
                STI                             ;enable interrupts
                CALL    DWORD PTR CS:OLDINT_21  ;execute original DOS int vec
                CLI                             ;disable interrupts
                CALL    CLEAR_VIDEO             ;clear the screen
                MOV     CS:SLEEP_FLAG,0         ;WAKE UP!
                PUSH    DX
                PUSH    AX
                PUSH    DS                      ;Save data segment
                PUSH    CS                      ;Set data segment to code seg.
                POP     DS
                CALL    INIT_SERIAL             ;Re-initialize serial port
                CALL    CLEAR_OUTBUFF           ;Clear buffers because ZCOPY
                CALL    CLEAR_INBUFF            ; sent hex 02's to sync systems
                POP     DS                      ;Restore the data segment
                POP     AX                      ;Restore registers
                POP     DX
                STI                             ;Enable interrupts
                RET     2                       ;Return to calling program
INT21           ENDP

;---------------------------------------------------------------------
;  Clear the video copy to force a new screen to be tranferred to
;  remote system
;---------------------------------------------------------------------
CLEAR_VIDEO:
                ; save necessary registers
                ;
                PUSH    CX
                PUSH    DI
        if manned_is_dumb_ansi
                PUSH    SI
        endif
                PUSH    ES
                PUSHF

                ; clear our shadow of video memory
                ; to match what we assume manned host has
                ;
                CLD                             ;Direction = forward
                PUSH    CS                      ;Push CS so ES can now
                POP     ES                      ; point at it .. ES->CS
                MOV     DI,WORD PTR VIDEO_COPY  ;ES:DI->Video copy
                MOV     AX,0720H                ;AL=SPACE,AH=white on black
                MOV     CX,CHARS_PER_SCR        ;Clear entire video copy
                REP     STOSW                   ;Store char and attribute

                ; force update of cursor pos to manned host
                ;
                MOV     CURSOR_POSITION,UNKNOWN_POS

        if manned_is_dumb_ansi
                ; clear screen on manned terminal
                ;
                MOV     AH,1                    ;use output buffer
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'['
                CALL    PUT_BUFF_DATA
                MOV     AL,'2'
                CALL    PUT_BUFF_DATA
                MOV     AL,'J'
                CALL    PUT_BUFF_DATA

                ; remember that we don't know the text attributes
                ; on the dumb ANSI manned terminal, by remembering
                ; it as something unlikely
                ;
                MOV     PREV_ATTRIBUTE,UNKNOWN_ATTR

                ; remember that we don't know the cursor position
                ; on the dumb ANSI manned terminal, by remembering
                ; it as one greater than both the row and column
                ;
                MOV     PREV_POSITION,UNKNOWN_POS

                ; disable wraparound so we can write to rightmost character
                ; on the last line of their screen, without causing scrolling
                ;
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'['
                CALL    PUT_BUFF_DATA
                MOV     AL,'?'
                CALL    PUT_BUFF_DATA
                MOV     AL,'7'
                CALL    PUT_BUFF_DATA
                MOV     AL,'l'                  ;lowercase 'L'
                CALL    PUT_BUFF_DATA

                ; map the US ASCII characters as font G0 so we can
                ; get normal lowercase letters after a SHIFTIN
                ;
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'('
                CALL    PUT_BUFF_DATA
                MOV     AL,'B'
                CALL    PUT_BUFF_DATA

                ; map the DEC linedrawing and special characters
                ; as font G1 so we can get them as BACKQUOTE and
                ; lowercase letters 'a' and 'j' through 'x'
                ; after a SHIFTOUT
                ;
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,')'
                CALL    PUT_BUFF_DATA
                MOV     AL,'0'
                CALL    PUT_BUFF_DATA

                ; remember that we don't know whether the graphic
                ; character set is currently selected on the dumb
                ; dumb ANSI manned terminal or not
                ;
                MOV     PREV_GRAPHIC,UNKNOWN_GRAPHIC

                ; if the /Ystring commandline argument was used
                ; or there is a default system name, send it to
                ; the manned host to change the xterm window title
                ; and icon name
                ;
                CMP     BYTE PTR SYSTEM_NAME,0
                JE      CV_NO_SYSTEM_NAME

                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,']'
                CALL    PUT_BUFF_DATA
                MOV     AL,'0'
                CALL    PUT_BUFF_DATA
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
                MOV     SI,OFFSET SYSTEM_NAME
                CALL    LOAD_ZSTRING
                MOV     AL,BELL
                CALL    PUT_BUFF_DATA
CV_NO_SYSTEM_NAME:
        endif

                ; restore registers
                ;
                POPF
                POP     ES
        if manned_is_dumb_ansi
                POP     SI
        endif
                POP     DI
                POP     CX
                RET

;----------------------------------------------------------------------
;  Interrupt handling routine for communications interrupt.  Provides
;  interrupt driven I/O.  Transmit or receive a character.
;----------------------------------------------------------------------
INT_COMM        PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
        if lockup_when_int_comm
                JMP     $
        endif
                PUSH    DS                      ;Save data segment
                PUSH    DX                      ;Save registers
                PUSH    BX
                PUSH    AX
        if beep_when_int_comm
                ; beep
                ;
                MOV     AX, 0E07h
                XOR     BX, BX
                INT     10h
        endif
                PUSH    CS                      ;Set data segment to code seg.
                POP     DS
                MOV     DX,COMM_PORT            ;Get port base address
                ADD     DX,2                    ;Point at interrupt ID register
                IN      AL,DX                   ;???acknowledge the interrupt
                TEST    AL,1                    ;if no interrupt, get out
                JNZ     IC_EXIT
                SUB     DX,2                    ;back to base I/O address
IC_10:
                ADD     DX,5                    ;Point to line status register
                IN      AL,DX
                SUB     DX,5                    ;Port base address
                TEST    AL,01H                  ;Received data ready?
                JNZ     IC_15                   ;yes - suck it in
                TEST    AL,20H                  ;Transmit holding register empty?
                JZ      IC_EXIT                 ;no - done with interrupt
;
;       transmit buffer empty, send a byte if you have one
;
                MOV     BX,OUT_BUFF_HEAD        ;Get the buffer head pointer
                CMP     BX,OUT_BUFF_TAIL        ;Test for data in buffer
                JE      IC_60                   ;If the same, no data so exit
                MOV     AL,[BX]                 ;Get the data
                INC     BX                      ;Point to next data in buffer
                CMP     OUT_BUFF_END,BX         ;Beyond end of buffer area?
                JNE     IC_50                   ;No, then don't reset
                MOV     BX,OUT_BUFF_BEGIN       ;Yes, reset to buffer begin
IC_50:
                MOV     OUT_BUFF_HEAD,BX        ;Save new head pointer
                OUT     DX,AL                   ;Send the data out the port
                JMP     IC_10                   ;Check for request pending
;
;       received data, get it and store in buffer
;
IC_15:
                IN      AL,DX                   ;Get data from receive register
                MOV     BX,IN_BUFF_TAIL         ;Get the buffer tail pointer
                MOV     [BX],AL                 ;Store the data in buffer
                INC     BX                      ;Point to next data storage
                CMP     IN_BUFF_END,BX          ;Beyond end of buffer area?
                JNE     IC_20                   ;No, then don't reset
                MOV     BX,IN_BUFF_BEGIN        ;Yes, reset to buffer begin
IC_20:
                CMP     BX,IN_BUFF_HEAD         ;Test for buffer full
                JE      IC_30                   ;If so,don't change ptr, sorry
                MOV     IN_BUFF_TAIL,BX         ;Save new tail pointer
                JMP     IC_10
IC_30:
                INC     RX_OVERRUNS             ;count these puppies
        if beep_when_rx_overruns
                ; beep
                ;
                PUSH    AX
                PUSH    BX
                PUSH    CX
                PUSH    DX
                MOV     AX, 0E07h
                XOR     BX, BX
                INT     10h
                POP     DX
                POP     CX
                POP     BX
                POP     AX
        endif
                JMP     IC_10
IC_60:
                MOV     COMM_INT_STA,0          ;Reset transmitting data flag
IC_EXIT:
                MOV     AL,20H                  ;Reset 8259
                OUT     20H,AL
                STI                             ;Enable interrupts
                POP     AX                      ;Restore registers
                POP     BX
                POP     DX
                POP     DS                      ;Restore data segment
                IRET
INT_COMM        ENDP

;----------------------------------------------------------------------
;  Create stack area and save all registers.
;----------------------------------------------------------------------
SET_STACK       PROC    NEAR
                ASSUME  CS:CSEG,DS:NOTHING,ES:NOTHING,SS:NOTHING
                MOV     TEMP_REG,BX             ;Save BX
                POP     BX                      ;Save the return address
                PUSH    AX                      ;Save AX
;
;       make my own stack
;
                CLI                             ;Disable interrupts
                MOV     AX,SS                   ;Put old stack segment in AX
                MOV     OLD_SS,AX               ;And save it
                MOV     AX,SP                   ;Put old stack pointer in AX
                MOV     OLD_SP,AX               ;And save it
                MOV     AX,CS                   ;Get current segment
                MOV     SS,AX                   ;And put into stack segment
                MOV     AX,STACK_TOP            ;Get top of stack address
                MOV     SP,AX                   ;And put into stack pointer
                STI                             ;Enable interrupts
;
;       save all the registers on the stack
;
                PUSH    CX
                PUSH    DX
                PUSH    SI
                PUSH    DI
                PUSH    DS
                PUSH    ES
                PUSH    BP
                MOV     AX,CS                   ;Get code segment
                MOV     DS,AX                   ;Set data segment to code seg
                MOV     ES,AX                   ;Set extra seg to code seg
                PUSH    BX                      ;Restore return address
                RET
SET_STACK       ENDP

;----------------------------------------------------------------------
;  Restore all registers and reset stack
;----------------------------------------------------------------------
RESET_STACK     PROC    NEAR
                ASSUME  CS:CSEG,DS:NOTHING,ES:NOTHING,SS:NOTHING
                POP     BX                      ;Save return address
;
;       restore the registers
;
                POP     BP                      ;Restore the registers
                POP     ES
                POP     DS
                POP     DI
                POP     SI
                POP     DX
                POP     CX
;
;       restore the original stack
;
                CLI                             ;Disable interrupts
                MOV     AX,OLD_SP               ;Get old stack pointer
                MOV     SP,AX                   ;And restore it
                MOV     AX,OLD_SS               ;Get old stack segment
                MOV     SS,AX                   ;And restore it
                STI                             ;Enable interrupts
                POP     AX                      ;Restore AX
                PUSH    BX                      ;Put return add back on stack
                MOV     BX,TEMP_REG             ;Restore BX
                RET
RESET_STACK     ENDP

;======================================================================
;  The unattended routine will execute the connect routine to establish
;  a connection with the manned system.  Once connected it will execute
;  the unattended routine to process incoming data and send any changed
;  video data to the manned system.
;======================================================================

;----------------------------------------------------------------------
;  CONNECT_UNATTENDED - Check for ring codes, answer the phone and check
;  the password.  If correct, set connected flag.
;----------------------------------------------------------------------
CONNECT_UNATTENDED PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:CSEG,SS:NOTHING
                MOV     BP,SP                   ;Save stack pointer for exit
                CMP     AUTOBAUD_FLAG,0
                JNZ     CU_AUTOBAUD_DONE
        if baud_rate_detection
                XOR     AH,AH                   ;Use input buffer
                MOV     CX,10D                  ;Number of char to check for CR
                MOV     CR_COUNT,5              ;Number of matches required
CU_10:
                CALL    GET_BUFF_DATA           ;Try to get data from buffer
                JC      CU_20                   ;If data, check for CR code
                JMP     CU_40                   ;else exit, try another speed
CU_20:
                CMP     AL,CR                   ;Check for CR code
                JNZ     CU_30                   ;No so skip over
                MOV     SPEED_COUNT,19          ;it's a CR, so speed is correct
                DEC     CR_COUNT                ;Found a match dec count
                JZ      CU_70                   ;5 out of ten then null connect
CU_30:
                LOOP    CU_10                   ;Keep trying
CU_40:
                DEC     SPEED_COUNT
                JZ      CU_50
                JMP     CU_EXIT
CU_50:
                DEC     CURRENT_SPEED           ;Try a slower speed
                CMP     CURRENT_SPEED,0FFH      ;Out of speeds?
                JNZ     CU_60                   ;No
                MOV     AL,SPEED_FLAG           ;Ran out of speeds, reset
                MOV     CURRENT_SPEED,AL        ; to starting speed
CU_60:
                MOV     SPEED_COUNT,19          ;reset tick counter
                MOV     AL,CURRENT_SPEED        ;set the slower speed
                CALL    SET_BAUD_RATE
                CALL    CLEAR_INBUFF            ;Empty the input buffer
                JMP     CU_EXIT                 ;No 5 out of 10, exit
        else
                XOR     AH,AH                   ;Use input buffer
                CALL    GET_BUFF_DATA           ;Try to get data from buffer
                JC      CU_65
                JMP     CU_EXIT                 ;if no data, we are done
CU_65:
        if manned_is_dumb_ansi
                CMP     AL,CR                   ;is it a carriage return?
        else
                CMP     AL,0                    ;is it a sync byte?
        endif
                JZ      CU_70
                JMP     CU_EXIT                 ;no - check again
        endif                                   ;baud_rate_detection
CU_70:
                MOV     AUTOBAUD_FLAG,1
CU_AUTOBAUD_DONE:
                CMP     PASSWORD_BAD,0
                JE      CU_HANGUP_NOT_DELAYING
                JMP     CU_HANGUP_DELAY
CU_HANGUP_NOT_DELAYING:
        if ask_for_password
                CMP     PASSWORD_FLAG,0
                JNZ     CU_PASSWORD_PROMPTED
                MOV     PASSWORD_FLAG,1
                MOV     PASSWORD_BUFFER,0       ;zero out the string
                MOV     SI,OFFSET PASSWORD_MSG  ;Point to enter password message
                CALL    LOAD_ZSTRING            ;Load it into output buffer

CU_PASSWORD_PROMPTED:
                ; read one character from input buffer, if possible
                ;
                XOR     AH,AH
                CALL    GET_BUFF_DATA
                JC      CU_PASSWORD_GOT_ONE
                JMP     CU_EXIT
CU_PASSWORD_GOT_ONE:

                ; get the current string length in BX
                ;
                MOV     BL,PASSWORD_FLAG
                DEC     BL
                XOR     BH,BH

                ; if this character is a carriage return or linefeed,
                ; it is time to check the whole string
                ;
                CMP     AL,CR
                JE      CU_PASSWORD_CHECKING
                CMP     AL,LF
                JE      CU_PASSWORD_CHECKING

                ; echo a star so they know we saw this character
                ;
                PUSH    AX
                MOV     AH,1
                MOV     AL,'*'
                CALL    PUT_BUFF_DATA
                POP     AX

                ; put the character into the buffer and look for more,
                ; but don't affect buffer if it is already full
                ;
                CMP     BL,PASSWORD_MAX
                JAE     CU_PASSWORD_PROMPTED
                INC     PASSWORD_FLAG
                MOV     PASSWORD_BUFFER[BX],AL
                MOV     BYTE PTR PASSWORD_BUFFER[BX+1],0
                JMP     CU_PASSWORD_PROMPTED

CU_PASSWORD_CHECKING:
                CALL    CHECK_PASSWORD          ;Check the password sent
                MOV     PASSWORD_FLAG,0         ;forget what they entered
                JZ      CU_PASSWORD_OK          ;Jump if correct
                INC     PASSWORD_TRIES
                CMP     PASSWORD_TRIES,3
                JB      CU_AUTOBAUD_DONE
                MOV     SI,OFFSET PASSWORD_MSG2 ;Point to failed password message
                CALL    LOAD_ZSTRING            ;Load it into output buffer
        else                                    ;ask_for_password
                JMP     CU_PASSWORD_OK
        endif

CU_HANGING_UP:
                ; this is another entrypoint, called from IP_90
                ;
                MOV     PASSWORD_BAD,1D         ;come back to PASSWORD_DELAY each tick
                JMP     CU_EXIT                 ;Done for this tick

CU_HANGUP_DELAY:
                ; if we haven't deasserted DTR and RTS yet...
                ;
                CMP     PASSWORD_BAD,1
                JNE     CU_OUTBUF_EMPTY

                ; if output buffer is empty...
                ;
                CLI
                MOV     AX,OUT_BUFF_HEAD
                CMP     AX,OUT_BUFF_TAIL
                STI
                JNE     CU_EXIT

                ; count 2/19 second's worth of timer ticks
                ;
                MOV     PASSWORD_TRIES,2d
                MOV     PASSWORD_BAD,2
                JMP     CU_EXIT

CU_OUTBUF_EMPTY:
                ; if the desired number of ticks have occurred...
                ;
                DEC     PASSWORD_TRIES
                JNZ     CU_EXIT

                ; if we haven't deasserted DTR and RTS yet...
                ;
                CMP     PASSWORD_BAD,2
                JNE     CU_ONHOOK_ALREADY

                ; deassert DTR and RTS
                ;
                MOV     AH,0
                MOV     DX,COMM_PORT
                CALL    SET_CARRIER

                ; count one more second
                ;
                MOV     PASSWORD_TRIES,19
                MOV     PASSWORD_BAD,3
                JMP     CU_EXIT

CU_ONHOOK_ALREADY:
                MOV     AH,1                    ;assert DTR and RTS
                MOV     DX,COMM_PORT
                CALL    SET_CARRIER
                MOV     AL,1                    ;Use setup string 2
                CALL    RESET_MODEM             ;Hangup and reset modem
                MOV     PASSWORD_BAD,0          ;restart normal password processing
                MOV     AUTOBAUD_FLAG,0         ;restart autobaud processing
                JMP     CU_EXIT                 ;Done, so exit

CU_PASSWORD_OK:
                MOV     CONNECT_FLAG,1
                CALL    CLEAR_VIDEO
        if manned_is_dumb_ansi
        else
                MOV     AH,1                    ;Use output buffer

                XOR     AL,AL                   ;Sync byte to send
                MOV     CX,5                    ;Send 5 of them
CU_100:
                CALL    PUT_BUFF_DATA           ;Send them
                LOOP    CU_100
;
; Notify remote computer the value of host's COMM_FLAG switch.  This allows
; the remote computer to properly send the ZCOPY command line to the host.
;
                MOV     AH,1                    ;Use output buffer
                MOV     AL,COMM_FLAG            ;Send host's comm port
                INC     AL                      ;Bump AL so sent byte will be
                                                ; in range from 1 thru 4.
                MOV     CX,10D                  ;Send 10 of them
CU_110:
                CALL    PUT_BUFF_DATA           ;Send Host Comm Port
                LOOP    CU_110                  ;Sent 10 ?
                XOR     AL,AL                   ;Sync byte to send
                MOV     CX,5D                   ;Send 5 of them
CU_120:
                CALL    PUT_BUFF_DATA           ;Send sync byte
                LOOP    CU_120                  ;Sent 5 ?
        endif                                   ;manned_is_dumb_ansi
CU_EXIT:
                RET
CONNECT_UNATTENDED ENDP

        if ask_for_password
;----------------------------------------------------------------------
;  Check the password in the buffer with correct password
;  Input - Length of user-enterewd password in BX
;  Output - Zero set - correct password
;           Zero cleared - Wrong password
;  Changes - SI, DI
;----------------------------------------------------------------------
CHECK_PASSWORD  PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:CSEG,SS:NOTHING
                PUSH    CX                        ;Save register
                MOV     CX,BX
                INC     CX                      ;check the terminating NUL too
                MOV     SI,OFFSET PASSWORD_BUFFER ;Pointer to buffer
                MOV     DI,OFFSET PASSWORD        ;Pointer to password
                CLD                             ;Compare forward
                REPZ    CMPSB                   ;Repeat while passwords match
CHECK_PASSWORD_DONE:
                POP     CX                      ;Restore register
                RET
CHECK_PASSWORD  ENDP
        endif

;----------------------------------------------------------------------
;  UNATTENDED - Process incoming data and check for changes in video
;  data, format and send to manned system.  Maintains 18 or less characters
;  in output buffer, ensures maximum throughput.
;----------------------------------------------------------------------
UNATTENDED      PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:CSEG,SS:NOTHING
                MOV     BP,SP                   ;Save stack pointer for exit
                CALL    INPUT_PROCESSING        ;Process any buffered data
;
;       Check for a change in the cursor position
;
                PUSH    ES                      ;Save ES
                MOV     AX,40H                  ;Set ES to BIOS data segment
                MOV     ES,AX
                MOV     BX,ES:[50H]             ;Get current cursor position
                MOV     CX,ES:[60H]             ;Get current cursor shape
                POP     ES                      ;Restore ES
        if manned_is_dumb_ansi
                MOV     CURSOR_POSITION,BX      ;Save new cursor position

                ; don't move the cursor on the dumb terminal yet,
                ; because it has to move if we send any text to it,
                ; and it might just end up in the right spot anyways
                ;

                ; no way to change the cursor's shape on ANSI dumb terminal
                ; (I think)
                ;

        else
                CMP     BX,CURSOR_POSITION      ;Compare with copy
                JZ      UA_5                    ;No change, skip
                MOV     CURSOR_POSITION,BX      ;Save new cursor position
                JMP     UA_7
UA_5:
                CMP     CX,CURSOR_SHAPE         ;Compare with copy
                JZ      UA_10                   ;No change, skip
UA_7:
                MOV     CURSOR_SHAPE,CX         ;Save new cursor shape
                MOV     AL,0D2H                 ;Sync bits, set position
                MOV     AH,1                    ;Use output buffer
                CALL    PUT_BUFF_DATA           ;Send sync byte
                MOV     AL,BL                   ;Low byte of position
                CALL    PUT_BUFF_DATA
                MOV     AL,BH                   ;High byte of position
                CALL    PUT_BUFF_DATA
                MOV     AL,CL                   ;Low byte of shape
                CALL    PUT_BUFF_DATA
                MOV     AL,CH                   ;High byte of shape
                CALL    PUT_BUFF_DATA
        endif
UA_10:
                MOV     VID_RAM_OFFSET,0        ;Start at address 0
                MOV     BLOCK_COUNT,0           ;Reset block counter
        if detect_scrolling
                MOV     BLOCK_ROW,0
                MOV     BLOCK_COLUMN,0
                MOV     SCROLL_FOUND,0
                MOV     SCROLL_OFFSET,0
                MOV     SCROLL_OFFSET_ROW,0
        endif
;
; Check the output buffer and keep it full enough so it's busy till
; next timer tick.
;
UA_30:
        if debug_unattended
        else
                MOV     AX,7                    ;1200 baud, at least 7 bytes
                MOV     CL,CURRENT_SPEED        ;the speed decides the size
                SHL     AX,CL                   ;larger for faster baud
                MOV     CX,AX                   ;at least this many bytes
                CLI
                MOV     AX,OUT_BUFF_HEAD        ;Get the head pointer
                MOV     BX,OUT_BUFF_TAIL        ;Get the tail pointer
                STI
                SUB     BX,AX                   ; BX = tail - head
                JAE     UA_40                   ; if tail>=head, we have fullness
                ADD     BX,OUT_BUFF_END         ; need to add (end - begin)
                SUB     BX,OUT_BUFF_BEGIN
UA_40:
                CMP     BX,CX                   ; fullness >= wanted?
                JB      UA_60
                JMP     UA_EXIT                 ; yes - no need to put more in
        endif
;
;       setup to compare the video copy with the video ram
;
UA_60:
                PUSH    DS                      ;Save data segment
                MOV     SI,VID_RAM_OFFSET       ;Get the current offset
                MOV     DI,TEMP_VIDEO_PTR       ;Dest. ES:DI temp video buffer
                MOV     AX,VIDEO_SEGMENT        ;Get video Segment register
                MOV     DS,AX                   ;Source DS:SI video RAM
;
;       transfer block_size words of data from video RAM to the temp buffer
;
                MOV     CX,BLOCK_SIZE           ;Get count of words to transfer
                CLD
                TEST    CS:DESNOW_FLAG,1        ;Check desnow flag
                JZ      UA_90
                SAL     CX,1                    ;Convert words to bytes
                MOV     DX,3DAH                 ;CGA status port
UA_70:
                IN      AL,DX                   ;Get status byte
                TEST    AL,1                    ;Test display enable
                JNZ     UA_70                   ;If in Hor. sync then wait
                CLI                             ;Disable interrupts
UA_80:
                IN      AL,DX                   ;Get status byte
                TEST    AL,1                    ;Test display enable
                JZ      UA_80                   ;Wait for Hor. sync
                MOVSB                           ;Transfer one byte
                STI                             ;Enable interrupts
                LOOP    UA_70                   ;Transfer block size words
                JMP     UA_100
UA_90:
                REP     MOVSW
UA_100:
                POP     DS                      ;Restore data segment
;
;       compare the block from video RAM with the video copy
;
UA_COMPARE:
                MOV     CX,BLOCK_SIZEX2         ;Number of bytes to compare
                MOV     SI,TEMP_VIDEO_PTR       ;Point to block of video data
                MOV     DI,VIDEO_COPY           ;Point to video copy
                ADD     DI,VID_RAM_OFFSET       ;Adjust for current block
        if detect_scrolling
                ADD     DI,SCROLL_OFFSET        ;Adjust for scrolling
        endif
                REPE    CMPSB                   ;Compare while equal
        if detect_scrolling
                JNE     UA_DIFFERENT
                JMP     UA_SAME                 ;Whole block matches
UA_DIFFERENT:
                CMP     SCROLL_FOUND,0          ;Did we find a scroll?
                JZ      UA_DIFFERENT_NO_SCROLL_YET
        if 1
        if manned_is_dumb_ansi
                ; if less than one whole line was the same...
                ;
                MOV     BL,SCROLL_DEST_ROW
                CMP     BLOCK_ROW,BL
                JA      UA_SCROLL_OK

                ; copy the shadow to the temp area
                ; for the first block on the line
                ;
                MOV     SI,VIDEO_COPY
                ADD     SI,SCROLL_DEST
                ADD     SI,SCROLL_OFFSET
                MOV     DI,TEMP_VIDEO_PTR
                MOV     CX,BLOCK_SIZE
                REP     MOVSW

                ; go back to the beginning of the line, and try
                ; a larger scroll, as if it miscompared there
                ;
                MOV     SCROLL_FOUND,0
                MOV     BX,SCROLL_DEST
                MOV     VID_RAM_OFFSET,BX
                MOV     BL,SCROLL_DEST_BLOCK
                MOV     BLOCK_COUNT,BL
                MOV     BL,SCROLL_DEST_ROW
                MOV     BLOCK_ROW,BL
                MOV     BLOCK_COLUMN,0
                JMP     UA_CHANGE_SCROLL

UA_SCROLL_OK:
        endif

                ; flush scroll info to manned host, and restart comparing
                ; the video data against the same or blank shadow
                ;
                CALL    SEND_SCROLL
                JMP     UA_COMPARE
        else                                    ;not 1
                ; joel debug nasty
                sti
UA_COMPARE_LOOP:
                jmp     UA_COMPARE_LOOP
        endif                                   ;1
UA_DIFFERENT_NO_SCROLL_YET:
                CMP     SCROLL_OFFSET,0         ;Did we start looking yet?
                JNZ     UA_CHANGE_SCROLL        ; yes - skip the anchoring
                CMP     BLOCK_COLUMN,0          ;Is this beginning of line?
                JE      UA_CAN_SCROLL
                JMP     UA_120                  ; no - scroll can't start here
UA_CAN_SCROLL:
                MOV     AL,CL                   ; remember attribute vs. char mismatch
                MOV     BX,VID_RAM_OFFSET       ;Remember destination of putative copy
                MOV     SCROLL_DEST,BX
                MOV     BL,BLOCK_ROW
                MOV     SCROLL_DEST_ROW,BL
                MOV     BL,BLOCK_COUNT
                MOV     SCROLL_DEST_BLOCK,BL
UA_CHANGE_SCROLL:
                ADD     SCROLL_OFFSET,BYTES_PER_LINE    ;Maybe one more line scrolled
                INC     SCROLL_OFFSET_ROW
                MOV     BX,SCROLL_OFFSET
                ADD     BX,VID_RAM_OFFSET
                CMP     BX,BYTES_PER_SCR        ;Did we hit end of shadow?
                JAE     UA_HIT_SHADOW_END
                JMP     UA_COMPARE              ;no - try comparing again
UA_HIT_SHADOW_END:
                MOV     CL,AL                   ;Restore attr vs. char miscompare
                JMP     UA_120                  ;Format & send block
UA_SAME:
                CMP     SCROLL_OFFSET,0         ;did we assume scrolling when we compared?
                JZ      UA_112
                MOV     SCROLL_FOUND,1          ;yes - remember to flush on next miscompare
UA_112:
        else    ; NOT detect_scrolling
                JNE     UA_120                  ;different?  format/send block
        endif
                CALL    UA_CHANGE_VID_PTR       ;next block, please
                JZ      UA_115                  ;end of screen?
                JMP     UA_60                   ;no - copy another block from video RAM
UA_115:
        if detect_scrolling
                CMP     SCROLL_FOUND,0          ;Were we assuming scrolling?
                JZ      UA_EXIT                 ;no - nothing to send to manned
                CALL    SEND_SCROLL
;
; now we need to give the manned PC the lines
; that fill the hole we just opened up
; (unless they really are blank)
;
; note: we cannot simply jump to UA_120,
; because the data at TEMP_VIDEO_PTR
; is for the line above the hole
;
                JMP     UA_30                   ;cache and compare again
        else    ; NOT detect_scrolling
                JMP     UA_EXIT                 ;then exit
        endif

;
;       data doesn't match, format and send to manned system
;
UA_120:
                INC     CX                      ;Adjust count
                AND     CX,1                    ;LSB indicates char or attr.
                CALL    TRANSFER_BLOCK          ;Prepare to send block of data
                CALL    UA_CHANGE_VID_PTR       ;next block, please
                JZ      UA_EXIT                 ;end of screen?
                JMP     UA_30                   ;no - check buffer fullness

UA_EXIT:
        if manned_is_dumb_ansi
                ; move the cursor on the manned PC to match ours,
                ; if necessary
                ;
                MOV     BX,CURSOR_POSITION
                CALL    SEND_CURSOR_ESCAPE
        endif

                TEST    SMILE_FLAG,1            ;Smile face enabled ?
                JE      UA_130                  ;No, skip smile face coding
;
; Redisplay the smile face in upper right corner
;
; Note: this couldn't have been done at the top of
; UNATTENDED because that would have caused a miscompare
; of 5th block on first line when we look for scrolling
;
                PUSH    ES                      ;Store ES
                MOV     AX,VIDEO_SEGMENT        ;AX = Video Segment
                MOV     ES,AX                   ;ES = Video Segment
                MOV     DI,158D                 ;DI-> Upper right corner
                MOV     AL,SMILE                ;Smiley Face code
                CALL    PUT_VIDEO_DATA          ;Write Video RAM
                POP     ES                      ;Restore ES
UA_130:
                RET


; move on to the next block
;
; affects AX, globals
;
; returns carry and zero flags by comparing current video
; location (plus scroll offset) against the end of screen
;
UA_CHANGE_VID_PTR:
                MOV     AX,VID_RAM_OFFSET       ;Get current block pointer
                ADD     AX,BLOCK_SIZEX2         ;Point to next block
                MOV     VID_RAM_OFFSET,AX       ;And save the pointer
                INC     BLOCK_COUNT             ;Increment current block count
        if detect_scrolling
                ADD     BLOCK_COLUMN,BLOCK_SIZE
                CMP     BLOCK_COLUMN,CHARS_PER_LINE
                JB      UA_BLOCK_COLUMN_OK
                SUB     BLOCK_COLUMN,CHARS_PER_LINE
                INC     BLOCK_ROW
UA_BLOCK_COLUMN_OK:
                ADD     AX,SCROLL_OFFSET        ;Adjust for scrolling
        endif
                CMP     AX,BYTES_PER_SCR        ;Check for end of video RAM
                RET

UNATTENDED      ENDP




        if manned_is_dumb_ansi
;----------------------------------------------------------------------
; send the ANSI escape sequence to change the foreground
; and background colors, blinking, and bold to the PC-style
; attribute byte (as it came from video RAM) in DH
;
; if zero flag is cleared, foreground and background colors
; will be swapped
;
; destroys BX, DH
;----------------------------------------------------------------------
SEND_ATTRIBUTE_ESCAPE  PROC    NEAR

                PUSH    AX

                ; save the zero flag so we can use it later
                ;
                PUSHF

                ; separate the FG and BG colors
                ; from each other and attributes
                ;
                MOV     AH,DH
                AND     AH,7                    ;FG in AH
                MOV     AL,DH
                SHR     AL,4
                AND     AL,7                    ;BG in AL

        if force_fg_ne_bg
                ; if FG color is same as BG (ignoring blink and bold),
                ; force BG to some other color,
                ; so it will contrast with FG, since we don't think blink
                ; is implemented at all on the dumb ANSI terminal, and we
                ; don't think bold is implemented as a brighter color
                ;
                CMP     AH,AL
                JNE     TB_FG_NE_BG
                XOR     AL,7
TB_FG_NE_BG:
        endif

                ; restore the zero flag
                ;
                POPF

                ; swap FG and BG colors if desired
                ;
                JZ      TB_NOT_REVERSE_VIDEO
                XCHG    AL,AH
TB_NOT_REVERSE_VIDEO:

                ; now blink and bold are in DH,
                ; but desired colors are in AX...
                ;

                ; now squish the colors back together
                ; with the blink and bold attributes
                ; so they can all be remembered in byte
                ;
                AND     DH,88h
                OR      DH,AH
                SHL     AL,4
                OR      DH,AL

                ; if this character has different attributes
                ; from those we last output to the terminal...
                ;
                CMP     DH,PREV_ATTRIBUTE
                JZ      SEND_ATTRIBUTE_DONE
                MOV     PREV_ATTRIBUTE,DH

                ; send the ANSI escape to change color/blink/bold
                ;
                MOV     AH,1
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'['
                CALL    PUT_BUFF_DATA
                MOV     AL,'0'                  ;turn off blink and bold
                CALL    PUT_BUFF_DATA
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
                TEST    DH,128                  ;MSB of BG nibble means blink
                JZ      SEND_ATTRIBUTE_NO_BLINK
                MOV     AL,'5'
                CALL    PUT_BUFF_DATA
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
SEND_ATTRIBUTE_NO_BLINK:
                TEST    DH,8                    ;MSB of FG nibble means bold
                JZ      SEND_ATTRIBUTE_NO_BOLD
                MOV     AL,'1'
                CALL    PUT_BUFF_DATA
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
SEND_ATTRIBUTE_NO_BOLD:
                ; setup to translate PC BG/FG colors to ANSI colors
                ;
                MOV     BX,OFFSET COLOR_XLATE

                MOV     AL,DH                   ;FG nibble is least significant
                AND     AL,7
                XLATB                           ;convert PC color to ANSI
                ADD     AL,30                   ;ANSI codes are 30 to 37
                CALL    SEND_DECIMAL_ASCII
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
                MOV     AL,DH                   ;BG nibble is most significant
                SHR     AL,4
                AND     AL,7
                XLATB                           ;convert PC color to ANSI
                ADD     AL,40                   ;ANSI codes are 40 to 47
                CALL    SEND_DECIMAL_ASCII
                MOV     AL,'m'                  ;must terminate the escape sequence!
                CALL    PUT_BUFF_DATA

SEND_ATTRIBUTE_DONE:
                POP     AX
                RET

SEND_ATTRIBUTE_ESCAPE ENDP
        endif



        if manned_is_dumb_ansi
;----------------------------------------------------------------------
; send the ANSI escape sequence to move the manned PC's cursor
; to the zero-based row passed in BH
; and the zero-based column passed in BL
;
; destroys AX
;----------------------------------------------------------------------
SEND_CURSOR_ESCAPE  PROC    NEAR

                ; if cursor row is beyond bottom of screen (hidden),
                ; don't bother to move the cursor
                ;
                ; also don't remember cursor's position, in case the
                ; last-known valid position will be used for next text
                ;
                CMP     BH,LINES_PER_SCR
                JAE     SEND_CURSOR_DONE

                ; if the requested position is same as where we
                ; think the dumb terminal's cursor already is,
                ; do nothing
                ;
                MOV     AX,PREV_POSITION
                CMP     BX,AX
                JE      SEND_CURSOR_DONE
                MOV     PREV_POSITION,BX

                ; if requested position is beginning of line
                ; after the one the dumb terminal is already on,
                ; just output a carriage-return and a line-feed
                ;
                INC     AH
                XOR     AL,AL
                CMP     AX,BX
                JNE     SEND_CURSOR_ROW_COL
                MOV     AH,1
                MOV     AL,CR
                CALL    PUT_BUFF_DATA
                MOV     AL,LF
                CALL    PUT_BUFF_DATA
                JMP     SEND_CURSOR_DONE

SEND_CURSOR_ROW_COL:
                ; send the escape character and a left bracket
                ;
                MOV     AH,1                        ; use output buffer
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'['
                CALL    PUT_BUFF_DATA

                ; if cursor needs to go to top left corner of screen,
                ; no need to send either row or column
                ;
                CMP     BX,0
                JE      SEND_CURSOR_TERMINATOR

                ; send the desired row
                ;
                MOV     AL,BH
                INC     AL
                CALL    SEND_DECIMAL_ASCII

                ; if requested position is beginning of line,
                ; no need to send the column, since the dumb
                ; terminal will default to one for that parm
                ;
                CMP     BL,0
                JE      SEND_CURSOR_TERMINATOR

                ; send a semicolon and the desired column
                ;
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
                MOV     AL,BL
                INC     AL
                CALL    SEND_DECIMAL_ASCII

SEND_CURSOR_TERMINATOR:
                MOV     AL,'H'
                CALL    PUT_BUFF_DATA

SEND_CURSOR_DONE:
                RET

SEND_CURSOR_ESCAPE  ENDP
        endif               ; manned_is_dumb_ansi

        if manned_is_dumb_ansi
;----------------------------------------------------------------------
; sends a sequence of ASCII characters that are the decimal
; representation of the binary number passed in AL
;
; for example, it outputs 0x31 0x32 0x33 which is "123" in
; ASCII when you pass 0x7B (123 decimal) in AL
;
; destroys AL
; leaves AH=1
;----------------------------------------------------------------------
SEND_DECIMAL_ASCII      PROC    NEAR

                ; save previous contents of BX register
                ;
                PUSH    BX

                ; save argument in BL register
                ;
                MOV     BL,AL

                ; send the decimal hundreds digit (if any)
                ;
                ; this has only 3 possible values, so we handle
                ; it with 2 comparisons
                ;
                ; also place a nonzero value in BH if we sent this
                ;

                XOR     AH,AH
                CMP     BL,200
                JB      SEND_DECIMAL_LT_200
                MOV     AH,1
                MOV     AL,'2'
                CALL    PUT_BUFF_DATA
                SUB     BL,200
                JMP     SEND_DECIMAL_LT_100
SEND_DECIMAL_LT_200:
                CMP     BL,100
                JB      SEND_DECIMAL_LT_100
                MOV     AH,1
                MOV     AL,'1'
                CALL    PUT_BUFF_DATA
                SUB     BL,100
SEND_DECIMAL_LT_100:
                MOV     BH,AH

                ; determine the tens digit but don't send it yet
                ;
                ; this and the ones digit have 100 possibilities
                ; together, so we do the 17-cycle division, although
                ; we could instead have done a lookup table
                ;
                XOR     AH,AH
                MOV     AL,BL
                MOV     BL,10
                DIV     BL
                               ; quotient in AL, remainder in AH
                MOV     BL,AH

                ; if tens digit needs to be sent...
                ; (tens is nonzero or hundreds were sent)
                ;
                OR      AL,AL
                JNZ     SEND_DECIMAL_NEED_TENS
                OR      BH,BH
                JZ      SEND_DECIMAL_WITHOUT_TENS
SEND_DECIMAL_NEED_TENS:
                ADD     AL,'0'
                MOV     AH,1
                CALL    PUT_BUFF_DATA
SEND_DECIMAL_WITHOUT_TENS:

                ; this ones digit will always be sent
                ;
                MOV     AL,BL
                ADD     AL,'0'
                MOV     AH,1
                CALL    PUT_BUFF_DATA

                ; restore the contents of BX register
                ;
                POP     BX

                RET

SEND_DECIMAL_ASCII      ENDP
        endif               ; manned_is_dumb_ansi

        if detect_scrolling
;----------------------------------------------------------------------
; send the scroll dest, source, and length to manned system,
; and scroll our copy of video memory
; and optionally fill the hole with blanks
;
; when using an ANSI dumb terminal as manned,
; if scroll_clears_lines is off, length will basically be forced
; to include all lines on screen below the source
;
; input: nothing
; output: SCROLL_OFFSET=0, SCROLL_OFFSET_ROW=0, SCROLL_FOUND=0
; changes: AX, BX, CX, DI, SI
;----------------------------------------------------------------------
SEND_SCROLL     PROC    NEAR

        if manned_is_dumb_ansi
                ; define the scroll region to start at SCROLL_DEST_ROW
                ; and continue through BLOCK_ROW+SCROLL_OFFSET_ROW
                ;
                MOV     AH,1
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'['
                CALL    PUT_BUFF_DATA
                MOV     AL,SCROLL_DEST_ROW
                INC     AL                      ;convert to 1-based
                CALL    SEND_DECIMAL_ASCII
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
                MOV     AL,BLOCK_ROW            ;0-based first diff row (a.k.a. 1-based last same row)
                ADD     AL,SCROLL_OFFSET_ROW
                MOV     BH,AL                   ;save bottom row to move cursor later
                CALL    SEND_DECIMAL_ASCII
                MOV     AL,'r'
                CALL    PUT_BUFF_DATA

                ; remember that the cursor just moved to the
                ; top left corner of the screen (this may be
                ; a bug in XTERM, but we have to live with it)
                ;
                MOV     PREV_POSITION, 0

                ; move the cursor to the bottom of that region
                ;
                ; column doesn't matter, so we choose the first,
                ; so it can be sent slightly more efficiently
                ;
                DEC     BH                      ;convert to 0-based
                XOR     BL,BL
                CALL    SEND_CURSOR_ESCAPE

                ; now send SCROLL_OFFSET_ROW linefeeds
                ;
                MOV     AH,1
                MOV     AL,LF
                MOV     CL,SCROLL_OFFSET_ROW
SEND_SCROLL_LF_LOOP:
                CALL    PUT_BUFF_DATA
                DEC     CL
                JNZ     SEND_SCROLL_LF_LOOP

                ; reset the scroll region to be the whole screen
                ;
                MOV     AH,1
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,'['
                CALL    PUT_BUFF_DATA
                MOV     AL,'r'
                CALL    PUT_BUFF_DATA

                ; sorry, cursor moved again
                ;
                MOV     PREV_POSITION, 0

                ; calculate length from source to end of matching lines
                ; but ensure it is a POSITIVE multiple of lines, since we
                ; told the dumb terminal to delete lines, not chars
                ;
                MOV     CL,BYTES_PER_LINE
                MOV     AL,BLOCK_ROW
                SUB     AL,SCROLL_DEST_ROW
                XOR     AH,AH
                MUL     CL
                MOV     CX,AX
        else                                    ;not manned_is_dumb_ansi
                ; send the scroll destination, offset and length,
                ; as a packet, to manned system
                ;
                MOV     AH,1                    ;Use output buffer
                MOV     AL,0E1H                 ;Sync bits, scroll screen w/dest
                CALL    PUT_BUFF_DATA           ;Send sync byte
                MOV     BX,SCROLL_DEST
                MOV     AL,BL                   ;Low  byte of destination
                CALL    PUT_BUFF_DATA
                MOV     AL,BH                   ;High byte of destination
                CALL    PUT_BUFF_DATA
                ADD     BX,SCROLL_OFFSET
                MOV     AL,BL                   ;Low  byte of source
                CALL    PUT_BUFF_DATA
                MOV     AL,BH                   ;High byte of source
                CALL    PUT_BUFF_DATA
                MOV     CX,VID_RAM_OFFSET       ;calculate length in bytes
                SUB     CX,SCROLL_DEST
                MOV     AL,CL                   ;Low  byte of length
                CALL    PUT_BUFF_DATA
                MOV     AL,CH                   ;High byte of length
                CALL    PUT_BUFF_DATA
        endif                                   ;manned_is_dumb_ansi

                ; scroll our copy of video memory
                ;
                ; note: CX already has byte length from above
                ;
                MOV     DI,VIDEO_COPY           ;dest is start of video copy
                ADD     DI,SCROLL_DEST
                MOV     SI,DI                   ;src is later in video copy
                ADD     SI,SCROLL_OFFSET
                SHR     CX,1                    ;Convert bytes to chars
                CLD                             ;pointers should increment
                REP     MOVSW                   ;copy lines

        if scroll_clears_lines OR manned_is_dumb_ansi
                ; fill the hole we just opened up with blanks
                ;
                ; note: the above scrolling left DI pointing
                ; at the hole, so there is no need to reload it
                ;
                MOV     AX,0720H                ;AL=SPACE,AH=white text, black BG
                MOV     CX,SCROLL_OFFSET        ;as many bytes as were opened up
                SHR     CX,1                    ;Convert bytes to chars (words)
                REP     STOSW                   ;Store char and attribute
        endif

                ; video offset = copy offset
                ;
                MOV     SCROLL_OFFSET,0
                MOV     SCROLL_OFFSET_ROW,0

                ; scroll region is no longer stored
                ;
                MOV     SCROLL_FOUND,0

                RET
SEND_SCROLL     ENDP
        endif

;----------------------------------------------------------------------
;  Get data from the unattended input buffer and process
;  Input - Nothing
;  Output - Keyboard buffer or shift status is updated
;  Changes - Nothing
;----------------------------------------------------------------------
INPUT_PROCESSING PROC    NEAR
                PUSH    AX                      ;Save register
        if manned_is_dumb_ansi
                PUSH    BX
        endif
                PUSH    DX
                PUSH    ES
                MOV     AX,40H                  ;Set ES to BIOS data area
                MOV     ES,AX
IP_10:
                CMP     CONNECT_FLAG,0          ;Are we connected
                JZ      IP_20                   ;No, don't check carrier
                CALL    CHECK_CARRIER           ;Yes check for carrier loss
                JNZ     IP_20
                JMP     IP_90                   ;Carrier loss, reset and exit
IP_20:
                MOV     AH,0                    ;Use input buffer
                CALL    GET_BUFF_DATA           ;Get a byte of data
                JC      IP_60                   ;If data then process
                JMP     IP_EXIT                 ;Otherwise exit
;
;       Check to see if expecting ASCII data
;
IP_60:
                CMP     ASCII_FLAG,0            ;Check ASCII flag
                JNZ     IP_70                   ;
                JMP     IP_120                  ; 0=not expecting data here
IP_70:
                CMP     ASCII_FLAG,1            ;Check for received 1st byte
                JNZ     IP_80
        if manned_is_dumb_ansi
                ; if escape key was doubled, put one in keyboard buffer
                ; and return to ground state
                ;
                CMP     AL,ESC_KEY
                JNZ     IP_70A
                XOR     AH,AH
                JMP     IP_110
IP_70A:
        endif
                MOV     KEY_ONE,AL              ;Save scan code, byte one

        if manned_is_dumb_ansi
                ; if not left bracket and not BELL, put both escape
                ; and this character into KB buffer and goto groundstate
                ;
                CMP     AL,BELL
                JZ      IP_70B
                CMP     AL,'['
                JZ      IP_70B
                XOR     AH,AH
                MOV     AL,ESC_KEY
                CALL    PUT_KEY_DATA
                MOV     AL,KEY_ONE
                JMP     IP_110
IP_70B:
        endif
                INC     ASCII_FLAG              ;Indicate receiving one byte
                JMP     IP_10                   ;Process next byte
IP_80:
                MOV     KEY_TWO,AL              ;Save ASCII code, byte two
        if manned_is_dumb_ansi
                ; if first key wasn't left square bracket,
                ; skip function key and arrow processing
                ;
                CMP     KEY_ONE,'['
                JE      IP_GOT_LEFT_SQ_BRACKET
                JMP     IP_NOT_ANSI
IP_GOT_LEFT_SQ_BRACKET:

                ; if we just got an ASCII digit...
                ;
                ; assumes '0' has lowest numeric value and
                ; '9' has highest
                ;
                CMP     AL,'0'
                JB      IP_NOT_DIGIT
                CMP     AL,'9'
                JA      IP_NOT_DIGIT

                ; if we already have 3 digits, don't store this one
                ; and force the count to saturate at 4 (ASCII_FLAG==6)
                ;
                CMP     ASCII_FLAG,5
                JB      IP_STORABLE_DIGIT
                MOV     ASCII_FLAG,6
                JMP     IP_10

IP_STORABLE_DIGIT:
                ; store it and wait for more digits or a terminator
                ;
                INC     ASCII_FLAG
                MOV     BL,ASCII_FLAG
                SUB     BL,3
                XOR     BH,BH
                MOV     KEY_DIGITS[ BX ],AL
                JMP     IP_10

IP_NOT_DIGIT:
                ; if this is a semicolon, wait for a valid terminator
                ; and discard the whole sequence
                ;
                ; otherwise, this must be a terminator now
                ;
                CMP     AL,';'
                JNE     IP_NOT_SEMICOLON
                MOV     ASCII_FLAG,6
                JMP     IP_10

IP_NOT_SEMICOLON:
                ; if terminator is between 'A' and 'D'...
                ;
                CMP     AL,'A'
                JB      IP_NOT_ARROW
                CMP     AL,'D'
                JA      IP_NOT_ARROW

                ; translate into a PC keycode and stuff into KB buffer
                ;
                ; allows SHIFT/CTRL/ALT to be applied once
                ;
                SUB     AL,'A'
                XOR     AH,AH
                SHL     AX,1                        ; each entry is 2 bytes big
                MOV     BX,AX
                XOR     AH,AH
                MOV     AL,SHIFT_FLAG               ; multiply this by size of
                MOV     SHIFT_FLAG,0
                SHL     AX,3                        ;    row (8 bytes)
                ADD     BX,AX
                MOV     AX,WORD PTR ARROW_XLATE[ BX ]
                JMP     IP_110

IP_NOT_ARROW:
                ; if terminator is tilde...
                ;
                CMP     AL,'~'
                JE      IP_IS_FUNC_OR_MOTION
                JMP     IP_NOT_FUNC_OR_MOTION
IP_IS_FUNC_OR_MOTION:

                ; make sure there are 3 or fewer digits
                ; (no digits is OK, means zero value)
                ;
                MOV     AL,0
                MOV     BL,ASCII_FLAG
                CMP     BL,2
                JNB     IP_NOT_TOO_FEW_DIGITS
                JMP     IP_INVALID_ANSI             ; <2 would be internal error, but check it anyways!
IP_NOT_TOO_FEW_DIGITS:
                JE      IP_CVT_DIGITS_DONE
                CMP     BL,5
                JNA     IP_NOT_TOO_MANY_DIGITS
                JMP     IP_INVALID_ANSI
IP_NOT_TOO_MANY_DIGITS:

                ; convert the ASCII digits into binary
                ;

                ; convert ASCII_FLAG to a ones-based count of digits,
                ; so it decrements to zero when no more digits, so it
                ; is possible to know when to terminate the loop
                ; without additional comparisons
                ;
                ; well that's not true; I could have jumped on carry
                ; which gets set when 0 decrements to -1, instead of
                ; jumping on zero...
                ;
                SUB     BL,2                        ; leave it one too high,
                XOR     BH,BH

                ; get ones digit, convert with just a subtraction
                ;
                MOV     AL,KEY_DIGITS[ BX - 1 ]
                SUB     AL,'0'

                ; if a tens digit exists, get it, scale it, add it
                ;
                DEC     BL
                JZ      IP_CVT_DIGITS_DONE
                MOV     AH,KEY_DIGITS[ BX - 1 ]
                SUB     AH,'0'
                SHL     AH,1                        ; multiply AL by 10
                MOV     DL,AH                       ;    as AL = AL*8 + AL*2
                SHL     AH,2
                ADD     AH,DL                       ; last step of multiply
                ADD     AL,AH                       ; combine with ones

                ; if hundreds digit exists, get it, shift it,
                ; check it for values 1 or 2; 0 or >=3 are errors
                ;
                DEC     BX
                JZ      IP_CVT_DIGITS_DONE
                MOV     AH,KEY_DIGITS[ BX - 1 ]
                CMP     AH,'1'
                JNE     IP_CVT_NOT_100
                ADD     AL,100
                JMP     IP_CVT_DIGITS_DONE
IP_CVT_NOT_100:
                CMP     AH,'2'
                JNE     IP_INVALID_ANSI
                ADD     AL,200
IP_CVT_DIGITS_DONE:

                ; range-check the binary value:
                ; function keys should be from 11 to 24
                ;
                CMP     AL,11
                JB      IP_ANSI_BELOW_FUNC
                CMP     AL,24
                JA      IP_ANSI_ABOVE_FUNC

                ; eliminate the dead codes between F5 and F6,
                ; and between F10 and F11
                ;
                CMP     AL,16
                JE      IP_INVALID_ANSI
                CMP     AL,22
                JE      IP_INVALID_ANSI

                ; now translate the function key into a PC keycode
                ; and put it into the keyboard buffer
                ;
                ; allows SHIFT/CTRL/ALT to be applied once
                ;
                SUB     AL,11                       ; now zero-based
                XOR     AH,AH
                SHL     AX,1                        ; each entry is 2 bytes
                MOV     BX,AX
                XOR     AH,AH
                MOV     AL,SHIFT_FLAG               ; multiply this by size of
                MOV     SHIFT_FLAG,0
                SHL     AX,2                        ;    row (28 bytes)
                ADD     BX,AX                       ;    as AX*4 + AX*8 + AX*16
                SHL     AX,1
                ADD     BX,AX
                SHL     AX,1
                ADD     BX,AX                       ; add/multiply done
                MOV     AX,WORD PTR FUNC_XLATE[ BX ]
                JMP     IP_110

IP_ANSI_BELOW_FUNC:
                ; range-check the binary value:
                ; motion keys should be from 2 to 6 inclusive
                ;
                CMP     AL,2
                JB      IP_ANSI_BELOW_MOTION
                CMP     AL,6
                JA      IP_ANSI_ABOVE_MOTION

                ; now translate the motion key into a PC keycode
                ; and put it into the keyboard buffer
                ;
                ; allows SHIFT/CTRL/ALT to be applied once
                ;
                SUB     AL,2                        ; now zero-based
                XOR     AH,AH
                SHL     AX,1                        ; each entry is 2 bytes
                MOV     BX,AX
                XOR     AH,AH
                MOV     AL,SHIFT_FLAG               ; multiply this by size of
                MOV     SHIFT_FLAG,0
                SHL     AX,1                        ;    row (10 bytes)
                ADD     BX,AX                       ;    as 2*AX + 8*AX
                SHL     AX,2
                ADD     BX,AX                       ; add/multiply done
                MOV     AX,WORD PTR MOTION_XLATE[ BX ]
                JMP     IP_110

IP_ANSI_BELOW_MOTION:
IP_ANSI_ABOVE_MOTION:
IP_ANSI_ABOVE_FUNC:
IP_NOT_FUNC_OR_MOTION:

IP_INVALID_ANSI:
                ; ignore the whole sequence, since if user wanted
                ; to cause the PC to accept a sequence starting
                ; with ESC [ they would have doubled the ESC
                ;
                MOV     ASCII_FLAG,0
                JMP     IP_10

IP_NOT_ANSI:
        endif
                MOV     AL,KEY_ONE              ;Get ASCII code
                MOV     AH,KEY_TWO              ;Get scan code
                CMP     AX,REBOOT_CODE          ;Check for remote reboot wanted
                JNZ     IP_NOT_REBOOT           ;No,then continue processing

        if disable_com_when_reboot
        if 0
                PUSH    BP
                PUSH    offset IP_81
                MOV     BP,SP
                JMP     IP_90
IP_81:
                POP     BP
        else
                CALL    RESET_COMM              ;mask IRQ, disable serial port, etc.
        endif
        endif

        if mask_all_when_reboot
                mov al, 0FFh
                out 21h, al
                out 0A1h, al
        endif

        if seoi_all_when_reboot
                cli

                ; send specific end-of-interrupt to both
                ; Intel 8259A priority interrupt controllers
                ; for all 8 possible interrupt levels
                ;
                mov al, 60h
IP_83:
                out 20h, al
                out 0A0h, al
                inc al
                cmp al, 67h
                jbe IP_83
        endif

        if disable_kb_when_reboot
IP_84:
                ; read the keyboard controller status port
                ;
                in al, 64h
                jmp $+2
                jmp $+2

                ; wait for the input buffer to be empty
                ;
                and al, 2
                jnz IP_84

                ; tell the keyboard controller to disable
                ; the keyboard
                ;
                mov al, 0ADH
                out 64h, al
                jmp $+2
                jmp $+2
        endif

        if sysflag_0_when_reboot
                ; disable interrupts
                ;
                cli

                ; wait for the input buffer to be empty
                ;
IP_85A:
                in al, 64h
                test al, 2
                jnz IP_85A

                ; give the keyboard controller the command
                ; to change its command byte
                ;
                mov al, 60H
                out 64h, al

                ; wait for the input buffer to be empty
                ;
IP_85B:
                in al, 64h
                test al, 2
                jnz IP_85B

                ; give the keyboard controller a command byte
                ; which has system flag (bit 2, value 04h)
                ; cleared so ROM BIOS thinks system was reset
                ;
        ;bad    mov al, 61h
                mov al, 0
                out 60h, al

                ; wait for the input buffer to be empty
                ; and for the system flag to be cleared
                ; at the status port
                ;
IP_85C:
                in al, 64h
                test al, 6
                jnz IP_85C
        endif

        if ctrl_alt_del_reboot
                MOV     WORD PTR ES:[72H],1234H ;soft reboot like local ctrl-alt-del
;               JMP     0F000H:0FFF0H           ;masm/tasm doesn't like far call
                DB      0EAH                    ; must hand-assemble opcode,
                DW      0FFF0H                  ; dest offset,
                DW      0F000H                  ; dest segment
        endif

        if triton_reboot_soft OR triton_reboot_hard
                MOV     DX,0CF9H                ;Intel 82371FX PIIX reset control register
        if triton_reboot_soft
                MOV     AL,4                    ;6=hard 4=soft
        else
                MOV     AL,6                    ;6=hard 4=soft
        endif
                OUT     DX,AL
                AND     AL, NOT 4               ;2=hard 0=soft
                OUT     DX,AL
        endif

        if kb_controller_reboot
IP_86:
                ; read the keyboard controller status port
                ;
                in al, 64h
                jmp $+2
                jmp $+2

                ; wait for the input buffer to be empty
                ;
                and al, 2
                jnz IP_86

                ; tell the keyboard controller to reset
                ; the CPU by pulsing the least significant
                ; bit of its output port low for 6 u secs
                ;
                mov al, 0FEH
                out 64h, al
                jmp $+2
                jmp $+2
        endif

IP_NOT_REBOOT:
                CMP     AX,EXIT_CODE            ;Check for exit code
                JNZ     IP_100                  ;No,then continue processing
IP_90:                                          ;Otherwise reset connect & exit
                TEST    SMILE_FLAG,1            ;Smile flag set ?
                JE      IP_95                   ;No, skip resetting smile face
                PUSH    AX                      ;Store AX
                PUSH    ES                      ;Store ES
                PUSH    DI                      ;Store DI
                MOV     AX,VIDEO_SEGMENT        ;AX = VIDEO_SEGMENT
                MOV     ES,AX                   ;ES = VIDEO_SEGMENT
                MOV     DI,158D                 ;Bump DI 79 chars
                MOV     AL,SPACE                ;Blank space
                CALL    PUT_VIDEO_DATA          ;Write Video RAM
                POP     DI                      ;Restore DI
                POP     ES                      ;Restore ES
                POP     AX                      ;Restore AX
IP_95:
        if manned_is_dumb_ansi
                ; clear the xterm title bar
                ;
                MOV     AH,1
                MOV     AL,ESC_KEY
                CALL    PUT_BUFF_DATA
                MOV     AL,']'
                CALL    PUT_BUFF_DATA
                MOV     AL,'0'
                CALL    PUT_BUFF_DATA
                MOV     AL,';'
                CALL    PUT_BUFF_DATA
                PUSH    SI
                MOV     SI,OFFSET OFFLINE_NAME
                CALL    LOAD_ZSTRING
                MOV     AL,BELL
                CALL    PUT_BUFF_DATA
                POP     SI
        endif

                MOV     AUTOBAUD_FLAG,1
                CALL    CU_HANGING_UP           ;drop DTR and RTS, bring 'em back later
                CALL    CLEAR_INBUFF            ;Clear input buffer
                CLI                             ;Disable interrupts
                MOV     CONNECT_FLAG,0          ;Reset the connect flag
                MOV     ASCII_FLAG,0            ;Reset ASCII flag
                MOV     SHIFT_FLAG,0            ;Reset shift flag
                MOV     BYTE PTR ES:[17H],0     ;Reset any shift status
                MOV     SP,BP                   ;Clean the stack
                RET                             ;And exit unattended routine
IP_100:
                CMP     AX,CLEAR_CODE           ;Clear code received
                JNZ     IP_NOT_CLEAR            ;No... skip clear code coding
                CALL    CLEAR_VIDEO             ;Rewrite screen
                MOV     ASCII_FLAG,0            ;Reset ASCII flag
                JMP     IP_10                   ;Get next character
IP_NOT_CLEAR:
        if manned_is_dumb_ansi
                ; check for a command to apply SHIFT or CTRL or ALT
                ; to the next single character
                ;
                CMP     AX,SHIFT_CODE
                JNE     IP_NOT_SHIFT_CODE
                MOV     SHIFT_FLAG,1
                MOV     ASCII_FLAG,0
                JMP     IP_10
IP_NOT_SHIFT_CODE:
                CMP     AX,CTRL_CODE
                JNE     IP_NOT_CTRL_CODE
                MOV     SHIFT_FLAG,2
                MOV     ASCII_FLAG,0
                JMP     IP_10
IP_NOT_CTRL_CODE:
                CMP     AX,ALT_CODE
                JNE     IP_NOT_ALT_CODE
                MOV     SHIFT_FLAG,3
                MOV     ASCII_FLAG,0
                JMP     IP_10
IP_NOT_ALT_CODE:

                ; shove all 3 into the keyboard buffer:
                ; escape, KEY_ONE, and KEY_TWO
                ;
                XOR     AH,AH
                MOV     AL,ESC_KEY
                CALL    PUT_KEY_DATA
                MOV     AL,KEY_ONE
                CALL    PUT_KEY_DATA
                MOV     AL,KEY_TWO
        endif
IP_110:                                         ;
                CALL    PUT_KEY_DATA            ;And stuff in keyboard buffer
                MOV     ASCII_FLAG,0            ;Reset ASCII flag for next data
                JMP     IP_10                   ;Process next byte

IP_120:
        if manned_is_dumb_ansi
                ; if this is the escape character, expect a valid
                ; sequence next time
                ;
                XOR     AH,AH
                CMP     AL,ESC_KEY
                JNE     IP_NOT_ESC_KEY
                INC     ASCII_FLAG
                JMP     IP_10
IP_NOT_ESC_KEY:

                ; if the character is DELETE, we want to map
                ; all 4 possible shifts for it, so it must be
                ; handled before the general just-special-if-ALT
                ; that we do for printable ASCII characters
                ;
                CMP     AL,DELETE
                JNE     IP_ASCII_NOT_DELETE

                ; translate into a PC keycode and stuff into KB buffer
                ;
                ; allows SHIFT/CTRL/ALT to be applied once
                ;
                ; this code sure is a lot simpler because we don't
                ; have to include the ASCII value in the index!
                ;
                XOR     BH,BH
                MOV     BL,SHIFT_FLAG               ; multiply this by size of
                MOV     SHIFT_FLAG,0
                SHL     BX,1                        ;    row (2 bytes)
                MOV     AX,WORD PTR DELETE_XLATE[ BX ]
                JMP     IP_110
IP_ASCII_NOT_DELETE:

                ; if the character is TAB, we want to map
                ; all 4 possible shifts for it, so it must be
                ; handled before the general just-special-if-ALT
                ; that we do for printable ASCII characters
                ;
                CMP     AL,TAB
                JNE     IP_ASCII_NOT_TAB

                ; translate into a PC keycode and stuff into KB buffer
                ;
                ; allows SHIFT/CTRL/ALT to be applied once
                ;
                ; this code sure is a lot simpler because we don't
                ; have to include the ASCII value in the index!
                ;
                XOR     BH,BH
                MOV     BL,SHIFT_FLAG               ; multiply this by size of
                MOV     SHIFT_FLAG,0
                SHL     BX,1                        ;    row (2 bytes)
                MOV     AX,WORD PTR TAB_XLATE[ BX ]
                JMP     IP_110
IP_ASCII_NOT_TAB:

                ; sorry, sorry, but there is one more non-ALT
                ; shifted regular ASCII character:
                ;
                ; CTRL-BREAK needs to become DELETE
                ;
                CMP     AL,BACKSPACE
                JNE     IP_ASCII_NOT_BACKSPACE
                CMP     SHIFT_FLAG,2
                JNE     IP_ASCII_NOT_CTRL_BACKSPACE
                MOV     SHIFT_FLAG,0
                XOR     AH,AH
                MOV     AL,DELETE
                JMP     IP_110
IP_ASCII_NOT_CTRL_BACKSPACE:
IP_ASCII_NOT_BACKSPACE:

                ; if we aren't in ALT shift state,
                ; treat the character we got as
                ; ASCII and put it into KB buffer
                ;
                ; but it clears the shift state either way
                ;
                CMP     SHIFT_FLAG,3
                MOV     SHIFT_FLAG,0
                JNE     IP_110

                ; deal with alphabetic characters first,
                ; since that is the largest range handled
                ; by one table, and we want to take the least
                ; time to handle the most frequent cases
                ;
                ; does NOT alias { | } to [ \ ]
                ;
                CMP     AL,'A'
                JB      IP_ASCII_BELOW_ALPHA
                CMP     AL,']'
                JB      IP_ASCII_IS_ALPHA
                CMP     AL,'a'
                JNB     IP_ASCII_NOT_BETWEEN_ALPHA
                JMP     IP_ASCII_BETWEEN_ALPHA
IP_ASCII_NOT_BETWEEN_ALPHA:
                CMP     AL,'z'
                JNA     IP_ASCII_IS_LOWERCASE
                JMP     IP_ASCII_ABOVE_ALPHA
IP_ASCII_IS_LOWERCASE:

                ; need to convert lowercase to uppercase
                ;
                SUB     AL,32

IP_ASCII_IS_ALPHA:
                ; normalize to zero-based, 2 bytes per entry
                ;
                SUB     AL,'A'
                SHL     AL,1

                ; get the entry and stuff the keyboard buffer
                ;
                XOR     AH,AH
                MOV     BX,AX
                MOV     AX,WORD PTR ALPHA_XLATE[ BX ]
                JMP     IP_110

IP_ASCII_BELOW_ALPHA:
                ; check for digits range
                ;
                CMP     AL,'0'
                JB      IP_ASCII_BELOW_DIGITS
                CMP     AL,'9'
                JA      IP_ASCII_ABOVE_DIGITS

                ; normalize to zero-based, 2 bytes per entry
                ;
                SUB     AL,'0'
                SHL     AL,1

                ; get the entry and stuff the keyboard buffer
                ;
                XOR     AH,AH
                MOV     BX,AX
                MOV     AX,WORD PTR DIGIT_XLATE[ BX ]
                JMP     IP_110

IP_ASCII_BELOW_DIGITS:
                ; check for onesie-twosies below '0'
                ;
                CMP     AL,CR
                JNE     IP_ASCII_NOT_ENTER
                MOV     AX,WORD PTR ENTER_XLATE
                JMP     IP_110
IP_ASCII_NOT_ENTER:
                CMP     AL,'-'
                JNE     IP_ASCII_NOT_DASH
                MOV     AX,WORD PTR DASH_XLATE
                JMP     IP_110
IP_ASCII_NOT_DASH:
                CMP     AL,'/'
                JNE     IP_ASCII_NOT_SLASH
                MOV     AX,WORD PTR SLASH_XLATE
                JMP     IP_110
IP_ASCII_NOT_SLASH:
                CMP     AL,'.'
                JNE     IP_ASCII_NOT_PERIOD
                MOV     AX,WORD PTR PERIOD_XLATE
                JMP     IP_110
IP_ASCII_NOT_PERIOD:
                CMP     AL,','
                JNE     IP_ASCII_NOT_COMMA
                MOV     AX,WORD PTR COMMA_XLATE
                JMP     IP_110
IP_ASCII_NOT_COMMA:
                CMP     AL,QUOTE
                JNE     IP_ASCII_NOT_QUOTE
                MOV     AX,WORD PTR QUOTE_XLATE
                JMP     IP_110
IP_ASCII_NOT_QUOTE:
                ; leave the ASCII value as-is
                ;
                JMP     IP_110

IP_ASCII_ABOVE_DIGITS:
                ; check for onesie-twosies between '9' and 'A', exclusive
                ;
                CMP     AL,'='
                JNE     IP_ASCII_NOT_EQUALS
                MOV     AX,WORD PTR EQUALS_XLATE
                JMP     IP_110
IP_ASCII_NOT_EQUALS:
                CMP     AL,';'
                JNE     IP_ASCII_NOT_SEMICOLON
                MOV     AX,WORD PTR SEMICOLON_XLATE
                JMP     IP_110
IP_ASCII_NOT_SEMICOLON:
                ; leave the ASCII value as-is
                ;
                JMP     IP_110

IP_ASCII_BETWEEN_ALPHA:
                ; check for onesie-twosies between ']' and 'a', exclusive
                ;
                CMP     AL,BACKQUOTE
                JNE     IP_ASCII_NOT_BACKQUOTE
                MOV     AX,WORD PTR BACKQUOTE_XLATE
                JMP     IP_110
IP_ASCII_NOT_BACKQUOTE:
                ; leave the ASCII value as-is
                ;
                JMP     IP_110

IP_ASCII_ABOVE_ALPHA:
                ; nothing we care about above 'z',
                ; so leave the ASCII value as-is
                ;
                JMP     IP_110
        else
;
;       Check to see if expecting shift data
;
                CMP     SHIFT_FLAG,0            ;Check shift flag
                JZ      IP_130                  ;0=not expecting data here
                MOV     ES:[17H],AL             ;And save in shift status
                MOV     SHIFT_FLAG,0            ;Reset shift flag for next data
                JMP     IP_10                   ;Process next byte
;
;       Check to see if it's a sync byte
;
IP_130:
                CMP     AL,0FEH                 ;Check for ASCII sync byte
                JNZ     IP_140                  ;If not then check for shift
                INC     ASCII_FLAG              ;Indicate received FEh
                JMP     IP_10                   ;Process next byte
IP_140:
                CMP     AL,0FDH                 ;Check for shift sync byte
                JNZ     IP_150                  ;If not then throw away
                INC     SHIFT_FLAG              ;Indicate received FDh
IP_150:
                JMP     IP_10                   ;Process till buffer empty
        endif
IP_EXIT:
                POP     ES                      ;Restore registers
                POP     DX
        if manned_is_dumb_ansi
                POP     BX
        endif
                POP     AX
                RET
INPUT_PROCESSING ENDP

;----------------------------------------------------------------------
;  Reset comm port.  Resets the comm port 8250 and disables interrupt
;  driven I/O.  Reset the 8259 mask.
;----------------------------------------------------------------------
RESET_COMM      PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    ES
                PUSH    DS
                PUSH    CS
                POP     DS
                CMP     INSTALLED_SEG,0         ;Maybe this is uninstall
                JZ      EM_10
                MOV     ES,INSTALLED_SEG        ;Get location of resident prog
                MOV     DX,ES:COMM_PORT         ;Get the comm port it used
                MOV     CL,BYTE PTR ES:SET_VEC  ;Determine mask bit to change
                JMP     EM_20
EM_10:
                MOV     DX,CS:COMM_PORT         ;Get port address
                MOV     CL,BYTE PTR SET_VEC     ;Determine mask bit to change
EM_20:
                INC     DX                      ;Interrupt enable register
                XOR     AL,AL
                OUT     DX,AL                   ;Disable all interrupts

        if enable_uart_fifos
                ; clear and disable the RX and TX FIFO's
                ; (only applicable to 16550A)
                ; in case another program only expects an 8250 UART
                ;
                INC     DX                      ;FIFO control register
                MOV     AL,00000110B
                OUT     DX,AL                   ;Disable all interrupts
        endif

                XOR     AH,AH
                SUB     DX,2
                CALL    SET_CARRIER

                SUB     CL,8                    ;convert interrupt # to IRQ level
                MOV     AH,1
                SHL     AH,CL                   ;Select the mask bit to set
                IN      AL,21H                  ;Get current 8259 int mask
                OR      AL,AH                   ;Set appropriate int bit
                OUT     21H,AL                  ;And set new 8259 mask

        if force_irq_priorities
                ; reorder IRQ priorities so they are the way (we assume)
                ; they were when we first loaded:   0 1 2 3 4 5 6 7
                ;
                ; see INIT_SERIAL for explaination of this
                ;
                MOV     AL, 0C0h
                OUT     20h, AL
        endif

                POP     DS
                POP     ES                      ;Restore segment
                RET
RESET_COMM      ENDP

;----------------------------------------------------------------------
;  Formats the data in temporary video buffer and puts it into the
;  output buffer
;  Input - CX=0 Transfer character data to output buffer
;          CX=1 Transfer attribute data to output buffer
;  Output - SCROLL_OFFSET=0
;  Changes - AX, BX, CX, SI, DI
;----------------------------------------------------------------------
TRANSFER_BLOCK  PROC    NEAR
        if manned_is_dumb_ansi
                ; save DX register
                ;
                PUSH    DX

                ; repeats are not implemented for dumb ANSI
                ; terminals yet, so don't bother looking for one
                ;

                ; convert the block number to a character row
                ; move the dumb terminal's cursor to the start of
                ; the block, so we can put text/attributes there
                ;
                MOV     BL,BLOCK_COLUMN
                MOV     BH,BLOCK_ROW
                CALL    SEND_CURSOR_ESCAPE
        else
                PUSH    CX                      ;Save the data type
;
; If the data is all the same only send it once, and set the repeat flag
;
                MOV     SI,TEMP_VIDEO_PTR       ;Point
                ADD     SI,CX                   ;Adjust for char. or attr.
                XOR     BL,BL                   ;Initialize for non-repeating
                MOV     CX,BLOCK_SIZE           ;Number of bytes to compare
                MOV     AH,[SI]                 ;Get the first byte in block
TB_10:
                LODSB                           ;Get next byte from the block
                INC     SI                      ;Adjust for word
                CMP     AH,AL                   ;Verify all the same
                LOOPZ   TB_10                   ;For the entire block
                JNZ     TB_20                   ;Not the same, send block
                MOV     BL,2                    ;set repeat flag
TB_20:
                POP     CX                      ;Restore data type
                MOV     AL,0FCH                 ;Start with base sync byte
                OR      AL,CL                   ;Include data type bit
                OR      AL,BL                   ;Include repeat bit
                MOV     AH,1                    ;Use output buffer
                CALL    PUT_BUFF_DATA           ;Put sync byte in output buff.
                MOV     AL,BLOCK_COUNT
                MOV     AH,1                    ;Use output buffer
                CALL    PUT_BUFF_DATA           ;Send the block number
        endif
                MOV     SI,TEMP_VIDEO_PTR       ;Point to block of video data
                MOV     DI,VIDEO_COPY           ;Point to video copy
                ADD     DI,VID_RAM_OFFSET       ;Adjust for block offset
        if manned_is_dumb_ansi
        else
                ADD     SI,CX                   ;Adjust for char. or attr.
                ADD     DI,CX                   ;Adjust for character or attribute
        endif
                MOV     CX,BLOCK_SIZE           ;Number of bytes to send
                CLD                             ;Forward direction
TB_30:
                LODSW                           ;Get the unmatched data
                MOV     DX,AX                   ;save video char and attr
        if manned_is_dumb_ansi
                STOSW                           ;save both char & attr in shadow

                ; determine whether we need to swap foreground
                ; and background attributes, based upon the character
                ;
                ; leaves zero flag cleared if swapping required
                ;
                MOV     BX,FONT_XLATE_ATTR
                XLATB
                TEST    AL,USE_REVERSE_VIDEO

                CALL    SEND_ATTRIBUTE_ESCAPE

                MOV     AH,1                    ;use output buffer

                ; decide whether we want the graphic font for this character
                ;
                TEST    AL,USE_GRAPHIC_FONT

                JZ      TB_NOT_GRAPHIC

                ; if previous character wasn't graphic,
                ; switch to the graphic character set
                ;
                CMP     PREV_GRAPHIC,1
                JE      TB_XLATE_CHAR
                MOV     PREV_GRAPHIC,1
                MOV     AL,SHIFTOUT
                CALL    PUT_BUFF_DATA
                JMP     TB_XLATE_CHAR

TB_NOT_GRAPHIC:

                ; if previous character was graphic,
                ; switch to the normal character set
                ;
                CMP     PREV_GRAPHIC,0
                JE      TB_XLATE_CHAR
                MOV     PREV_GRAPHIC,0
                MOV     AL,SHIFTIN
                CALL    PUT_BUFF_DATA

TB_XLATE_CHAR:
                ; translate the character into something XTERM has
                ;
                MOV     AL,DL                   ;restore just the character
                MOV     BX,FONT_XLATE
                XLATB

TB_SEND_CHAR:
                CALL    PUT_BUFF_DATA           ;send this character

                ; increment the column we think the dumb terminal
                ; has its cursor positioned at, but saturate at
                ; the width we think our screen is, because we
                ; disabled auto-wraparound during CONNECT_UNATTENDED
                ;
                MOV     AL,BYTE PTR PREV_POSITION
                CMP     AL,CHARS_PER_LINE
                JAE     TB_COLUMN_SATURATED
                INC     AL
                MOV     BYTE PTR PREV_POSITION,AL

TB_COLUMN_SATURATED:
        else                                    ;not manned_is_dumb_ansi
                OR      BL,BL                   ;Check the repeat flag
                JNZ     TB_40                   ;Don't send repeat data
                MOV     AH,1                    ;Use the output buffer
                CALL    PUT_BUFF_DATA           ;send this character
                STOSB                           ;Save video data in copy
                INC     DI                      ;Adjust for word offset
TB_40:
        endif                                   ;manned_is_dumb_ansi
                DEC     CX
                JZ      TB_45                   ;send BLOCK_SIZE bytes of data
                JMP     TB_30
TB_45:

        if manned_is_dumb_ansi
                ; restore DX register
                ;
                POP     DX
        else
                OR      BL,BL                   ;If repeat data then send one
                JZ      TB_50
                MOV     AH,1
                CALL    PUT_BUFF_DATA
        endif
TB_50:
        if detect_scrolling
                MOV     SCROLL_OFFSET,0         ;video offset = shadow offset
                MOV     SCROLL_OFFSET_ROW,0
                MOV     SCROLL_FOUND,0          ;didn't find any scroll
        endif
                RET
TRANSFER_BLOCK  ENDP

;----------------------------------------------------------------------
;  Put a byte of data into the unattended keyboard buffer.
;  Input - AL contains ASCII to be put into buffer.
;          AH contains scan code to be put into buffer.
;  Output - Carry Set - Byte placed in buffer successfully
;           Carry Reset - Buffer full, byte not stored in buffer
;  Changes - Nothing
;----------------------------------------------------------------------
PUT_KEY_DATA    PROC    NEAR
                PUSH    BX                      ;Save registers
                PUSH    DI

                ; if this is a carriage return
                ; fix up the keycode portion, since Borland's
                ; turbo debugger looks there during dialog boxes
                ;
                CMP     AL,CR
                JNE     PK_NOT_CR
                MOV     AH,KEYCODE_ENTER
PK_NOT_CR:

                CLI                             ;Don't allow interrupts
                MOV     BX,ES:[1CH]             ;Get the buffer tail pointer
                MOV     DI,BX                   ;Save the tail pointer
                INC     BX                      ;Point to next data storage
                INC     BX
                CMP     BX,3EH                  ;Beyond end of buffer area?
                JNE     PK_10                   ;No, then don't reset
                MOV     BX,1EH                  ;Yes, reset to buffer begin
PK_10:
                CMP     BX,ES:[1AH]             ;Test for buffer full
                JE      PK_FULL                 ;If the same, don't save it
                                                ; exit, carry is already reset
                MOV     ES:[DI],AX              ;Store the data in buffer
                MOV     ES:[1CH],BX             ;Save new tail pointer
                STC                             ;Indicate data stored OK
        if manned_is_dumb_ansi
                JMP     PK_EXIT
        endif
PK_FULL:
        if manned_is_dumb_ansi
                ; cause the manned terminal to beep
                ;
                PUSHF
                PUSH    AX
                MOV     AH,1                    ;use output buffer
                MOV     AL,BELL
                CALL    PUT_BUFF_DATA
                POP     AX
                POPF
        endif
PK_EXIT:
                POP     DI
                POP     BX
                STI                             ;Enable interrupts
                RET
PUT_KEY_DATA   ENDP

;======================================================================
;  COMMON ROUTINES - These routines are common to both the unattended
;  processing portion of the program and the manned processing portion.
;======================================================================

;----------------------------------------------------------------------
; Store the video data in AL in the memory location pointed to by DI
; INPUT - AL video data, ES:DI video RAM destination
; OUTPUT - Nothing
; Changes - DI is incremented
;----------------------------------------------------------------------
PUT_VIDEO_DATA  PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
;
; Check desnow flag, if set wait for horizontal sync to put in video RAM
;
                PUSH    BX                          ; Save Registers
                PUSH    DX
                TEST    DESNOW_FLAG,1               ; Check Desnow flag
                JZ      PV_30                       ; No, skip over
                CLD                                 ; Forward
                XCHG    AX,BX                       ; Save video data
                MOV     DX,3DAH                     ; CGA status port
PV_10:
                IN      AL,DX                       ; Get status byte
                TEST    AL,1                        ; Test display enable
                JNZ     PV_10                       ; If in Hor. sync, wait
PV_20:
                IN      AL,DX                       ; Get status byte
                TEST    AL,1                        ; Test display enable
                JZ      PV_20                       ; Wait for Hor. sync
                XCHG    AX,BX                       ; Get video data
PV_30:
                STOSB                               ; Put into video RAM
                POP     DX                          ; Restore registers
                POP     BX
                RET
PUT_VIDEO_DATA  ENDP

;----------------------------------------------------------------------
;  Wait for a buffer to be empty.
;  Input - AH - Buffer to use 0=Input buffer, 1=Output buffer.
;  Output - none
;  Changes - nothing
;----------------------------------------------------------------------
WAIT_BUFF_EMPTY PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    BX                      ;Save registers
                PUSH    SI

                CMP     AH,0                    ;Check which buffer to use
                JNZ     WBE_10                  ;Jump for output buffer
                MOV     SI,OFFSET IN_BUFF_HEAD  ;Point to input buffer
                JMP     WBE_20                  ;Skip over out buffer
WBE_10:
                MOV     SI,OFFSET OUT_BUFF_HEAD ;Point to output buffer
WBE_20:
                CLI                             ;Don't allow interrupts
                MOV     BX,[SI]                 ;Get the buffer head pointer
                CMP     BX,2[SI]                ;Test for data in buffer
                STI                             ;Interrupts OK now
                JNE     WBE_20                  ;Try again if not empty

                POP     SI                      ;Restore registers
                POP     BX
                RET
WAIT_BUFF_EMPTY ENDP

;----------------------------------------------------------------------
;  Get a byte of data from a buffer.  Byte pointed to by head pointer
;  is next data byte.  If head=tail, no data in buffer.
;  Input - AH - Buffer to use 0=Input buffer, 1=Output buffer.
;  Output - Carry Set - Byte from buffer is in AL
;           Carry Reset - No data in buffer
;  Changes - AL
;
;  we check for emptiness BEFORE messing with pointer wrap
;----------------------------------------------------------------------
GET_BUFF_DATA   PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    BX                      ;Save registers
                PUSH    SI
                CMP     AH,0                    ;Check which buffer to use
                JNZ     GD_10                   ;Jump for output buffer
                MOV     SI,OFFSET IN_BUFF_HEAD  ;Point to input buffer
                JMP     GD_20                   ;Skip over out buffer
GD_10:
                MOV     SI,OFFSET OUT_BUFF_HEAD ;Point to output buffer
GD_20:
                CLI                             ;Don't allow interrupts
                MOV     BX,[SI]                 ;Get the buffer head pointer
                CMP     BX,2[SI]                ;Test for data in buffer
                JE      GD_EXIT                 ;  no - exit; carry is already clear
                MOV     AL,[BX]                 ;Get the data
                INC     BX                      ;Point to data in buffer
                CMP     6[SI],BX                ;Beyond end of buffer area?
                JNE     GD_30                   ;No, then don't reset
                MOV     BX,4[SI]                ;Yes, reset to buffer begin
GD_30:
        if escape_bad_chars
                CMP     AH,0                    ;If was output buffer...
                JNZ     GD_40                   ;  yes - leave data as-is
                CMP     AL,BAD_ESCAPE           ;If this is the escape character...
                JNE     GD_40                   ;  no - return the data
                CMP     BX,2[SI]                ;Test for data in buffer
                JE      GD_EXIT                 ;  no - exit; carry is already clear
                MOV     AL,[BX]                 ;Get the data
                ADD     AL,BAD_OFFSET           ;Unmangle the data
                INC     BX                      ;Point to data in buffer
                CMP     6[SI],BX                ;Beyond end of buffer area?
                JNE     GD_40                   ;No, then don't reset
                MOV     BX,4[SI]                ;Yes, reset to buffer begin
GD_40:
        endif
                MOV     [SI],BX                 ;Save new head pointer
                STC                             ;Indicate data is in AL
GD_EXIT:
                POP     SI                      ;Restore registers
                POP     BX
                STI                             ;Enable interrupts
                RET
GET_BUFF_DATA   ENDP

;----------------------------------------------------------------------
;  Put a byte of data into a buffer.  Byte is stored at location
;  pointed to by tail pointer.  If necessary, byte is escaped.
;  Input - AL contains data to be put into buffer.
;          AH - Buffer to use 0=Input buffer, 1=Output buffer
;  Output - Carry Set - byte placed in buffer successfully
;           Carry Reset - Buffer full, byte not stored in buffer
;  Changes - Nothing
;
;  we check for fullness AFTER messing with pointer wrap
;----------------------------------------------------------------------
PUT_BUFF_DATA PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save registers
                PUSH    BX
                PUSH    DX
                PUSH    SI
                PUSH    DI
                PUSH    DS
                PUSH    CS                      ;Set data segment to CS
                POP     DS

                CMP     AH,0                    ;Check which buffer to use
                JNZ     PD_10                   ;Jump for output buffer
                MOV     SI,OFFSET IN_BUFF_HEAD  ;Point to input buffer
        if escape_bad_chars
                ; we do NOT want to escape characters being put
                ; into the input buffer, so we leave AH zero
                ;
        endif
                JMP     PD_20                   ;Skip over out buffer
PD_10:
                MOV     SI,OFFSET OUT_BUFF_HEAD ;Point to output buffer
        if escape_bad_chars
                ; remember whether this char needs escaping
                ;
                MOV     BX,BAD_CHARS
                XOR     AH,AH
                ADD     BX,AX
                MOV     AH,[BX]
        endif
PD_20:
                CLI                             ;Don't allow interrupts
                MOV     BX,2[SI]                ;Get the buffer tail pointer
                MOV     DI,BX                   ;Save the tail pointer
                INC     BX                      ;Point to next data storage
                CMP     6[SI],BX                ;Beyond end of buffer area?
                JNE     PD_30                   ;No, then don't reset
                MOV     BX,4[SI]                ;Yes, reset to buffer begin
PD_30:
                CMP     BX,[SI]                 ;Test for buffer full
        if escape_bad_chars
                JE      PD_34                   ;  yes - exit, carry is reset
                OR      AH,AH                   ;Does char need escaping?
                JZ      PD_38                   ;  no - go ahead with data char
                MOV    BYTE PTR [DI],BAD_ESCAPE ;Store the escape in buffer just in case
                SUB     AL,BAD_OFFSET           ;Mangle the data so it will pass
                MOV     DI,BX                   ;Data will go where BX points now
                INC     BX                      ;Point to next data storage
                CMP     6[SI],BX                ;Beyond end of buffer area?
                JNE     PD_32                   ;No, then don't reset
                MOV     BX,4[SI]                ;Yes, reset to buffer begin
PD_32:
                CMP     BX,[SI]                 ;Test for buffer full
        endif
                JNE     PD_38                   ;  no - go ahead with data char

PD_34:
                INC     TX_OVERRUNS             ;count these puppies
        if beep_when_tx_overruns
                ; beep
                ;
                PUSH    AX
                PUSH    BX
                PUSH    CX
                PUSH    DX
                MOV     AX, 0E07h
                XOR     BX, BX
                INT     10h
                POP     DX
                POP     CX
                POP     BX
                POP     AX
        endif
        if reinit_when_tx_overruns OR revector_when_tx_overruns
        if revector_when_tx_overruns
                ; reinstall the interrupt descriptor table register
                ; so the interrupt vector table isn't in high memory
                ;
                .286p
                DB      66h
                LIDT    IDT_LIMIT

                ; reinstall interrupt vector in low memory
                ;
                ; we must do this ourselves since (1) DOS vector
                ; may also be screwed up, and (2) DOS is unsafe
                ; to call in an interrupt handler anyways
                ;
                PUSH ES
                PUSH AX
                PUSH BX
                XOR BX, AX
                MOV ES, BX
                MOV BL, [SET_VEC]         ;actual addr=vector * 4
                SHL BX, 1
                SHL BX, 1

                ; need interrupts disabled here
                CLI
                MOV AX, OFFSET INT_COMM
                MOV ES:[BX], AX
                MOV AX, CS
                MOV ES:[BX + 2], AX
                ;STI

                POP BX
                POP AX
                POP ES
        endif
        if reinit_when_tx_overruns
                ; re-init serial port, IRQ masking
                ; in case a nasty exit from protected mode
                ; has screwed up the serial port, but left
                ; interrupts enabled
                ;
                ; P.S. you can tell interrupts are enabled
                ; if the keyboard responds, even though all
                ; you can do is reboot with CTRL-ALT-DELETE
                ; or cause a beep because host KB buffer is
                ; full
                ;
                CALL    INIT_SERIAL
        endif
        else
                ; clear the carry flag manually
                ; since the increment screwed it up
                ;
                CLC

                JMP     PD_40
        endif

PD_38:
                MOV     [DI],AL                 ;Store the data in buffer
                MOV     2[SI],BX                ;Save new tail pointer
                STC                             ;Indicate data stored ok
PD_40:
                PUSHF                           ;Save the flags
                CMP     COMM_INT_STA,0          ;Transmit int. running?
                JNZ     PD_60                   ;Yes, so exit
                MOV     AX,OUT_BUFF_HEAD        ;Is data in output buffer
                CMP     AX,OUT_BUFF_TAIL
                JZ      PD_60                   ;No, so exit
                MOV     COMM_INT_STA,1          ;Set transmitting data flag
                MOV     DX,COMM_PORT            ;Get port base address
                ADD     DX,5                    ;Line status register
PD_50:
                IN      AL,DX                   ;Make sure holding reg. empty
                TEST    AL,00100000B            ;Test the hold reg empty flag
                JZ      PD_50                   ;Loop if not empty
                SUB     DX,5                    ;Port base register
                MOV     AH,1                    ;Use the output buffer
                CALL    GET_BUFF_DATA           ;Get data from output buffer
                OUT     DX,AL                   ;Send the data out the port
PD_60:
                STI                             ;Enable interrupts
                POPF                            ;Restore flags
                POP     DS                      ;Restore registers
                POP     DI
                POP     SI
                POP     DX
                POP     BX
                POP     AX
                RET
PUT_BUFF_DATA   ENDP

;----------------------------------------------------------------------
;  Clear the input buffer
;  Input - Nothing
;  Output - Nothing
;  Changes - Nothing
;----------------------------------------------------------------------
CLEAR_INBUFF    PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save register
                CLI                             ;Disable interrupts
                MOV     AX,IN_BUFF_TAIL         ;Get buffer tail pointer
                MOV     IN_BUFF_HEAD,AX         ;Make head equal tail
                STI                             ;Enable interrupts
                POP     AX
                RET
CLEAR_INBUFF    ENDP

;----------------------------------------------------------------------
;  Clear the output buffer
;  Input - Nothing
;  Output - Nothing
;  Changes - Nothing
;----------------------------------------------------------------------
CLEAR_OUTBUFF   PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save register
                CLI                             ;Disable interrupts
                MOV     AX,OUT_BUFF_TAIL        ;Get buffer tail pointer
                MOV     OUT_BUFF_HEAD,AX        ;Make head equal tail
                STI                             ;Enable interrupts
                POP     AX
                RET
CLEAR_OUTBUFF   ENDP

;----------------------------------------------------------------------
;  Reset the modem and send the setup string to initialize
;  Input - AL - 0 use setup string 1, 1 use setup string 2
;  Output - Nothing
;  Changes - Nothing
;----------------------------------------------------------------------
RESET_MODEM     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                TEST    NULL_MODEM,1            ;If null modem don't reset
                JZ      RM_10
                MOV     CONNECT_FLAG,0          ;Reset connect flag
        if manned_is_dumb_ansi
        else
;
; After resetting connect flag, see if unattended and three bad passwords
; received, if so send an exit code to the remote computer so that the
; remote computer can exit to DOS.
;
                TEST    MANNED_FLAG,1           ;If host, then send exit code
                JNE     RM_5                    ; to remote system to exit
                CMP     PASSWORD_BAD,0          ;Three bad passwords ?
                JE      RM_5                    ;No, skip sending exit code
                CALL    CLEAR_OUTBUFF           ;Clear out buff to insure room
                MOV     AX,EXIT_CODE            ;Set up AL and AH to send
                MOV     AL,AH                   ; the exit code thru out buff.
                MOV     AH,1                    ;Use output buffer
                CALL    PUT_BUFF_DATA           ;Send char (exit_code)
RM_5:                                           ;
        endif
                JMP     RM_EXIT                 ;and return to caller
RM_10:
                PUSH    SI                      ;Save registers
                PUSH    DX
                PUSH    AX
                CMP     CONNECT_FLAG,0          ;Is modem connected?
                JZ      RM_20                   ;No then send setup only
                CALL    CLEAR_OUTBUFF           ;Empty the output buffer
                MOV     CONNECT_FLAG,0          ;Reset connect flag
                MOV     AL,1                    ;Wait a second for guard time
                CALL    DELAY
                MOV     SI,OFFSET MODEM_ESCAPE  ;Send modem escape code
                CALL    LOAD_ZSTRING
                MOV     AL,2                    ;Wait
                CALL    DELAY
                MOV     SI,OFFSET MODEM_HANGUP  ;Send modem hangup code
                CALL    LOAD_ZSTRING
                MOV     AL,1                    ;Wait for a second
                CALL    DELAY
RM_20:
                MOV     AL,SPEED_FLAG           ;Get speed flag
                MOV     CURRENT_SPEED,AL        ;Save as next starting speed
                CALL    SET_BAUD_RATE
                MOV     SPEED_COUNT,19
                MOV     SI,OFFSET MODEM_ATTENTION ;Point to modem attention
                CALL    LOAD_ZSTRING              ;Put it into output buffer
                MOV     AL,1                    ;Wait for a second
                CALL    DELAY
                POP     AX                      ;Get setup string to use
                CMP     AL,0                    ;Test for string 1
                JNZ     RM_30                   ;No, then use string 2
                MOV     SI,OFFSET MODEM_SETUP1  ;Point to modem setup string 1
                JMP     RM_40
RM_30:
                MOV     SI,OFFSET MODEM_SETUP2  ;Point to modem setup string 2
RM_40:
                CALL    LOAD_ZSTRING            ;Load setup string to modem
                MOV     AL,1                    ;Wait a second
                CALL    DELAY
                MOV     SI,OFFSET MODEM_SETUP3  ;Point to modem setup string 3
                CALL    LOAD_ZSTRING            ;Load setup string to modem
                CMP     SPEED_FLAG,2            ;If speed is 1200 or 2400
                JL      RM_50                   ; skip setup string 4
                MOV     AL,1                    ;Wait a second
                CALL    DELAY
                MOV     SI,OFFSET MODEM_SETUP4  ;Point to modem setup string 4
                CALL    LOAD_ZSTRING            ;Load setup string to modem
RM_50:
                MOV     AL,1                    ;Wait a second
                CALL    DELAY
                MOV     SI,OFFSET MODEM_SETUP5  ;Point to custom setup string 5
                CALL    LOAD_ZSTRING            ;Load setup string to modem
                CALL    CLEAR_INBUFF            ;Clear the input buffer
                POP     DX                      ;Restore registers
                POP     SI
RM_EXIT:
                RET
RESET_MODEM     ENDP

;----------------------------------------------------------------------
;  deasserts DTR and RTS if AH is zero, otherwise asserts them
;
;  when OUT2 is deasserted and a character comes in on the serial
;  port, no further characters cause an interrupt, which effectively
;  disables the port from then on, so it is unsafe to change it here
;
;  ???maybe OUT2 is some sort of flow control or interrupt disable???
;
;  Input - I/O base address of serial port in DX
;          AH nonzero to assert, zero to deassert control lines
;  Output - Nothing
;  Changes - Nothing
;----------------------------------------------------------------------
SET_CARRIER PROC NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save registers
                PUSH    DX
                ADD     DX,4                    ;Modem control register
                IN      AL,DX
                AND     AL,11111100B            ;turn off DTR and RTS
                OR      AH,AH
                JZ      SET_CARRIER_DEASSERT
                OR      AL,11B
SET_CARRIER_DEASSERT:
                OUT     DX,AL
                POP     DX                      ;Restore registers
                POP     AX
                RET
SET_CARRIER ENDP

;----------------------------------------------------------------------
;  Check carrier reads the carrier status signal and sets Z flag to
;  indicate status
;  Input - Nothing
;  Output - Zero flag 0 - Carrier
;           Zero flag 1 - No Carrier detected
;  Changes - Nothing
;----------------------------------------------------------------------
CHECK_CARRIER   PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save registers
        if respect_control_lines
        else
                TEST    NULL_MODEM,1            ;If null modem don't reset
                JZ      CC_10
                CMP     CONNECT_FLAG,0
                JZ      CC_EXIT                 ;not connected, send no carrier
                MOV     AH,1                    ;reset zero flag
                CMP     AH,0
                JMP     CC_EXIT                 ;and return to caller
CC_10:
        endif
                PUSH    DX
                MOV     DX,COMM_PORT            ;Get the comm base address
                ADD     DX,6                    ;Modem status register
                IN      AL,DX                   ;Get the current status
                TEST    AL,10000000B            ;Data carrier detect
                POP     DX                      ;Restore registers
CC_EXIT:
                POP     AX
                RET
CHECK_CARRIER   ENDP

;----------------------------------------------------------------------
;  DELAY - delay approximate number of seconds in AL
;  Input - AL
;  Output - Nothing (just waits till AL is zero)
;  Changes - Nothing
;----------------------------------------------------------------------
DELAY           PROC    NEAR
                PUSH    CX                      ;Save registers
                PUSH    DX
                PUSH    DI
                PUSH    AX
                XOR     AH,AH                   ;Read system time
                INT     1AH
                STI                             ;Enable interrupts Ver 1.1
                MOV     DI,DX                   ;Save low tick count
                MOV     SI,CX                   ;Save high tick count
                POP     AX                      ;Get number of seconds to delay
                PUSH    AX
                XOR     CX,CX                   ;Zero CX
                MOV     CL,AL                   ;Put seconds into loop counter
D_10:
                ADD     DI,19D                  ;Approximate counts in a second
                ADC     SI,0                    ;Add carry to SI
                LOOP    D_10
D_20:
                XOR     AH,AH                   ;Read system time
                INT     1AH
                STI                             ;Enable interrupts Ver 1.1
                CMP     SI,CX
                JNE     D_20
                CMP     DI,DX                   ;End of delay time
                JGE     D_20                    ;No, keep checking
                POP     AX                      ;Restore registers
                POP     DI
                POP     DX
                POP     CX
                RET
DELAY           ENDP


;----------------------------------------------------------------------
;  String at SI is placed in output buffer to be sent out serial port
;  Input - SI points to zero terminated string
;  Output - Nothing
;  Changes - SI
;----------------------------------------------------------------------
LOAD_ZSTRING    PROC    NEAR
                PUSH    AX                      ;Save register
                MOV     AH,1                    ;Use output buffer
                CLD                             ;Forward
LZ_10:
                LODSB                           ;Get a byte of data
                CMP     AL,0                    ;Check for zero
                JZ      LZ_EXIT                 ;Yes, then exit
                CALL    PUT_BUFF_DATA           ;No, put in output buffer
                JMP     LZ_10                   ;Process next data
LZ_EXIT:
                POP     AX                      ;Restore register
                RET
LOAD_ZSTRING    ENDP

;----------------------------------------------------------------------
;  Check for a key in the keyboard buffer, if one is there, get it
;  Input - Nothing
;  Output - Zero flag = 1 no key in buffer
;           Zero flag = 0 key is in AX
;  Changes - AX
;----------------------------------------------------------------------
GET_KEYSTROKE   PROC    NEAR
        ASSUME  CS:CSEG,DS:NOTHING,ES:NOTHING,SS:NOTHING
                MOV     AH,1                    ;Check for keystroke
                INT     16H                     ;Keyboard BIOS
                JZ      GK_EXIT                 ;No key so exit
                PUSHF                           ;Save the zero flag
                XOR     AH,AH                   ;Get the keystroke
                INT     16H
                POPF                            ;Restore the zero flag
GK_EXIT:
                RET
GET_KEYSTROKE   ENDP

        if escape_bad_chars
;----------------------------------------------------------------------
;  Initialize the table of characters needing escaping
;  Input - CX points to starting buffer location
;  Output - BAD_ESCAPE table and its pointer are initialized
;  Changes - CX, AX
;----------------------------------------------------------------------
INIT_ESCAPING   PROC NEAR
                push es
                push di
                push bx

                ; remember the table address
                ;
                mov BAD_CHARS, cx

                mov bx, cx
                mov di, cx

                ; by default, no characters need escaping
                ;
                cld                             ; go forward
                mov cx, 128d
                mov ax, cs
                mov es, ax
                xor ax, ax
                rep stosw

                ; cause characters 0, ^Q, ^S, ^], 157, 255,
                ; and the escape character itself to be escaped
                ;
                mov byte ptr [bx], 1
                mov byte ptr [bx + XON], 1
                mov byte ptr [bx + XOFF], 1
                mov byte ptr [bx + TELNET_ESCAPE + 128D], 1
                mov byte ptr [bx + 255d], 1
                mov byte ptr [bx + BAD_ESCAPE], 1

                ; allocate room for table
                ;
                mov cx, bx
                add cx, 256D

                ; restore registers
                ;
                pop bx
                pop di
                pop es

                ret
INIT_ESCAPING   endp
        endif

        if manned_is_dumb_ansi
;----------------------------------------------------------------------
;  Initialize the table of characters needing escaping
;  Input - CX points to starting buffer location
;  Output - BAD_ESCAPE table and its pointer are initialized
;  Changes - CX, AX
;----------------------------------------------------------------------
INIT_XLATE      PROC NEAR
                push es
                push di
                push bx

                ; save pointer to first table
                ;
                mov bx, cx

                ; by default, all characters translate to normal spaces
                ;
                cld                             ; go forward
                mov di, cx
                mov cx, 128d
                mov ax, cs
                mov es, ax
                mov ax, (SPACE * 256) + SPACE
                rep stosw
                mov cx, 128d                    ; DI already points at attributes
                mov ax, USE_NOTHING
                rep stosw

                ; normal 7-bit ASCII maps to itself
                ;
                mov di, bx
                add di, EXCLAIM
                mov al, EXCLAIM
INIT_XLATE_MORE_NORMAL:
                mov [di], al
                inc di
                inc al
                cmp al, DELETE
                jb INIT_XLATE_MORE_NORMAL

                ; special mappings
                ;
                mov byte ptr [bx + DOT], XTERM_SMALL_ZERO

                mov byte ptr [bx + REV_DOT], XTERM_SMALL_ZERO
                mov byte ptr [bx + REV_DOT + 256], USE_REVERSE_VIDEO

                mov byte ptr [bx + CIRCLE], XTERM_SMALL_ZERO

                mov byte ptr [bx + REV_CIRCLE], XTERM_SMALL_ZERO
                mov byte ptr [bx + REV_CIRCLE + 256], USE_REVERSE_VIDEO

                mov byte ptr [bx + UP_TRIANGLE], '^'
                mov byte ptr [bx + UP_ARROW], '^'

                mov byte ptr [bx + DOWN_TRIANGLE], 'v'
                mov byte ptr [bx + DOWN_ARROW], 'v'

                mov byte ptr [bx + LEFT_TRIANGLE], '<'
                mov byte ptr [bx + LEFT_ARROW], '<'

                mov byte ptr [bx + RIGHT_TRIANGLE], '>'
                mov byte ptr [bx + RIGHT_ARROW], '>'

                mov byte ptr [bx + MATH_MUCHLESS], XTERM_MATH_MUCHLESS
                mov byte ptr [bx + MATH_MUCHGREATER], XTERM_MATH_MUCHGREATER
                mov byte ptr [bx + MATH_NEGATION], XTERM_MATH_NEGATION

                mov byte ptr [bx + MONEY_CENTS], XTERM_MONEY_CENTS
                mov byte ptr [bx + MONEY_POUNDS], XTERM_MONEY_POUNDS
                mov byte ptr [bx + MONEY_YEN], XTERM_MONEY_YEN

                mov byte ptr [bx + UPSIDE_EXCLAIM], XTERM_UPSIDE_EXCLAIM
                mov byte ptr [bx + UPSIDE_QUESTION], XTERM_UPSIDE_QUESTION

                mov byte ptr [bx + SECTION_SIGN], XTERM_SECTION_SIGN
                mov byte ptr [bx + PARAGRAPH_SIGN], XTERM_PARAGRAPH_SIGN

                mov byte ptr [bx + TINY_DOT], XTERM_TINY_DOT

                mov byte ptr [bx + CENTER_PART_BOX + 256], USE_REVERSE_VIDEO

                ; temporary kluge?
                ;
                ; copy linedraw translation table to this one
                ; so even less code has to execute
                ;
                mov si, OFFSET LINEDRAW_XLATE
                mov di, bx
                add di, LIGHT_GRAY_BOX
                mov cx, UP_FULL_BOX - LIGHT_GRAY_BOX + 1
                rep movsb

                mov di, bx
                add di, LIGHT_GRAY_BOX + 256
                mov cx, UP_FULL_BOX - LIGHT_GRAY_BOX + 1
                mov al, USE_GRAPHIC_FONT
                rep stosb

                ; onesie-twosie graphic characters
                ;
                mov byte ptr [bx + DIAMOND], XTERM_LINEDRAW_DIAMOND
                mov byte ptr [bx + DIAMOND + 256], USE_GRAPHIC_FONT
                mov byte ptr [bx + FULL_BOX], SPACE
                mov byte ptr [bx + FULL_BOX + 256], USE_REVERSE_VIDEO

                ; restore pointer to tables
                ;
                mov cx, bx

                ; remember the table addresses,
                ; and allocate room for tables
                ;
                mov FONT_XLATE, cx
                add cx, 256d
                mov FONT_XLATE_ATTR, cx
                add cx, 256d

                ; restore registers
                ;
                pop bx
                pop di
                pop es

                ret
INIT_XLATE      endp
        endif

;----------------------------------------------------------------------
;  Initialize the buffer pointers for the input and output buffers.
;  Input - CX points to starting buffer location
;  Output - Input and output buffer points are initialized
;  Changes - BX, CX
;----------------------------------------------------------------------
INIT_BUFFERS    PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     BX,OFFSET IN_BUFF_HEAD  ;In buffer will be here
                MOV     WORD PTR [BX],CX        ; Set head pointer to buffer
                MOV     WORD PTR 2[BX],CX       ;Set tail pointer to buffer
                MOV     WORD PTR 4[BX],CX       ;Set begin of buffer
                ADD     CX,IN_BUFF_SIZE         ;CX Points to end of in buffer
IB_10:
                MOV     WORD PTR 6[BX],CX       ;Set end of buffer
                MOV     BX,OFFSET OUT_BUFF_HEAD ;Out buffer after in buffer
                MOV     WORD PTR [BX],CX        ; Set head pointer to buffer
                MOV     WORD PTR 2[BX],CX       ;Set tail pointer to buffer
                MOV     WORD PTR 4[BX],CX       ;Set begin of buffer
                ADD     CX,OUT_BUFF_SIZE        ;CX Points to end of out buffer
                MOV     WORD PTR 6[BX],CX       ;Set end of buffer
                RET
INIT_BUFFERS    ENDP

;----------------------------------------------------------------------
;  Change the interrupt 8 vector to the interrupt service routine of
;  PCREMOTE
;  Input - Nothing
;  Output - Interrupt vector 8 points to INT8
;  Changes - AX, DX, BX, ES
;----------------------------------------------------------------------
MODIFY_INT8     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
;
;       Change interrupt 8 vector
;
                MOV     AX,3508H                ;Get interrupt 8h vector
                INT     21H
                MOV     OLDINT_8,BX             ;And save it
                MOV     OLDINT_8[2],ES
                MOV     AX,2508H                ;Set interrupt 8h vector
                MOV     DX,OFFSET INT8          ; to point to new routine
                INT     21H
                RET
MODIFY_INT8     ENDP

;----------------------------------------------------------------------
;  Change the interrupt 21 vector to the interrupt service routine of
;  PCREMOTE
;  Input - Nothing
;  Output - Interrupt vector 21 points to INT21
;  Changes - AX, DX, BX, ES
;----------------------------------------------------------------------
MODIFY_INT21    PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AX,3521H                ;Get interrupt 21H vector
                INT     21H
                MOV     WORD PTR OLDINT_21,BX   ;And save it
                MOV     WORD PTR OLDINT_21[2],ES
                MOV     AX,2521H                ;Set interrupt 21H vector
                MOV     DX,OFFSET INT21         ; to point to new routine
                INT     21H
                RET
MODIFY_INT21    ENDP

;----------------------------------------------------------------------
;  Change the communication interrupt vector to the interrupt service
;     routine of PCREMOTE
;  Input - SET_VEC contains interrupt vector number
;  Output - communications interrupt vector points to INT_COMM
;  Changes - AX, DX, BX, ES
;----------------------------------------------------------------------
MODIFY_INTCOMM  PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    ES
                MOV     AL,BYTE PTR SET_VEC     ;Get the interrupt number
                MOV     AH,35H                  ;Get the old vector
                INT     21H
                MOV     WORD PTR OLDINT_COMM,BX ;And save it
                MOV     WORD PTR OLDINT_COMM[2],ES
                POP     ES
REMODIFY_INTCOMM:
                MOV     AL,BYTE PTR SET_VEC     ;Get the interrupt number
                MOV     AH,25H                  ;Set interrupt for comm vector
                MOV     DX,OFFSET INT_COMM      ; to point to new routine
                INT     21H
                RET
MODIFY_INTCOMM  ENDP

;----------------------------------------------------------------------
;  Set baud rate
;  Input - AL -0 1200 baud, 1 2400 baud, 2 4800 baud, 3 9600 baud
;               4 19200 baud, 5 38400 baud
;  Output - Nothing
;  Changes - Nothing
;----------------------------------------------------------------------
SET_BAUD_RATE   PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    DX                      ;Save registers
                PUSH    BX
                PUSH    SI
                PUSH    AX
                MOV     DX,COMM_PORT            ;Get port address
                ADD     DX,3                    ;Line control register
                MOV     AL,83H                  ;Toggle port address to
                OUT     DX,AL                   ; prepare to set baud rate
                SUB     DX,2                    ;Baud rate divisor MSB port
                POP     AX                      ;Restore baud rate
                PUSH    AX
                XOR     AH,AH                   ;Zero AH
                MOV     SI,AX                   ;Save in index register
                SHL     SI,1                    ;Multiply by 2, word address
                MOV     BX,OFFSET BAUD_RATE     ;Point to baud rates
                MOV     AL,[BX+SI]              ;Get baud rate MSB
                OUT     DX,AL                   ; and set it
                DEC     DX                      ;Baud rate divisor LSB port
                MOV     AL,1[BX+SI]             ;Get baud rate LSB
                OUT     DX,AL                   ; and set it
                ADD     DX,3                    ;Line control register
                MOV     AL,3                    ;8 data bits,1 stop,no parity
                OUT     DX,AL                   ;Set data bit pattern
                                                ; and toggle port address
                POP     AX                      ;Restore registers
                POP     SI
                POP     BX
                POP     DX
                RET
SET_BAUD_RATE   ENDP

;----------------------------------------------------------------------
;  Initialize the serial port
;  Input - COMM_PORT contains port address,
;          COMM_FLAG 0-comm1 1-comm2 2-comm3 3-comm4
;  Output - Serial port initialize for interrupt driven I/O
;  Changes - Nothing
;----------------------------------------------------------------------
INIT_SERIAL     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    DX                      ;Save registers
                PUSH    CX
                PUSH    BX
                PUSH    AX

                ; reinstall the interrupt descriptor table register
                ; so the interrupt vector table isn't in high memory
                ;
;
;       disable the interrupts on the 8250 and initialize DTR and RTS
;
                CLI                             ;Disable interrupts
                MOV     DX,COMM_PORT            ;Get UART base address
                ADD     DX,4                    ;Modem control register
                MOV     AL,00001011B            ;Set DTR, RTS, and OUT2
                OUT     DX,AL
;
;       set the baud rate of the UART and initialize line control register
;
                CMP     CONNECT_FLAG,0          ;If connect then use current
                JZ      IS_10
                MOV     AL,CURRENT_SPEED
                JMP     IS_20
IS_10:
                MOV     AL,SPEED_FLAG           ;Get speed flag
IS_20:
                MOV     CURRENT_SPEED,AL        ;Save the current speed
                MOV     SPEED_COUNT,19
                CALL    SET_BAUD_RATE

                ; clear and disable the RX and TX FIFO's
                ; (only applicable to 16550A)
                ;
                MOV     DX,COMM_PORT
                ADD     DX,2                    ;FIFO control register
                MOV     AL,00000110B
                OUT     DX,AL

        if enable_uart_fifos
                ; enable both TX and RX FIFO's and trigger IRQ at 8 full
                ;
                JCXZ    $+2
                JCXZ    $+2
                MOV     AL,10000001B
                OUT     DX,AL
        endif
;
;       set 8259 mask to enable the comm port interrupt
;
                MOV     CL,BYTE PTR SET_VEC     ;Get the interrupt number
                SUB     CL,8                    ; and adjust
                MOV     AH,1                    ;Mask for the interrupt
                SHL     AH,CL
                NOT     AH                      ;Invert
                IN      AL,21H                  ;Get current 8259 int mask
                AND     AL,AH                   ;Mask appropriate int bit
                OUT     21H,AL                  ;And set new 8259 mask

        if force_irq_priorities
                ; tell the master 8259 that IRQ3 (for COM2 or COM4) is
                ; highest priority by telling it that IRQ2 (for slave
                ; 8259) is lowest
                ;
                ; no way to set priorities for each IRQ individually;
                ; we can only rotate the existing ordering
                ;
                ; BEFORE (low) 7 6 5 4 3 2 1 0 (high priority)
                ; AFTER        2 1 0 7 6 5 4 3
                ;
                ; actually, this code is general enough to work for any
                ; IRQ level on master 8259, so when we are using COM1
                ; we get priorities   3 2 1 0 7 6 5 4   from low to high
                ;
                ; we do this so serial interrupts will still work and the
                ; manned PC can still get an image of the unattended screen
                ; and force a reboot of unattended PC, even though an
                ; interrupt handler for another IRQ (like ATM card) forgets
                ; to issue end-of-interrupt (EOI) command to 8259 (like when
                ; an interrupt handler causes a program exit)
                ;
                ; you may think it is nasty and intrusive, but it is better
                ; than having our timer (IRQ 0, INT 08h, highest priority
                ; originally) interrupt handler gratuitously issue the NSEOI
                ; or SEOI command on behalf of all IRQ's between it and the
                ;
                ; and either of these code changes is much more convenient
                ; to me as a human than having to phone up an operator and
                ; ask him to please find my PC at his site, read the screen
                ; to me, and press the reset button
                ;
                MOV     AL, CL
                SUB     AL, 1
                AND     AL, 7
                OR      AL, 0C0h
                OUT     20h, AL
        endif

;
;       enable the data received interrupt and reset the 8250
;
                MOV     DX,COMM_PORT
                ADD     DX,1                    ;Point to int enable reg
                MOV     AL,3                    ;Enable data received
                OUT     DX,AL                   ; and transmit empty int
                DEC     DX                      ;Point to base address
                MOV     CX,7                    ;Reset the serial port
IS_30:
                IN      AL,DX                   ;Read registers to reset
                INC     DX
                LOOP    IS_30

                CMP     COMM_INT_STA,0          ;Transmit int. running?
                JZ      IS_50                   ; no, so exit
                MOV     AX,OUT_BUFF_HEAD        ;Is data in output buffer
                CMP     AX,OUT_BUFF_TAIL
                JZ      IS_50                   ;No, so exit
                MOV     DX,COMM_PORT            ;Get port base address
                ADD     DX,5                    ;Line status register
IS_40:
                IN      AL,DX                   ;Make sure holding reg. empty
                TEST    AL,00100000B            ;Test the hold reg empty flag
                JZ      IS_40                   ;Loop if not empty
                SUB     DX,5                    ;Port base register
                MOV     AH,1                    ;Use the output buffer
                CALL    GET_BUFF_DATA           ;Get data from output buffer
                OUT     DX,AL                   ;Send the data out the port
IS_50:
                STI                             ;Enable interrupts
                POP     AX                      ;Restore registers
                POP     BX
                POP     CX
                POP     DX
                RET
INIT_SERIAL     ENDP

;======================================================================
;  Initialize routines.  The initialize routine for unattended is first
;  because it must remain resident.
;======================================================================

;-----------------------------------------------------------------------------
;  Initialize the unattended program.
;-----------------------------------------------------------------------------
INIT_UNATTENDED PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:CSEG,SS:NOTHING
                CALL    IU_10                   ;Init pointers,clear video copy
                CALL    MODIFY_INT8             ;Change INT 8 vector
                CALL    MODIFY_INT21            ;Change INT 21 vector
;
;       Terminate-Stay-Resident
;
                MOV     DX,VIDEO_COPY           ;Start of video copy
                ADD     DX,BYTES_PER_SCR        ;Allow room for video copy
                INT     27H                     ;Terminate-Stay-Resident
;
;       Initialize buffer pointers and video copy pointer
;
IU_10:
                PUSH    ES                      ;Save extra segment
                MOV     CX,OFFSET INITIALIZE    ;CX points to begin of buffer
                ADD     CX,STACK_SIZE           ;Leave room for TSR stack
        if escape_bad_chars
                CALL    INIT_ESCAPING           ;Initialize table of bad chars
        endif
        if manned_is_dumb_ansi
                CALL    INIT_XLATE              ;Init char translation table
        endif
                CALL    INIT_BUFFERS            ;Initalize the buffer pointers
                INC     CX                      ;Point to copy of video ram
                MOV     TEMP_VIDEO_PTR,CX       ;Save address of temp video buf
                ADD     CX,BLOCK_SIZEX2         ;Save room for temp video data
                MOV     VIDEO_COPY,CX           ;Save address of video copy

                MOV     AL,1                    ;Use setup string 2
                MOV     PASSWORD_BAD,0D         ;Do not send exit code
                CALL    RESET_MODEM             ;Reset the modem
                POP     ES                      ;Restore extra segment
                RET
INIT_UNATTENDED ENDP

;-----------------------------------------------------------------------------
;  INITIALIZE - Initialize the program.  Determine whether it is manned
;  or unattended by processing the command line.  Initialize the serial
;  port.  This area is overwritten by PCREMOTE host mode to conserve memory.
;-----------------------------------------------------------------------------
INITIALIZE      PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
;
;       Display copyright notice.
;
                MOV     DX,OFFSET COPYRIGHT     ;Display copyright notice
                MOV     AH,9                    ;Display string
                INT     21H
;
;       check to see if the program is already in memory
;
                MOV     BYTE PTR START,0        ;Zero word to avoid false match
                XOR     BX,BX                   ;Initialize search segment
                MOV     AX,CS                   ;Save current segment in AX
                CLD                             ;Clear direction flag
I_10:
                INC     BX                      ;Increment search segment
                CMP     AX,BX                   ;Reached current segment?
                JE      I_20                    ;Yes, PCREMOTE not resident
                MOV     ES,BX                   ;Point ES to search segment
                MOV     SI,OFFSET START         ;Start of compare area
                MOV     DI,SI                   ;Make offsets equal
                MOV     CX,16D                  ;Check 16 characters
                REPE    CMPSB                   ;Compare the strings
                JNE     I_10                    ;Compare failed
                MOV     INSTALLED_SEG,BX        ;Save the segment of resident

;               MOV     AH,09H                  ;Print string
;               MOV     DX,OFFSET PROG_RES      ;Display program resident mess.
;               INT     21H
;               INT     20H                     ;Terminate

;
;       Process command line for switches.
;
I_20:
                PUSH    CS                      ;Restore ES
                POP     ES
                MOV     BX,80H                  ;Point to command line length
                MOV     AH,[BX]                 ;Get command line length
I_30:
                OR      AH,AH                   ;Check for commands
                JNZ     I_40
                JMP     I_210                   ;None, so don't look
I_40:
                INC     BX                      ;Point to next data
                MOV     AL,[BX]                 ;Get data
                DEC     AH                      ;Decrement counter
I_50:
                CMP     AL,"?"                  ;Check for help menu (usage)
                JNE     I_60                    ;No, check for slash
                JMP     HELP                    ;Yes, help is requested !
I_60:
                CMP     AL,"/"                  ;Check for slash
                JNE     I_30                    ;No, jump to check next data
I_60_GOT_SLASH:
                OR      AH,AH                   ;Check for data after slash
                JNZ     I_70
                JMP     I_210                   ;No, so don't process
I_70:
                INC     BX                      ;Point to next data
                MOV     AL,[BX]                 ;Get data
                DEC     AH                      ;Decrement counter
                CMP     AL,"?"                  ;Check for help after a slash
                JNE     I_80                    ;No, check for comm2
                JMP     HELP                    ;Yes, help is requested !
I_80:
                CMP     AL,"1"                  ;Check for comm1
                JNE     I_90                    ;No, check for another switch
                MOV     COMM_FLAG,0             ;Set comm1 flag
                JMP     I_30                    ;Process next switch
I_90:
                CMP     AL,"2"                  ;Check for comm2
                JNE     I_100                   ;No, check for another switch
                MOV     COMM_FLAG,1             ;Set comm2 flag
                JMP     I_30                    ;Process next switch
I_100:
                CMP     AL,"3"                  ;Check for comm port 3
                JNE     I_110                   ;No, jump to check next data
                MOV     COMM_FLAG,2             ;Set comm3 flag
                JMP     I_30                    ;Process next switch
I_110:
                CMP     AL,"4"                  ;Check for comm port 4
                JNE     I_120                   ;No, jump to check next data
                MOV     COMM_FLAG,3             ;Set comm4 flag
                JMP     I_30                    ;Process next switch
I_120:
                OR      AL,20H                  ;Force data to lower case
                CMP     AL,"m"                  ;Check for manned switch
                JNE     I_130                   ;No, check for another switch
                XOR     MANNED_FLAG,1           ;Set manned flag
                JMP     I_30                    ;Process next switch
I_130:
                CMP     AL,"b"                  ;Check for baud rate
                JNE     I_170                   ;No, check for another switch
                OR      AH,AH                   ;Check for data after flag
                JZ      I_210                   ;No, so don't process
                INC     BX                      ;Point to next data
                MOV     AL,[BX]                 ;Get data
                DEC     AH                      ;Decrement counter
                SUB     AL,"0"                  ;parm must be >= "0"
                JB      I_50
                CMP     AL,7                    ;parm must be <= "7"
                JA      I_50
                MOV     SPEED_FLAG,AL           ;1200 to 115200 baud
                JMP     I_30
I_170:
                CMP     AL,"d"                  ;Check for desnow
                JNE     I_180                   ;No, process next switch
                XOR     DESNOW_FLAG,1           ;Toggle desnow flag
                JMP     I_30                    ;Process next switch
I_180:
                CMP     AL,"n"                  ;Check for null modem
                JNE     I_190                   ;No, process next switch
                XOR     NULL_MODEM,1            ;Toggle null modem flag
                JMP     I_30                    ;Process next switch
I_190:
                CMP     AL,"s"                  ;Check for smile face flag
                JNE     I_200                   ;No, process next switch
                XOR     SMILE_FLAG,1            ;Yes, toggle smile flag
                JMP     I_30                    ;No, check for more switches
I_200:
                CMP     AL,"u"                  ;Check for uninstall flag
                JNE     I_205                   ;No, process next switch
                JMP     UNINSTALL               ;If pcremote resident
I_205:
        if manned_is_dumb_ansi
                CMP     AL,"y"                  ;Check for systemname flag
                JNE     I_207                   ;No, process next switch
                MOV     DI,OFFSET SYSTEM_NAME
                MOV     CX,OFFSET SYSTEM_NAME_END

I_PARSE_STRING_ARG:
                MOV     BYTE PTR [DI],0
I_PARSE_STRING_MORE:
                OR      AH,AH                   ;Check for data after flag
                JZ      I_210                   ;No, so don't process
                INC     BX                      ;Point to next data
                MOV     AL,[BX]                 ;Get data
                DEC     AH                      ;Decrement counter
        if slash_delimits_strings
                CMP     AL,'/'                  ;parm must be not be slash
                JNE     I_206_NOT_SLASH
                JMP     I_60_GOT_SLASH
I_206_NOT_SLASH:
        else
                CMP     AL,SPACE                ;parm must be not be whitespace
                JNE     I_206_NOT_SPACE
                JMP     I_50
I_206_NOT_SPACE:
                CMP     AL,TAB
                JNE     I_206_NOT_TAB
                JMP     I_50
I_206_NOT_TAB:
        endif                                   ;slash_delimits_strings
                CMP     DI,CX
                JAE     I_PARSE_STRING_MORE     ;discard rest of string
                STOSB
                MOV     BYTE PTR [DI],0
                JMP     I_PARSE_STRING_MORE     ;keep looking for whitespace

I_207:
        endif                                   ;manned_is_dumb_ansi
        if ask_for_password
                CMP     AL,"p"                  ;Check for password flag
                JNE     I_208                   ;No, process next switch
                MOV     DI,OFFSET PASSWORD
                MOV     CX,OFFSET PASSWORD_END
                JMP     I_PARSE_STRING_ARG

I_208:
        endif                                   ;ask_for_password
                JMP     I_30                    ;Process next switch

I_210:
;
;  If PCREMOTE is resident and this is host mode then terminate
;
                CMP     INSTALLED_SEG,0         ;Check for resident segment
                JZ      I_215                   ;none, then ok to continue
                TEST    MANNED_FLAG,1           ;Check for manned mode
                JNZ     I_215                   ;yes, then continue
                MOV     AH,09H                  ;Print string
                MOV     DX,OFFSET PROG_RES      ;Display program resident mess.
                INT     21H
                INT     20H                     ;Terminate
;----------------------------------------------------------------------
;  Initialize the serial port
;----------------------------------------------------------------------
I_215:
;
;       get the comm port base address using 2* comm port flag as offset
;
                PUSH    DS                      ;Save data segment
                XOR     BX,BX                   ;Zero BX
                MOV     BL,COMM_FLAG            ;Get comm port flag
                MOV     SI,BX                   ;Save in index register
                SHL     SI,1                    ;Multiply by 2, word address
                MOV     AX,40H                  ;Point DS to BIOS data area
                MOV     DS,AX
                XOR     BX,BX                   ;Point to comm port address
                MOV     AX,[BX+SI]              ;Get comm port address
                POP     DS                      ;Restore data segment
                CMP     AX,0                    ;Make sure there is a entry
                JNZ     I_230
                CMP     COMM_FLAG,2             ;Is this comm 3?
                JNZ     I_220
                MOV     AX,WORD PTR COMM_PORT3  ;Yes, get the comm port 3 addr
                JMP     I_230
I_220:
                CMP     COMM_FLAG,3             ;Is this comm 4?
                JNZ     I_225                   ;No, display error message
                MOV     AX,WORD PTR COMM_PORT4  ;Yes, get the comm port 4 addr
                JMP     I_230
I_225:
                MOV     AH,09H                  ;Print string
                MOV     DX,OFFSET BAD_PORT      ;Display unknown port message
                INT     21H
                INT     20H                     ;Terminate
I_230:
                MOV     WORD PTR COMM_PORT,AX   ;Save comm port address
;
;       modify the interrupt vector for the comm port
;
                CMP     COMM_FLAG,0             ;Determine INT vector to change
                JNZ     I_240
                MOV     AL,4                    ;Comm 1 interrupt 4
I_240:
                CMP     COMM_FLAG,1
                JNZ     I_250
                MOV     AL,3                    ;Comm 2 interrupt 3
I_250:
                CMP     COMM_FLAG,2
                JNZ     I_260
                MOV     AL,COMM3_INT            ;Comm 3 interrupt
I_260:
                CMP     COMM_FLAG,3
                JNZ     I_270
                MOV     AL,COMM4_INT            ;Comm 4 interrupt
I_270:
                ADD     AL,8                    ;Interrupt number
                CLI
                MOV     BYTE PTR SET_VEC,AL     ;
                CALL    MODIFY_INTCOMM          ;Change comm interrupt vector
                CALL    INIT_SERIAL             ;Initialize the serial port
;
;       determine color or monochrome and save appropriate video segment
;
                MOV     AH,0FH                  ;Determine video mode
                INT     10H                     ;By using BIOS int 10
                CMP     AL,7                    ;Check for monochrome
                JZ      MONOCHROME                    ;Jump if it is
                MOV     WORD PTR VIDEO_SEGMENT,0B800H ;Nope, it's CGA or EGA
                JMP     I_280                         ;Skip over
MONOCHROME:
                MOV     WORD PTR VIDEO_SEGMENT,0B000H ;It's a monochrome
I_280:
;
;       If manned then execute manned initialization, otherwise execute
;       unattended initialization.
;
                TEST    MANNED_FLAG,1           ;Manned or unattended
                JZ      I_290
                JMP     INIT_MANNED             ;Init the manned routines
I_290:
                JMP     INIT_UNATTENDED         ;Init the unattended routines
INITIALIZE      ENDP

;----------------------------------------------------------------------
;  Uninstall the resident (host) portion of PCREMOTE.  Make sure another
;  program did not change any of the interrupt vectors, if so, then the
;  vectors cannot be restored properly and the program cannot be uninstalled.
;  Tom - if this looks like some of your code, it is, I borrowed from
;  PRN2FILE.
;----------------------------------------------------------------------
UNINSTALL       PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AL,08H                  ;Check the timer interrupt
                CALL    CHECK_SEG
                JNE     CANT_UNINSTALL          ;If changed, can't uninstall
                MOV     AL,21H                  ;Check DOS interrupt vector
                CALL    CHECK_SEG
                JNE     CANT_UNINSTALL          ;If changed, can't uninstall
                MOV     ES,INSTALLED_SEG        ;Get the segment of resident
                MOV     AL,BYTE PTR ES:SET_VEC  ;Check comm interrupt vector
                CALL    CHECK_SEG
                JNE     CANT_UNINSTALL          ;If changed, can't uninstall
                MOV     ES,INSTALLED_SEG
                ASSUME  DS:NOTHING, ES:NOTHING
                LDS     DX,DWORD PTR ES:OLDINT_8;Get original timer vector
                MOV     AX,2508H                ;and change it back
                INT     21H
                LDS     DX,DWORD PTR ES:OLDINT_21;Get original DOS vector
                MOV     AX,2521H                ;and change it back
                INT     21H
; comm vector changed last so PCREMOTE doesn't go to sleep
                CALL    RESET_COMM              ;Reset the comm port 8250
                LDS     DX,DWORD PTR ES:OLDINT_COMM ;Get original comm vector
                MOV     AX,ES:SET_VEC           ;and change it back
                INT     21H
; release the memory allocated to the resident portion of PCREMOTE
                MOV     ES,WORD PTR ES:2CH      ;Get the environment segment
                MOV     AH,49H                  ;Free allocated memory
                INT     21H
                JC      RELEASE_ERR
                MOV     ES,INSTALLED_SEG        ;Resident program segment
                NOT     WORD PTR ES:START       ;Modify so can't find again
                MOV     AH,49H                  ;Free allocated memory
                INT     21H
                JC      RELEASE_ERR             ;Error releasing the memory
                MOV     DX,OFFSET YES_UNINSTALL ;Display uninstalled mess
                JMP     UNI_10
CANT_UNINSTALL:
                MOV     DX,OFFSET NO_UNINSTALL  ;Display can't uninstall mess
                JMP     UNI_10
RELEASE_ERR:
                MOV     DX,OFFSET CANT_RELEASE  ;Display can't release mess
                JMP     UNI_10
UNI_10:
                PUSH    CS                      ;Restore the data segment
                POP     DS
                MOV     AH,09H                  ;Print string
                INT     21H
                MOV     AX,4C00H
                INT     21H                     ;Terminate
UNINSTALL       ENDP

;----------------------------------------------------------------------
;  Check to see if an interrupt vector points to the installed program
;  segment.
;  Input:  AL contains interrupt to check.
;  Output: Zero flag = 1 if yes, else Zero flag = 0
;  Destroys: AH, ES
;----------------------------------------------------------------------
CHECK_SEG       PROC    NEAR
                MOV     AH,35H                  ;Get vector
                INT     21H
                MOV     AX,ES                   ;Save returned segment
                CMP     AX,INSTALLED_SEG        ;Is it installed segment
                RET
CHECK_SEG       ENDP

;----------------------------------------------------------------------
;  Initialize the manned program.  Initialize buffers, and change INT 8
;  vector.
;----------------------------------------------------------------------
INIT_MANNED     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                CALL    GET_COMSPEC
                MOV     CX,OFFSET LAST_BYTE     ;CX points to begin of avail mem
        if escape_bad_chars
                CALL    INIT_ESCAPING           ;Initialize table of bad chars
        endif
                CALL    INIT_BUFFERS            ;Initalize the buffer pointers
                MOV     AX,VIDEO_SEGMENT        ;Set ES to point to video RAM
                MOV     ES,AX
                MOV     AH,3                    ;Get cursor shape
                MOV     BH,0
                INT     10H
                MOV     CURSOR_SHAPE,CX         ;Save it for when we exit
                JMP     CONNECT_MANNED          ;Run manned portion of program
INIT_MANNED     ENDP

;======================================================================
;  Data used by manned portion of program only.  Not required to be
;  resident for unattended mode.
;======================================================================
HOST_PORT       EQU     BYTE PTR CONNECT_UNATTENDED ;Host port number (1-4) DB
OLD_CURSOR      EQU     WORD PTR HOST_PORT+1    ;Cursor position save field DW
UP_LENGTH       EQU     WORD PTR OLD_CURSOR+2   ;Upload string length DW
DN_LENGTH       EQU     WORD PTR UP_LENGTH+2    ;Download string length DW
PATH_LENGTH     EQU     WORD PTR DN_LENGTH+2    ;Pathname string length DW
LAST_KEY        EQU     BYTE PTR PATH_LENGTH+2  ;Last key pressed save field DB
EXEC_VAL        EQU     BYTE PTR LAST_KEY+1     ;Shell Variable (1-3) DB
                                                ; 1 = Remote shell to DOS
                                                ; 2 = Remote Download
                                                ; 3 = Remote Upload
UPDN_STRING     EQU     BYTE PTR EXEC_VAL+1     ;Upload string (filename)
PATH_STRING     EQU     BYTE PTR UPDN_STRING+44 ;Path string (drive & path)
ORGDSK          EQU     BYTE PTR PATH_STRING+44 ;1 byte
ORGPTH          EQU     BYTE PTR ORGDSK+1       ;256 bytes
BLOCK_DATA_COUNT DB     0                       ;Number of bytes left to rec.
TYPE_TRANSFER   DB      0FFH                    ;FF - No data transfer in prog.
                                                ;00 - Received char. data
                                                ;01 - Received attr. data
SHIFT_STATUS    DB      0                       ;Current  status of shift keys
PREV_SHIFT_STS  DB      0                       ;Previous status of shift keys
RPT_STATUS      DB      0                       ;00 - No data repeat
                                                ;02 - Repeat the char in block
CUR_STATUS      DB      0                       ;0=No cursor data being rec.
                                                ; otherwise, byte count
CUR_POS         DW      0                       ;received cursor position
CUR_SHAPE       DW      0                       ;received cursor shape
    if detect_scrolling
SCRL_STATUS     DB      0                       ;0=No scroll data being rec.
                                                ; otherwise, byte count
SCRL_DEST       DW      0                       ;RX scroll destination
SCRL_SOURCE     DW      0                       ;RX scroll source
SCRL_LENGTH     DW      0                       ;RX scroll length
    endif
    if simple_dumb_terminal
DUMBTERM_ON     DB      1
DUMBTERM_MSG    DB      "Running dumb terminal.  Use clear key to stop me.",CR,LF,"$"
    endif
ENTER_NUMBER    DB      CR,LF,CR,LF,"Enter phone number:$"
NO_CARRIER      DB      CR,LF,"No carrier.$"
MODEM_ERROR     DB      CR,LF,"Error.$"
TERMINATE_MESS  DB      CR,LF,"Returning to DOS.$"
PROG_RES        DB      CR,LF,"PCREMOT2 is already resident.$"
BAD_PORT        DB      CR,LF,"BIOS does not know about that COM port.",CR,LF,"Maybe it was disabled by CMOS setup?",CR,LF,"$"
NO_UNINSTALL    DB      CR,LF,"Cannot uninstall PCREMOT2.$"
YES_UNINSTALL   DB      CR,LF,"PCREMOT2 uninstalled successfully.$"
CANT_RELEASE    DB      CR,LF,"Error release PCREMOT2 resident memory.$"
SHELL_PROMPT    DB      " Shell to DOS initiated.  Type 'EXIT' at any DOS "
                DB      "prompt to return to PCREMOT2. ",0
TRANSFER_PROMPT DB      " File transfer initiated ... ",0
PARM_$_1200     DB      "/a /8 /r",0            ;ZCOPY parm 1200 baud
PARM_$_2400     DB      "/a /7 /r",0            ;ZCOPY parm 2400 baud
PARM_$_4800     DB      "/a /6 /r",0            ;ZCOPY parm 4800 baud
PARM_$_9600     DB      "/a /5 /r",0            ;ZCOPY parm 9600 baud
PARM_$_19200    DB      "/a /4 /r",0            ;ZCOPY parm 19200 baud
PARM_$_38400    DB      "/a /3 /r",0            ;ZCOPY parm 38400 baud
PARM_$_57600    DB      "/a /2 /r",0            ;ZCOPY parm 57600 baud
PARM_$_115200   DB      "/a /1 /r",0            ;ZCOPY parm 115200 baud
PARM_STRING     DB      "         ",0   ;Actual ZCOPY parm
ASCII_CODE      DB      "~`!1@2#3$4%5^6&7*8(9)0_-+|\=QWERTYUIOP{[}]ASDFGHJKL:;"
                DB      34D,39D,CR,"ZXCVBNM<,>.?/ "
SCAN_CODE       DB      1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12
                DB      13,13,14,14,17,18,19,20,21,22,23,24,25,26,27,27,28,28
                DB      31,32,33,34,35,36,37,38,39,40,40,41,41,43,46,47,48,49
                DB      50,51,52,53,53,54,54,55,55,61
ZCOPY           DB      "ZCOPY ",0              ;ZCOPY command line
REMOTE_COM      DB      "COM ",0                ;Remote Comm value
HOST_COM        DB      "COM ",0                ;Host Comm value
UPDNLOAD_COM    DB      " /c zcopy  ", 90 DUP(?)
HELP_SCREEN     DB      CR,LF
                DB      " Usage  :  PCREMOT2 [/#] [/B#] [/M] [/D] [/S] [/N]"
        if ask_for_password
                DB      " [/Pstring]"
        endif
        if manned_is_dumb_ansi
                DB      " [/Ystring]"
        endif
                DB      CR,LF,CR,LF
        if ask_for_password
                DB      " [ /Pstring ]    sets password "
                DB      CR,LF
        endif
        if manned_is_dumb_ansi
                DB      " [ /Ystring ]    sets system name (ends up on xterm title bar) "
                DB      CR,LF
        endif
                DB      " [ /#  ]   - # = Communication port number ( 1 - 4 ) "
                DB      CR,LF
                DB      " [ /B# ]   - # = Baud rate 0=1200, 1=2400, 2=4800, 3=9600 "
                DB      CR,LF
                DB      "                 null modem only  4=19200, 5=38400, 6=57600, 7=115200 "
                DB      CR,LF
                DB      " [ /M  ]   - Manned mode (used on remote computer) "
                DB      CR,LF
                DB      " [ /D  ]   - Desnow flag (used on CGA monitors) "
                DB      CR,LF
                DB      " [ /S  ]   - Smile face disable (used on host computer) "
                DB      CR,LF
                DB      " [ /N  ]   - Null modem cable not used (must init modem) "
                DB      CR,LF
                DB      " [ /U  ]   - Uninstall (used on host computer to remove from memory) "
                DB      CR,LF,CR,LF,"$",26
EXIT_MESSAGE    DB      "                        "
                DB      "  Ŀ  "
                DB      "   OK to EXIT? (Y/N)  "
                DB      "    "
                DB      "                        "
BOX_CHAR        DW      164     ;location for character
                DB      ''     ;Character for box
                DW      256
                DB      ''
                DW      324
                DB      ''
                DW      416
                DB      ''
                DW      484
                DB      ''
                DW      576
                DB      ''
                DW      644
                DB      ''
                DW      804
                DB      ''
                DW      964
                DB      ''
                DW      1124
                DB      ''
                DW      736
                DB      ''
                DW      896
                DB      ''
                DW      1056
                DB      ''
                DW      1216
                DB      ''
                DW      1284
                DB      ''
                DW      1376
                DB      ''
                DW      1444
                DB      ''
                DW      1536
                DB      ''
                DW      1604
                DB      ''
                DW      1696
                DB      ''
TRANSFER_MESS   DB      "FILE TRANSFER REQUESTED",0
                DB      " ",0
                DB      " (U)pload - Send file(s) to Host System.",0
                DB      "                   or",0
                DB      " (D)ownload - Receive file(s) from Host.",0
UPLOAD_MESS     DB      "   UPLOAD REQUESTED",0
                DB      "Enter the full pathname of the file(s) on ",0
                DB      "the remote system which you want to upload",0
                DB      "to the host system.",0
                DB      ":",0
DOWNLOAD_MESS   DB      "  DOWNLOAD REQUESTED",0
                DB      "Enter the full pathname of the file(s) on ",0
                DB      "the host system which you want to download",0
                DB      "to the remote system.",0
                DB      ":",0
UP_PATH_MESS    DB      "   UPLOAD REQUESTED",0
                DB      "Enter the path on the host system to place ",0
                DB      "the uploaded file(s).  No target, other",0
                DB      "than directory path, may be specified.",0
                DB      ":",0
DN_PATH_MESS    DB      "  DOWNLOAD REQUESTED",0
                DB      "Enter the path on the remote system to place",0
                DB      "the downloaded file(s).  No target, other",0
                DB      "than directory path, may be specified.",0
                DB      ":",0
PRESS_ESC       DB      "Press <ESC> to return to PCREMOT2",0
;
;-----------------------------------------------------------------------
;
COMSPC          LABEL   DWORD
COM_OFS         DW      0000
COM_SEG         DW      0000

PAR_BLK         DW      0000
                DW      0000                ; OFFSET NUL_TAIL
                DW      0000
                DW      0000
                DD      0FFFFH
NOEXEC          DB      00
;--------------------------------------------------------------------
;
COM_VAR         DB      'COMSPEC='
NUL_TAIL        DB      00
;
; DON'T REMOVE OR PHYSICALLY REARRANGE THIS
;
GODPMT          DB      '$P$G',00
PMTSTR          DB      'PROMPT=',00
PMTLOC          DW      0000
DFTPMT          DB      'PROMPT='
PCRSTR          DB      'DOS SHELL- '
PMTLGT          DW      $-OFFSET PCRSTR

;======================================================================
;  The manned routine will execute the connect manned routine to call the
;  unattended system.  These routines are only used in the manned mode.
;======================================================================

;----------------------------------------------------------------------
; User requested a help screen by including a question mark in the
; command line.  Therefore, display the usage for PCREMOTE and end
; the program.
;----------------------------------------------------------------------
HELP            PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    CS                      ;
                POP     DS                      ;DS->CS
                MOV     AH,09H                  ;String print INT21 (09)
                MOV     DX,OFFSET HELP_SCREEN   ;DS:DX->Help screen variable
                INT     21H                     ;Display the help screen
                INT     20H                     ;Terminate the program
HELP            ENDP

;----------------------------------------------------------------------
;  Connect manned asks for the phone number to call, dials the number
;  and waits for connect.  Once connected it sends bursts of 20 CRs at
;  one second intervals.  When it receives alpha data it processes video
;  data and waits for sync byte (00), then transfers control to the
;  manned routine.
;----------------------------------------------------------------------
CONNECT_MANNED  PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    ES                      ;
                PUSH    CS                      ;
                POP     ES                      ;
                MOV     BX,1000H                ;
                MOV     AH,4AH                  ;
                INT     21H                     ;
                POP     ES                      ;
CM_10:
                TEST    NULL_MODEM,1            ;If null modem skip number
                JNZ     CM_50
                XOR     AL,AL                   ;Use setup string 1
                CALL    RESET_MODEM             ;Reset the modem first
                MOV     DX,OFFSET ENTER_NUMBER  ;Point to enter number mess.
                MOV     AH,9                    ;Display message
                INT     21H
                MOV     DX,OFFSET UPDN_STRING   ;Unused area as keyboard buffer
                MOV     BX,DX                   ;Enter max characters for input
                MOV     BYTE PTR [BX],40D
                MOV     AH,0AH                  ;Buffered keyboard input
                INT     21H                     ;Get the phone number
                MOV     AL,1                    ;Wait a second
                CALL    DELAY
                MOV     SI,OFFSET UPDN_STRING   ;Point to phone number
                INC     SI                      ;Amount of data in key buffer
                XOR     AX,AX                   ;Zero AX
                CLD                             ;Forward
                LODSB                           ;Get count
                OR      AL,AL                   ;Check for no input
                JNZ     CM_20                   ;If there is then continue
                JMP     EXIT_MANNED             ;No, then exit program
CM_20:
                PUSH    SI                      ;Save pointer
                MOV     SI,OFFSET TONE_DIAL     ;Send the tone dial command
                CALL    LOAD_ZSTRING
                POP     SI                      ;Restore pointer
                ADD     SI,AX                   ;Point to end of buffer
                INC     SI                      ;Include CR code
                MOV     BYTE PTR [SI],0         ;And put a zero on end
                DEC     SI                      ;Adjust for previous INC
                SUB     SI,AX                   ;Back to begin of buffer
                CALL    LOAD_ZSTRING            ;Dial the phone number
                MOV     AL,1                    ;Wait a second
                CALL    DELAY
                CALL    CLEAR_INBUFF            ;Clear input buffer
CM_30:
                CALL    GET_KEYSTROKE           ;Check for a keystroke
                JZ      CM_40                   ;None, so wait
                CMP     AX,EXIT_CODE            ;Check for exit keycode
                JNZ     CM_40                   ;No
                JMP     CM_10                   ;Yes, then enter new number
CM_40:
                CALL    CHECK_CARRIER           ;Check for a connect signal
                JZ      CM_60                   ;No, then check for err codes
CM_50:
                CALL    CLEAR_SCREEN            ;Yes, blank the screen
    if simple_dumb_terminal
                MOV     DX,OFFSET DUMBTERM_MSG  ;Display terminate call mess.
                MOV     AH,09H
                INT     21H
                MOV     DUMBTERM_ON,1           ;Remember dumb terminal mode ON
    endif
                MOV     CONNECT_FLAG,1          ;Set the connect flag
                JMP     CM_90                   ;Execute manned
CM_60:
                MOV     AH,0                    ;Use input buffer
                CALL    GET_BUFF_DATA           ;Get data from buffer
                JNC     CM_30                   ;No data, try again
                CMP     AL,"3"                  ;No carrier code?
                JNZ     CM_80                   ;No, check for other code
CM_70:
                MOV     DX,OFFSET NO_CARRIER    ;Send no carrier message
                MOV     AH,09H
                INT     21H
                JMP     CM_10                   ;Yes, reset modem and try again
CM_80:
                CMP     AL,"8"                  ;2400 no answer code
                JZ      CM_70                   ;Yes, display no carrier mess
                CMP     AL,"4"                  ;Error code?
                JNZ     CM_30                   ;No, check for other code
                MOV     DX,OFFSET MODEM_ERROR   ;Send modem command error mess
                MOV     AH,09H
                INT     21H
                JMP     CM_10                   ;Yes, reset modem and try again
CM_90:
        if baud_rate_detection
                MOV     AL,1                    ;Wait for a second
                CALL    DELAY
                MOV     CR_COUNT,30D            ;Send 30 CRs at a time
        else
CM_160:
        endif
CM_100:
                CALL    CHECK_CARRIER           ;Check for carrier loss
                JNZ     CM_110                  ;No, then continue
                JMP     EXIT_MANNED             ;Yes, then exit
CM_110:
                CALL    GET_KEYSTROKE           ;Check for a keystroke
                JZ      CM_116                  ;None, so wait
                CMP     AX,EXIT_CODE            ;Check for exit keycode
                JNZ     CM_112                  ;No
                JMP     EXIT_MANNED             ;Yes, give up
CM_112:
        if simple_dumb_terminal
                CMP     DUMBTERM_ON,0           ;still in dumb terminal mode?
                JZ      CM_118                  ;no - send sync
                CMP     AX,CLEAR_CODE           ;Want to exit terminal mode?
                JNZ     CM_114
                MOV     DUMBTERM_ON,0           ;yes - remember dumbterm off
                CALL    CLEAR_SCREEN            ;      blank the screen
                JMP     CM_118                  ;      sync this time
CM_114:
                OR      AL,AL                   ;was keystroke ASCII?
                JNZ     CM_120                  ;yes - send it instead of sync,
                                                ;  regardless whether buf empty
                JMP     CM_130                  ;no - don't send anything
        endif
CM_116:
        if simple_dumb_terminal
                CMP     DUMBTERM_ON,0           ;still in dumb terminal mode?
                JNZ     CM_130                  ;yes - don't send anything
        endif
CM_118:
                MOV     AX,OUT_BUFF_TAIL        ;Is output buffer empty?
                CMP     AX,OUT_BUFF_HEAD
                JNZ     CM_130                  ;No, skip CR
        if baud_rate_detection OR manned_is_dumb_ansi
                MOV     AL,CR                   ;Send a CR for speed sync
        else
                MOV     AL,0                    ;Send a sync byte
        endif
CM_120:
                MOV     AH,1                    ;Use output buffer
                CALL    PUT_BUFF_DATA
        if baud_rate_detection
                DEC     CR_COUNT                ;Decrement CR counter
                JZ      CM_90
        endif
CM_130:
        if ask_for_password
                MOV     AH,0                    ;Use input buffer
                CALL    GET_BUFF_DATA           ; and check for data
                JNC     CM_100                  ;No data so keep checking
                CMP     AL,'@'                  ;Alpha message?
                JL      CM_100                  ;No, then wait for it
        endif
CM_140:
        if ask_for_password
                TEST    NULL_MODEM,1            ;Using a null modem ?
                JZ      CM_150                  ;No... no need for exit check
                PUSH    AX                      ;Store AX just in case
                MOV     AH,AL                   ;Set up AX for exit compare
                XOR     AL,AL                   ;Reset AL to zero
                CMP     EXIT_CODE,AX            ;Did host send exit_code ?
                POP     AX                      ;Restore AX just in case no
                JNZ     CM_150                  ;No check for sync byte
                JMP     EXIT_MANNED             ;Yes, exit back to DOS
        endif
CM_150:
        if ask_for_password
                CMP     AL,0                    ;Is data a sync byte
                JE      CM_155                  ;  yes - continue
                JMP     CM_270                  ;  no  - put data on screen
CM_155:
        endif
        if baud_rate_detection
CM_160:
        endif
                MOV     AH,0                    ;Input buffer
                CALL    GET_BUFF_DATA           ;anything there?
                JNC     CM_160                  ;no - keep waiting
        if simple_dumb_terminal
                CMP     DUMBTERM_ON,0           ;still in dumb terminal mode?
                JZ      CM_165                  ;NO - want sync
                MOV     DL,AL                   ;Prepare to display character
                MOV     AH,2                    ; using DOS
                INT     21H                     ;Display it
                JMP     CM_160                  ;keep waiting for exit or clear
CM_165:
        endif
                CMP     AL,0                    ;Sync byte ?
                JNE     CM_NOT_SYNC
                JMP     CM_160                  ;Yes
CM_NOT_SYNC:

;
; A character was received that is not a sync byte.  This is the host trying
; to tell the remote computer what Comm port the host is currently running
; through. Because garbage (line noise) is always a threat, remote must
; accurately determine which Comm port was sent.
;
                PUSH    CX                      ;Store register
                PUSH    DX                      ;Store register
                XOR     CX,CX                   ;Reset counters to zero
                XOR     DX,DX                   ;Reset counters to zero
                JMP     CM_180                  ;Process AL data
CM_170:
                CALL    GET_KEYSTROKE           ;Check for a keystroke
                JZ      CM_175                  ;None, so wait
                CMP     AX,EXIT_CODE            ;Check for exit keycode
                JNE     CM_175                  ;  no - look for commport or sync
                JMP     EXIT_MANNED             ;  yes - exit program
CM_175:
                MOV     AH,0                    ;Use input buffer
                CALL    GET_BUFF_DATA           ;Receive a character
                JNC     CM_170                  ;Did we get one yet ?
                CMP     AL,0                    ;Sync byte ?
                JE      CM_220                  ;Yes, determine host comm
CM_180:
                CMP     AL,1D                   ;One received ?
                JNE     CM_190                  ;No
                INC     CL                      ;Yes, bump Comm 1 counter
CM_190:
                CMP     AL,2D                   ;Two received ?
                JNE     CM_200                  ;No
                INC     CH                      ;Yes, bump Comm 2 counter
CM_200:
                CMP     AL,3D                   ;Three received ?
                JNE     CM_210                  ;No
                INC     DL                      ;Yes, bump Comm 3 counter
CM_210:
                CMP     AL,4D                   ;Four received ?
                JNE     CM_215                  ;No, get another character
                INC     DH                      ;Yes, bump Comm 4 counter
                JMP     CM_170                  ;Get another character
CM_215:                                         ;Got unrecognizable char
                OR      CX,CX                   ;Have we ever got 1 or 2?
                JNZ     CM_170                  ;  yes - wait for another char
                OR      DX,DX                   ;Have we ever got 3 or 4?
                JNZ     CM_170                  ;  yes - wait for another char
                JMP     CM_160                  ;  no  - wait for a NULL again
CM_220:
                MOV     AL,'1'                  ;Assume Comm 1 for now
                CMP     CH,CL                   ;Comm2 count > Comm1 count ?
                JLE     CM_230                  ;No
                MOV     AL,'2'                  ;Yes, assume Comm2 for now
                MOV     CL,CH                   ;Save Comm2 counter in CL
CM_230:
                CMP     DL,CL                   ;Comm3 count > CL (current)
                JLE     CM_240                  ;No
                MOV     AL,'3'                  ;Yes, assume Comm3 for now
                MOV     CL,DL                   ;Save Comm3 counter in CL
CM_240:
                CMP     DH,CL                   ;Comm4 count > CL (current)
                JLE     CM_250                  ;No
                MOV     AL,'4'                  ;Yes, Host port = Comm4
CM_250:
                MOV     HOST_PORT,AL            ;Save largest port counter
                POP     DX                      ;Restore register
                POP     CX                      ;Restore register
                CALL    MAKE_PARM_STRING        ;Make ZCOPY parm w/ baud & comm
CM_260:
                XOR     AH,AH                   ;Use input buffer
                CALL    GET_BUFF_DATA           ;Receive a character
                JNC     CM_260                  ;Did we receive one yet ?
                CMP     AL,0                    ;Sync Byte ?
                JNE     CM_270                  ;Yes
                CALL    CLEAR_SCREEN            ;Blank the screen
                JMP     MANNED                  ;We're in, let's run program
CM_270:
                MOV     DL,AL                   ;Prepare to display character
                MOV     AH,2                    ; using DOS
                INT     21H                     ;Display it
CM_280:
                CALL    CHECK_CARRIER           ;Check for carrier loss
                JNZ     CM_290                  ;No, then continue
                JMP     EXIT_MANNED             ;Yes, then exit
CM_290:
                CALL    GET_KEYSTROKE           ;Check for a keystroke
                JZ      CM_310                  ;If none, skip next routine
;
;       Check for exit code, if it is then exit
;
                CMP     AX,EXIT_CODE            ;Check for exit code
                JNZ     CM_300                  ;No, so continue
                CALL    CONFIRM_EXIT            ;Yes, confirm exit
                JNZ     CM_310                  ;No, so continue, don't send
                JMP     EXIT_MANNED             ;Otherwise reset and exit
CM_300:
                CALL    SEND_KEYSTROKE          ;Send to unattended system
CM_310:
                MOV     AH,0                    ;Use input buffer
                CALL    GET_BUFF_DATA           ; and check for data
                JNC     CM_280                  ;No data, check for keys
                JMP     CM_140                  ;Check for sync byte
CONNECT_MANNED  ENDP

;----------------------------------------------------------------------
; MAKE_PARM_STRING creates the ZCOPY parameters using the current baud
; rate of the remote system. This PROC also creates REMOTE_COM using
; the COMM_FLAG.  REMOTE_COM is used in MAKE_UP_COM & MAKE_DN_COM.
;----------------------------------------------------------------------
MAKE_PARM_STRING        PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Store registers & flags
                PUSH    CX
                PUSH    DS
                PUSH    ES
                PUSH    DI
                PUSH    SI
                PUSHF
                CLD                             ;Direction flag -> forward
                PUSH    CS
                PUSH    CS
                POP     DS                      ;DS->CS
                POP     ES                      ;ES->CS
                MOV     DI,OFFSET PARM_STRING   ;DI->Parm string's first char
                MOV     SI,OFFSET PARM_$_1200   ;SI->1st char of 1200 var.
                CMP     SPEED_FLAG,1D           ;2400 Baud ?
                JNE     MPS_10                  ;Yes, use 2400 baud variable
                MOV     SI,OFFSET PARM_$_2400   ;SI->1st char of 2400 var.
MPS_10:
                CMP     SPEED_FLAG,2D           ;4800 Baud ?
                JNE     MPS_20                  ;Yes, use 4800 baud variable
                MOV     SI,OFFSET PARM_$_4800   ;SI->1st char of 4800 var.
MPS_20:
                CMP     SPEED_FLAG,3D           ;9600 Baud ?
                JNE     MPS_30
                MOV     SI,OFFSET PARM_$_9600   ;SI->1st char of 9600 var.
MPS_30:
                CMP     SPEED_FLAG,4D           ;19200 Baud ?
                JNE     MPS_32
                MOV     SI,OFFSET PARM_$_19200  ;SI->1st char of 19200 var.
MPS_32:
                CMP     SPEED_FLAG,5D           ;38400 Baud ?
                JNE     MPS_34
                MOV     SI,OFFSET PARM_$_38400  ;SI->1st char of 38400 var.
MPS_34:
                CMP     SPEED_FLAG,6D           ;57600 Baud ?
                JNE     MPS_35
                MOV     SI,OFFSET PARM_$_57600  ;SI->1st char of 57600 var.
MPS_35:
                CMP     SPEED_FLAG,7D           ;115200 Baud ?
                JNE     MPS_36
                MOV     SI,OFFSET PARM_$_115200 ;SI->1st char of 115200 var.
MPS_36:
                MOV     CX,8D                   ;Move 8 letters
MPS_40:
                LODSB                           ;Pick off char. and place it
                STOSB                           ; in the real ZCOPY parm string
                LOOP    MPS_40                  ;11 characters done ?
;
; Remote comm must be set up using COMM_FLAG
;
                MOV     DI,OFFSET REMOTE_COM    ;ES:DI-> (C)om var. 1st char
                ADD     DI,3D                   ;ES:DI-> COM( ) var. 4th char
                MOV     AL,'1'                  ;Assume remote using com1
                CMP     COMM_FLAG,1D            ;Using com2 ?
                JNE     MPS_50                  ;No
                MOV     AL,'2'                  ;Remote using com2
MPS_50:
                CMP     COMM_FLAG,2D            ;Using com3 ?
                JNE     MPS_60                  ;No
                MOV     AL,'3'                  ;Remote using com3
MPS_60:
                CMP     COMM_FLAG,3D            ;using com4?
                JNE     MPS_70                  ;No
                MOV     AL,'4'                  ;Remote using com4
MPS_70:
                STOSB                           ;Store com (1-4) variable
;
; HOST_COM must be set up using the HOST_PORT variable.
;
                MOV     AL,HOST_PORT            ;Ready AL for STOSB
                MOV     DI,OFFSET HOST_COM      ;ES:DI->1st char of HOST_COM
                ADD     DI,3D                   ;ES:DI->4th char of HOST_COM
                STOSB                           ;Store HOST_COM into variable
                POPF                            ;Restore flags & registers
                POP     SI
                POP     DI
                POP     ES
                POP     DS
                POP     CX
                POP     AX
                RET
MAKE_PARM_STRING        ENDP

;----------------------------------------------------------------------
;  Manned portion of the program.  Not RAM resident.  Sends keystrokes
;  to the unattended computer.  Data from unattended computer is
;  decoded and processed.
;----------------------------------------------------------------------
MANNED          PROC    NEAR
                ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
;
;       Check for loss of carrier
;
                CALL    CHECK_CARRIER           ;No, check for carrier loss
                JNZ     M_10                    ;No loss so continue
                JMP     EXIT_MANNED             ;Otherwise reset and exit
;
;       Check to see if shift status has changed, if so, put new status
;       in the output buffer
;
M_10:
                MOV     AH,2                    ; Prepare shift status
                INT     16H                     ; Get shift status
                CMP     SHIFT_STATUS,AL         ; See if the same
                JE      M_20                    ; If so, next routine
                XCHG    SHIFT_STATUS,AL         ; Save new status
                MOV     PREV_SHIFT_STS,AL       ; Also save old in case shelling or xferring
                MOV     AL,0FDH                 ; Sync, expect shift
                MOV     AH,1                    ; Put into out buff
                CALL    PUT_BUFF_DATA           ;
                MOV     AL,SHIFT_STATUS         ; Get shift status
                CALL    PUT_BUFF_DATA           ; Send to host system
;
;       Check for keystoke, if so, put the ASCII code and scan code
;       into the output buffer, check it for exit code
;
M_20:
                CALL    GET_KEYSTROKE           ;Check for keystroke
                JNZ     M_30
                JMP     M_90                    ;If none skip this routine
;
; Check for shell to DOS, if it is then shell & return
;
M_30:
                CMP     AX,SHELL_CODE           ;Shell key pressed ?
                JNE     M_40                    ; No....
                MOV     EXEC_VAL,1D             ;Tell EXEC to shell
                CALL    EXEC_ZCOPY              ;EXEC to shell
                CALL    CLEAR_INBUFF            ;Ignore anything received while shelled
                CALL    INIT_SERIAL             ;Re-initialize serial port
                CALL    CLEAR_SCREEN            ;Clear the remote screen
                MOV     AX,CLEAR_CODE           ;Send clear_code to
                CALL    SEND_KEYSTROKE          ; rewrite entire screen
                JMP     M_90                    ;Process next char
;
; Check for exit code, if it is then exit
;
M_40:
                CMP     AX,EXIT_CODE            ;Check for exit code
                JNZ     M_50                    ;No, so continue
                PUSH    DX                      ; Store DX
                CALL    GET_CURSOR              ; Get cursor pos.
                MOV     OLD_CURSOR,DX           ; Save Cursor pos.
                MOV     DH,LINES_PER_SCR        ; Place cursor
                MOV     DL,79D                  ;                   off screen
                CALL    PUT_CURSOR              ; Set cursor pos.
                POP     DX                      ; Restore DX
                CALL    CONFIRM_EXIT            ;Yes, confirm exit
                PUSH    DX                      ; Store DX
                MOV     DX,OLD_CURSOR           ; Restore cursor pos.
                CALL    PUT_CURSOR              ; Set cursor pos.
                POP     DX                      ; Restore DX
                JNZ     M_90                    ;No, so continue, don't send
                JMP     EXIT_MANNED             ;Otherwise reset and exit
;
; Check for transfer code, if it is then ask to upload or download
;
M_50:                                             ;
                CMP     AX,TRANSFER_CODE          ; Check for transfer
                JNZ     M_70                      ; No, so continue
                PUSH    DX                        ; Store register
                CALL    GET_CURSOR                ; Get cursor position
                MOV     OLD_CURSOR,DX             ; Save cursor in old cursor
                MOV     DH,LINES_PER_SCR          ; Cursor Row = 25
                MOV     DL,79D                    ; Cursor Col = 79
                CALL    PUT_CURSOR                ; Position Cursor
                POP     DX                        ; Restore DX register
                CALL    TRANSFER_TYPE             ; Transfer file
                CMP     AX,CLEAR_CODE             ; Transferred files ?
                JNE     M_60                      ; Yes-reset remote scrn
                CALL    INIT_SERIAL               ; Re-initialize serial port
                CALL    CLEAR_INBUFF
                CALL    CLEAR_OUTBUFF
                JMP     M_70                      ; Send clear screen code
M_60:
                PUSH    DX                        ; Push DX register
                MOV     DX,OLD_CURSOR             ; Put cursor back
                CALL    PUT_CURSOR                ; where it was
                POP     DX                        ; Restore DX register
                JMP     M_90                      ; Transfer complete
;
; Check for clear and resend entire screen code.
;
M_70:                                             ;
                CMP     AX,CLEAR_CODE             ; Clear screen pressed?
                JNZ     M_80                      ; No...
                PUSH    AX                        ; Yes.... save key
                CALL    CLEAR_SCREEN              ; Clear remote screen
                POP     AX                        ; Restore key
M_80:                                             ;
                CALL    SEND_KEYSTROKE            ; Send key to host
;
; Check for receive data, if available, get data and decode
;
M_90:
                MOV     AH,0                    ;Get data from input buffer
                CALL    GET_BUFF_DATA
                JC      M_100                   ;Data is there, so process
                JMP     M_EXIT                  ;No data,so loop again
M_100:
                CMP     TYPE_TRANSFER,0FFH      ;Check transfer type
                JZ      M_160                   ;If ff, check for sync byte
                CMP     VID_RAM_OFFSET,0FFFFH   ;Received block no. yet?
                JNZ     M_120                   ;Yes, then process as data
                CMP     AL,MAX_BLOCK_NUM        ;Avoid corrupting non-screen
                JB      M_110                   ;  RAM when line garbage occurs
                MOV     TYPE_TRANSFER,0FFH      ;Ignore & look for new sync
                JMP     M_EXIT                  ;No data,so loop again
M_110:
                MOV     BLOCK_DATA_COUNT,BLOCK_SIZE ;Data counter
                MOV     BX,BLOCK_SIZEX2         ;Get block size in bytes
                MUL     BX                      ;Block number X block size
                ADD     AL,TYPE_TRANSFER        ;Adjust for char. or attr.
                MOV     VID_RAM_OFFSET,AX       ;Save pointer to video RAM
                JMP     M_EXIT
M_120:
                MOV     DI,VID_RAM_OFFSET       ;Point to video RAM
                CMP     RPT_STATUS,0            ;Check the repeat flag
                JZ      M_140                   ;Normal data transfer
                MOV     CL,BLOCK_DATA_COUNT     ;Number to repeat
                XOR     CH,CH
M_130:
                CALL    PUT_VIDEO_DATA          ;Store the data in video RAM
                INC     DI                      ;Point to next video RAM loc.
                LOOP    M_130
                JMP     M_150                   ;End of data, reset variables
M_140:
                CALL    PUT_VIDEO_DATA          ;Store the data in video RAM
                INC     DI                      ;Point to next video RAM loc.
                MOV     VID_RAM_OFFSET,DI       ;And save pointer
                DEC     BLOCK_DATA_COUNT        ;Dec. data counter
                JZ      M_150                   ;End of data, reset variables
                JMP     M_EXIT                  ;Otherwise continue
M_150:
                MOV     TYPE_TRANSFER,0FFH      ;Reset type transfer flag
                JMP     M_EXIT
M_160:
                CMP     CUR_STATUS,0            ;Check cursor status
                JZ      M_170                   ;If 0, check for sync byte
                XOR     BX,BX                   ;Zero BX for index
                MOV     BL,CUR_STATUS           ;Use as index for cursor data
                MOV     DI,OFFSET CUR_STATUS    ;Base address for cursor data
                MOV     [DI+BX],AL              ;Save the register data
                INC     BL                      ;Increment count
                MOV     CUR_STATUS,BL           ;And save it
                CMP     BL,5                    ;Check to see if we have 5 bytes yet
                JZ      M_165
                JMP     M_EXIT                  ;No, so wait till enough data
M_165:
                MOV     BH,0                    ;Always use page one
                MOV     DX,CUR_POS              ;Get cursor position
                MOV     AH,2                    ;Set cursor position function
                INT     10H
                MOV     CX,CUR_SHAPE            ;Get cursor shape
                MOV     AH,1                    ;Set cursor shape function
                INT     10H
                MOV     CUR_STATUS,0            ;Reset cursor status
                JMP     M_EXIT
M_170:
        if detect_scrolling
                CMP     SCRL_STATUS,0           ;Waiting for scroll count?
                JZ      M_180                   ;If 0, check for sync byte
                XOR     BX,BX                   ;Zero BX for index
                MOV     BL,SCRL_STATUS          ;Use as index for scroll data
                MOV     DI,OFFSET SCRL_STATUS   ;Base address for scroll data
                MOV     [DI+BX],AL              ;Save the register data
                INC     BL                      ;Increment count
                MOV     SCRL_STATUS,BL          ;And save it
                CMP     BL,7                    ;Do we have all bytes yet?
                JNZ     M_EXIT                  ;No, so wait till enough data
;
; scroll up SCRL_LENGTH bytes of video memory
; from SCRL_SOURCE to SCRL_DEST
;
                PUSH    ES
                PUSH    DS
                MOV     DI,SCRL_DEST            ;dest = maybe top of screen
                MOV     SI,SCRL_SOURCE          ;src = a few lines down
                MOV     CX,SCRL_LENGTH          ;length from packet
                SHR     CX,1                    ;convert bytes to chars(words)
                MOV     AX,VIDEO_SEGMENT        ;Get video Segment register
                MOV     DS,AX                   ;Source DS:SI video RAM
                MOV     ES,AX
                CLD                             ;pointers should increment
                REP     MOVSW
                POP     DS
        if scroll_clears_lines
;
; fill the screen hole we just opened up with blanks
;
; note: ES:DI already points at left of first line to blank
;
                MOV     AX,720H                 ;AL=SPACE,AH=Normal attribute
                MOV     CX,SCRL_DEST            ;As many bytes as were scrolled
                SUB     CX,SCRL_SOURCE
                SHR     CX,1                    ;Convert bytes to chars (words)
                REP     STOSW                   ;Store char and attribute
        endif

                POP     ES
                MOV     SCRL_STATUS,0           ;Reset scroll status
                JMP     M_EXIT
        endif
M_180:
                PUSH    AX                      ;Save the data
                AND     AL,0FCH                 ;Mask out data type and repeat
                CMP     AL,0FCH                 ;Video data sync?
                JNZ     M_190                   ;No, check for cursor sync
                POP     AX                      ;Restore data
                PUSH    AX
                AND     AL,1                    ;Save data type
                MOV     TYPE_TRANSFER,AL        ;Set transfer type flag
                POP     AX                      ;Restore data
                AND     AL,2                    ;Save repeat flag
                MOV     RPT_STATUS,AL           ;Set repeat flag
                MOV     VID_RAM_OFFSET,0FFFFH   ;Prepare to receive block no.
                JMP     M_EXIT
M_190:
                POP     AX                      ;Restore data
                CMP     AL,0D2H                 ;Check for cursor data sync
                JNZ     M_200                   ;If not, then try scrolling
                MOV     CUR_STATUS,1            ;Set cursor status byte
                JMP     M_EXIT
M_200:
        if detect_scrolling
                CMP     AL,0E1H                 ;Check for scroll with dest
                JNZ     M_210                   ;If not, then try scroll w/o dest
                MOV     SCRL_STATUS,1           ;Remember we want dst,src,len
                JMP     M_EXIT
M_210:
                CMP     AL,0E0H                 ;Check for scroll without dest
                JNZ     M_EXIT                  ;If not, then throw it away
                MOV     SCRL_STATUS,3           ;Remember we want src,len
                MOV     SCRL_DEST,0
        endif
M_EXIT:
                JMP     MANNED                  ;Do it again
MANNED          ENDP

;----------------------------------------------------------------------
;  Exit manned.  The exit code is sent, interrupt vector reset, and
;  the modem is hung up and reset
;  Input - Nothing
;  Output - DOS
;  Changes - everything
;----------------------------------------------------------------------
EXIT_MANNED     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AX,EXIT_CODE            ;Get the exit code
                CALL    SEND_KEYSTROKE          ;Send the exit code
                CALL    CLEAR_SCREEN             ;Blank the screen
                MOV     DX,OFFSET TERMINATE_MESS ;Display terminate call mess.
                MOV     AH,09H
                INT     21H
                MOV     CX,CURSOR_SHAPE         ;Restore cursor shape
                MOV     AH,1
                INT     10H
                XOR     AL,AL                   ;Reset modem and exit
                CALL    RESET_MODEM
                MOV     AL,1                    ;Wait a second
                CALL    DELAY
                CALL    RESET_COMM
                LDS     DX,DWORD PTR OLDINT_COMM ;Get original comm vector
                MOV     AX,SET_VEC              ;and change it back
                INT     21H
                INT     20H                     ;Terminate program
EXIT_MANNED     ENDP

;----------------------------------------------------------------------
;  Clear screen.  This routine will blank the video display
;  Input - Nothing
;  Output - Screen is cleared
;  Changes - AX, BX, CX, DX
;----------------------------------------------------------------------
CLEAR_SCREEN    PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     BX,0                    ;Use video page one
                MOV     DX,0                    ;Position cursor at row 0 col 0
                MOV     AH,2                    ;Set cursor function
                INT     10H
                MOV     AX,0700H                ;Scroll down, clear screen
                MOV     BH,07H                  ;White on black
                MOV     CX,0                    ;Upper left corner
                MOV     DH,LAST_SCR_LINE        ;Lower right corner
                MOV     DL,79D
                INT     10H                     ;Video BIOS call
                RET
CLEAR_SCREEN    ENDP

;----------------------------------------------------------------------
;  Display OK to exit message and wait for response
;  Input - Nothing
;  Output - Zero flag set - exit, zero flag reset - do not exit
;  Changes - CX, SI, DI
;----------------------------------------------------------------------
CONFIRM_EXIT    PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save the exit code
CE_10:
                MOV     AX,OUT_BUFF_HEAD        ;Wait until the out buffer
                CMP     AX,OUT_BUFF_TAIL        ; is empty
                JNZ     CE_10
;
;  Save the existing video data in the output buffer
;
                CALL    SWITCH_DS_ES
                MOV     DI,ES:OUT_BUFF_END      ;Use output buffer as temp stor.
                MOV     SI,0                    ;Upper left corner of video
                MOV     CX,5                    ;Transfer 5 lines
CE_20:
                PUSH    CX                      ;Save line counter
                MOV     CX,24D                  ;Transfer 24 char. per line
                REPNZ   MOVSW
                ADD     SI,112D                 ;Next line
                POP     CX                      ;Restore line counter
                LOOP    CE_20
;
;  Display the ok to exit message
;
                CALL    SWITCH_DS_ES
                MOV     DI,0                    ;Upper left corner of video
                MOV     SI,OFFSET EXIT_MESSAGE  ;Exit message
                MOV     CX,5                    ;Display 5 lines
CE_30:
                PUSH    CX                      ;Save line counter
                MOV     CX,24D                  ;Display 20 characters per line
CE_40:
                LODSB
                CALL    PUT_VIDEO_DATA          ;Display the character
                MOV     AL,70H                  ;White on black attribute
                CALL    PUT_VIDEO_DATA          ;Store attribute
                LOOP    CE_40
                ADD     DI,112D                 ;Next line
                POP     CX                      ;Restore line counter
                LOOP    CE_30
;
;  Wait for keystroke and convert to lower case
;
CE_50:
                CALL    GET_KEYSTROKE           ;Check for keystroke
                JZ      CE_50                   ;None there, try again
                OR      AL,20H                  ;Convert to lower case
;
;  Restore the saved video data
;
                MOV     SI,OUT_BUFF_END         ;Old video is in output buffer
                MOV     DI,0                    ;Upper left corner of video
                MOV     CX,5                    ;Transfer 5 lines
CE_60:
                PUSH    CX                      ;Save line counter
                MOV     CX,24D                  ;Transfer 24 char. per line
                REPNZ   MOVSW
                ADD     DI,112D                 ;Next line
                POP     CX                      ;Restore line counter
                LOOP    CE_60
                CMP     AL,'y'                  ;Check for yes, all others no
                POP     AX                      ;Restore exit code
                RET
CONFIRM_EXIT    ENDP

;---------------------------------------------------------------
; Draw the box and display the text in it.
; Input - SI points to the text strings
; Output - Current video stored in output buffer, text displayed
;          in box.
; Changes - nothing
;---------------------------------------------------------------
DISPLAY_BOX     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save registers
                PUSH    BX
                PUSH    CX
                PUSH    DI
                PUSH    SI
;
; Save the current video area in the output buffer
;
                CALL    SWITCH_DS_ES
                MOV     DI,ES:OUT_BUFF_END      ;Use output buffer as temp stor.
                XOR     SI,SI                   ;Upper left of video
                MOV     CX,12                   ;Save 12 lines
DB_10:
                PUSH    CX                      ;Save the line counter
                MOV     CX,51                   ;Save 51 columns
                REPNZ   MOVSW
                ADD     SI,58                   ;Adjust to start of next line
                POP     CX                      ;Restore line counter
                LOOP    DB_10
                CALL    SWITCH_DS_ES
;
; Blank that area of the screen
;
                XOR     DI,DI                   ;Upper left of video
                MOV     CX,12                   ;Blank 12 lines
DB_20:
                PUSH    CX                      ;Save line counter
                MOV     CX,51                   ;Blank 51 characters
DB_30:
                MOV     AL,SPACE                ;Space character
                CALL    PUT_VIDEO_DATA          ;Display the character
                MOV     AL,70H                  ;Attribute
                CALL    PUT_VIDEO_DATA
                LOOP    DB_30
                ADD     DI,58                   ;Adjust to start of next line
                POP     CX                      ;Restore line counter
                LOOP    DB_20
;
; Draw the box
;
                MOV     DI,166                  ;First line of box
                CALL    DRAW_LINE               ;Display the horizontal line
                MOV     DI,486                  ;Second line of box
                CALL    DRAW_LINE               ;Display the horizontal line
                MOV     DI,1286                 ;Third line of box
                CALL    DRAW_LINE               ;Display the horizontal line
                MOV     DI,1606                 ;Fourth line of box
                CALL    DRAW_LINE               ;Display the horizontal line
                MOV     BX,OFFSET BOX_CHAR      ;Get a pointer to box corners
                MOV     CX,20                   ;Corners and vertical lines
DB_40:
                MOV     DI,[BX]                 ;Get offset to char location
                ADD     BX,2                    ;Point to character
                MOV     AL,[BX]                 ;Get the character
                INC     BX                      ;Next corner
                CALL    PUT_VIDEO_DATA
                LOOP    DB_40
;
; Draw the text in the box
;
                POP     SI                      ;Get pointer to text
                PUSH    SI
                MOV     DI,348                  ;Location for title line
                CALL    DRAW_TEXT               ;Display the line of text
                MOV     DI,648                  ;Next line of text
                CALL    DRAW_TEXT               ;Display the line of text
                MOV     DI,808                  ;Next line of text
                CALL    DRAW_TEXT               ;Display the line of text
                MOV     DI,968                  ;Next line of text
                CALL    DRAW_TEXT               ;Display the line of text
                MOV     DI,1128                 ;Next line of text
                CALL    DRAW_TEXT               ;Display the line of text
                MOV     DI,1458                 ;Next line of text
                MOV     SI,OFFSET PRESS_ESC     ;Display the press ESC to exit
                CALL    DRAW_TEXT               ;Display the line of text
                POP     SI                      ;Restore registers
                POP     DI
                POP     CX
                POP     BX
                POP     AX
                RET
DISPLAY_BOX     ENDP

;---------------------------------------------------------------
; Draw horizontal line
; Input ES:DI points to location to start drawing the line
; Output - nothing
; Changes - AL, CX, DI
;---------------------------------------------------------------
DRAW_LINE       PROC    NEAR
                MOV     CX,45                   ;45 Columns in line
                MOV     AL,''                  ;Line draw character
DL_10:
                CALL    PUT_VIDEO_DATA          ;Display the character
                INC     DI                      ;Adjust past attribute
                LOOP    DL_10
                RET
DRAW_LINE       ENDP

;---------------------------------------------------------------
; Display line of text
; Input - SI points to text DI points to video location
; Output - line of text is displayed on the screen
; Changes - AL, SI, DI
;---------------------------------------------------------------
DRAW_TEXT       PROC    NEAR
                LODSB                           ;Get the text
                CMP     AL,0                    ;Is this the end of the string
                JZ      DT_10                   ;Yes
                CALL    PUT_VIDEO_DATA          ;Display the character
                INC     DI
                JMP     DRAW_TEXT
DT_10:
                RET
DRAW_TEXT       ENDP

;---------------------------------------------------------------
; Restore the video saved to display a box
; Input - nothing
; Output - nothing
; Changes - nothing
;---------------------------------------------------------------
RESTORE_BOX     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save registers
                PUSH    CX
                PUSH    DI
                PUSH    SI
                MOV     SI,OUT_BUFF_END         ;Old video in output buffer
                XOR     DI,DI                   ;Upper left corner of screen
                MOV     CX,12                   ;Line counter
RB_10:
                PUSH    CX                      ;Save line counter
                MOV     CX,51                   ;51 characters per line
                REPNZ   MOVSW
                ADD     DI,58                   ;Next line
                POP     CX                      ;Restore line counter
                LOOP    RB_10
                POP     SI                      ;Restore registers
                POP     DI
                POP     CX
                POP     AX
                RET
RESTORE_BOX     ENDP

;---------------------------------------------------------------
; Display (U)pload or (D)ownload message and wait for response
; Input - Nothing
; Output -
; Changes - CX, SI, DI
;---------------------------------------------------------------
TRANSFER_TYPE   PROC    NEAR
       ASSUME   CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AX,OUT_BUFF_HEAD        ;Wait until the out
                CMP     AX,OUT_BUFF_TAIL        ;buffer is empty
                JNZ     TRANSFER_TYPE
;
; Display the (U)pload or (D)ownload message
;
                MOV     SI,OFFSET TRANSFER_MESS ;Transfer message
                CALL    DISPLAY_BOX
;
; Wait for a keystroke and convert to lower case
;
TT_10:
                CALL    GET_KEYSTROKE           ;Check for a keystroke
                JZ      TT_10                   ;None there, try again
                OR      AL,20H
                CALL    RESTORE_BOX
                CMP     AL,'u'                  ;Upload Requested ?
                JZ      TT_20                   ;Yes
                JMP     TT_90                   ;No, so continue
TT_20:                                          ;UPLOAD CODING HERE
                MOV     LAST_KEY,0              ;Clear last key pressed
                CALL    UPLOAD                  ;Ask for upload file
                CMP     LAST_KEY,ESC_KEY        ;ESC pressed ?
                JNZ     TT_30                   ;No
                JMP     TT_180                  ;Yes - return to PCREMOTE
TT_30:
                MOV     LAST_KEY,0              ;Clear last key pressed
                MOV     SI,OFFSET UP_PATH_MESS  ; Path message for upload
                CALL    PATH                    ;Get destination Path
                CMP     LAST_KEY,ESC_KEY        ;ESC pressed ?
                JNZ     TT_40                   ;No
                JMP     TT_180                  ;Yes - return to PCREMOTE
TT_40:
                MOV     LAST_KEY,0              ;Clear last key pressed
                CALL    MAKE_UP_COM             ;Create ZCOPY EXEC com
                CLD                             ;Forward direction ptr.
                MOV     SI,OFFSET ZCOPY         ;DS:SI -- ZCOPY variable
                CALL    SEND_STRING             ;Send ZCOPY command line
                MOV     SI,OFFSET HOST_COM      ;DS:SI -- COM variable
                CALL    SEND_STRING             ;Send COM to host
                MOV     AL,SPACE                ;AL = SPACE
                CALL    GET_SEND_CODE           ;Get & send scan-code
                MOV     SI,OFFSET PATH_STRING   ;DS:SI -- Path variable
                MOV     CX,PATH_LENGTH          ;CX = Length of PATH
                CMP     CX,0                    ;Default Path ?
                JE      TT_60                   ;Yes...
TT_50:                                          ;No...
                LODSB                           ;AL = next character
                CALL    GET_SEND_CODE           ;Get & send scan-code
                LOOP    TT_50                   ;All characters sent ?
                MOV     AL,SPACE                ;AL = SPACE
                CALL    GET_SEND_CODE           ;Get & send scan-code
TT_60:
                MOV     SI,OFFSET PARM_STRING   ;DS:SI -- PARM variable
                MOV     CX,9D
TT_70:
                LODSB                           ;AL = Next character
                CALL    GET_SEND_CODE           ;Get & send scan-code
                LOOP    TT_70                   ;All characters sent ?
                MOV     AL,CR                   ;AL = carriage return
                CALL    GET_SEND_CODE           ;Get & send scan-code
TT_80:
                MOV     AX,OUT_BUFF_HEAD        ;Get the buff head ptr
                CMP     AX,OUT_BUFF_TAIL        ;Test for data
                JNZ     TT_80                   ;Wait until buffer empty
                MOV     EXEC_VAL,3D             ;Tell EXEC to upload
                CALL    EXEC_ZCOPY              ;Shell to DOS & EXEC ZCOPY
                JMP     TT_170                  ;Transfer complete, exit
TT_90:
                CMP     AL,'d'                  ;Download Requested ?
                JZ      TT_100                  ;Yes
                JMP     TT_180                  ;No so return to PCREMOTE
TT_100:                                         ;DOWNLOAD CODING HERE
                MOV     LAST_KEY,0              ;Clear last key pressed
                CALL    DOWNLOAD                ;Ask for filename
                CMP     LAST_KEY,ESC_KEY        ;ESC pressed ?
                JNZ     TT_110                  ;No
                JMP     TT_180                  ;Yes, return to PC_REMOTE
TT_110:
                MOV     LAST_KEY,0              ;Clear last key pressed
                MOV     SI,OFFSET DN_PATH_MESS  ; Path message for download
                CALL    PATH                    ;Get destination Path
                CMP     LAST_KEY,ESC_KEY        ;ESC pressed ?
                JNZ     TT_120                  ;No
                JMP     TT_180                  ;Yes, return to PC_REMOTE
TT_120:
                MOV     LAST_KEY,0              ;Clear last key pressed
                CALL    MAKE_DN_COM             ;Create ZCOPY EXEC com
                CLD                             ;Forward direction flag
                MOV     SI,OFFSET ZCOPY         ;DS:SI -- ZCOPY variable
                CALL    SEND_STRING             ;Send ZCOPY command
                MOV     SI,OFFSET UPDN_STRING   ;DS:SI -- file variable
                MOV     CX,DN_LENGTH            ;CX = length of filename
                CMP     CX,0                    ;Default Filename
                JE      TT_140                  ;Yes...
TT_130:                                         ;No...
                LODSB                           ;AL = next character
                CALL    GET_SEND_CODE           ;Get & send scan-code
                LOOP    TT_130                  ;All characters sent ?
                MOV     AL,SPACE                ;AL = SPACE
                CALL    GET_SEND_CODE           ;Get & send scan-code
TT_140:
                MOV     SI,OFFSET HOST_COM      ;DS:SI -- Com variable
                CALL    SEND_STRING             ;Send COM variable
                MOV     AL,SPACE                ;AL = SPACE
                CALL    GET_SEND_CODE           ;Get & send scan-code
                MOV     SI,OFFSET PARM_STRING   ;DS:SI -- PARM variable
                MOV     CX,9D
TT_150:
                LODSB                           ;AL = next character
                CALL    GET_SEND_CODE           ;Get & send scan-code
                LOOP    TT_150                  ;All characters sent ?
                MOV     AL,CR                   ;AL = carriage return
                CALL    GET_SEND_CODE           ;Get & send scan-code
TT_160:
                MOV     AX,OUT_BUFF_HEAD        ;Get buffer head ptr
                CMP     AX,OUT_BUFF_TAIL        ;Test for data in buff
                JNZ     TT_160                  ;Wait for buffer empty
                MOV     EXEC_VAL,2D             ;Tell EXEC to download
                CALL    EXEC_ZCOPY              ;Shell to DOS - EXEC ZCOPY
                JMP     TT_170
TT_170:
                MOV     AX,CLEAR_CODE           ;Tell the caller to clear
                JMP     TT_190                  ;the remote's screen
TT_180:
                MOV     AX,TRANSFER_CODE ;Tell the caller that
TT_190:                                         ;screen doesn't need to
                RET                             ;be rewritten.
TRANSFER_TYPE   ENDP

;------------------------------------------------------------
; Display upload message then wait for filename(s).
; Input - Nothing
; Output -
; Changes - CX,SI,DI
;
;------------------------------------------------------------
UPLOAD          PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AX,OUT_BUFF_HEAD        ;Wait until the out
                CMP     AX,OUT_BUFF_TAIL        ;buffer is empty
                JNZ     UPLOAD
                MOV     SI,OFFSET UPLOAD_MESS   ;Upload message
                CALL    DISPLAY_BOX
                PUSH    DX                      ;Save register
                MOV     DH,7                    ;Put cursor in row 8
                MOV     DL,5                    ;Put cursor in column 6
                CALL    PUT_CURSOR              ;Set Cursor
                POP     DX                      ;Restore register
                PUSH    ES                      ;Store ES
                PUSH    DS                      ;Store DS
                POP     ES                      ;ES = DS
                MOV     DI,OFFSET UPDN_STRING   ;ES:DI -> UP_string
                MOV     CX,42D                  ;MAX length = 42
                CALL    GETSTRING               ;Receive file
                MOV     LAST_KEY,AL             ;Save last key
                MOV     UP_LENGTH,CX            ;Save length
                CMP     UP_LENGTH,0             ;Any file selected ?
                JNE     UP_10                   ;Yes
                MOV     LAST_KEY,ESC_KEY        ;No...Return (esc)
UP_10:
                POP     ES                      ;Restore ES
                CALL    RESTORE_BOX
                RET
UPLOAD          ENDP

;------------------------------------------------------------
; Display download message then wait for response.
; Input - Nothing
; Output -
; Changes - CX,SI,DI
;------------------------------------------------------------
DOWNLOAD        PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AX,OUT_BUFF_HEAD        ;Wait until the
                CMP     AX,OUT_BUFF_TAIL        ;out buffer is empty
                JNZ     DOWNLOAD
                MOV     SI,OFFSET DOWNLOAD_MESS ;Download message
                CALL    DISPLAY_BOX
                PUSH    DX                      ;Save DX register
                MOV     DH,7                    ;Put cursor in row 8
                MOV     DL,5                    ;Put cursor in column 6
                CALL    PUT_CURSOR              ;Put cursor there
                POP     DX                      ;Restore register
                PUSH    ES                      ;Store ES
                PUSH    DS                      ;Store DS
                POP     ES                      ;ES = DS
                MOV     DI,OFFSET UPDN_STRING   ;ES:DI -> DN_string
                MOV     CX,42D                  ;MAX length = 42
                CALL    GETSTRING               ;Get filename
                MOV     LAST_KEY,AL             ;Save last key
                MOV     DN_LENGTH,CX            ;Save length
                CMP     DN_LENGTH,0             ;File selected ?
                JNE     DN_10                   ;Yes
                MOV     LAST_KEY,ESC_KEY        ;No... return (esc)
DN_10:
                POP     ES                      ;Restore ES
                CALL    RESTORE_BOX
                RET
DOWNLOAD        ENDP

;----------------------------------------------------------------------
; Display the path request screen & prompt user to input path name.
; Input - SI points to display message for transfer
; Output - Path Message Displayed on screen.
; Changes - CX,SI,DI
;----------------------------------------------------------------------
PATH            PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                MOV     AX,OUT_BUFF_HEAD        ; Wait until the
                CMP     AX,OUT_BUFF_TAIL        ; out buffer is empty
                JNZ     PATH
                CALL    DISPLAY_BOX
                PUSH    DX                      ; Save DX Register
                MOV     DH,7                    ; Put cursor in row 8
                MOV     DL,5                    ; Put cursor in col 6
                CALL    PUT_CURSOR              ; Put cursor there
                POP     DX                      ; Restore DX
                PUSH    ES                      ; Store ES
                PUSH    DS                      ; Store DS
                POP     ES                      ; ES = DS
                MOV     DI,OFFSET PATH_STRING   ; ES:DI -> Path_string
                MOV     CX,42D                  ; Max length = 42
                CALL    GETSTRING               ; Get path
                MOV     LAST_KEY,AL             ; Save last key
                MOV     PATH_LENGTH,CX          ; Save length
;
; Validate and correct (modify) pathname if ZCOPY will not accept the
; string entered for the pathname.
; i.e.- C:\ is acceptable whereas C: is unacceptable
; C:\DOS is acceptable whereas C:\DOS\ is unacceptable.
;
                CMP     LAST_KEY,ESC_KEY        ; ESC key pressed ?
                JE      PATH_20                 ; Yes, so skip pathname edit
                MOV     SI,OFFSET PATH_STRING   ; SI->1st char of path name
                ADD     SI,PATH_LENGTH          ; SI->1 char after pathname
                DEC     SI                      ; SI->Last char of pathname
                LODSB                           ; Pick off last char
                CMP     AL,':'                  ; Is last char a colon
                JNE     PATH_10                 ; No, check other chars.
                PUSH    SI                      ; Save SI into DI
                POP     DI                      ; DI = SI
                MOV     AL,'\'                  ; Add backslash to pathname
                STOSB                           ; Store the backslash
                INC     PATH_LENGTH             ; Add one to length of path
                JMP     PATH_20                 ; Editting is complete
PATH_10:
                CMP     AL,'\'                  ; Is last char a backslash
                JNE     PATH_20                 ; No, skip pathname edit
                DEC     SI                      ; SI -> Last char again
                DEC     SI                      ; SI -> 2nd last char
                LODSB                           ; Pick off 2nd last char
                CMP     AL,':'                  ; 2nd last char a colon ?
                JE      PATH_20                 ; Yes, so skip edit.
                PUSH    SI                      ; Store SI for DI
                POP     DI                      ; DI-> Last char of pathname
                MOV     AL,SPACE                ; Clear last char of path
                STOSB                           ; Store blank
                DEC     PATH_LENGTH             ; Subtract 1 from path length
PATH_20:
                POP     ES
                CALL    RESTORE_BOX
                RET
PATH            ENDP

;----------------------------------------------------------------------
; Create the Upload command string that the REMOTE computer EXECS
;----------------------------------------------------------------------
MAKE_UP_COM     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Store registers
                PUSH    CX
                PUSH    DS
                PUSH    ES
                PUSH    DI
                PUSH    SI
                PUSHF
                PUSH    CS
                PUSH    CS
                POP     DS                      ;DS -> CS
                POP     ES                      ;ES -> CS
                CLD                             ;Forward Direction Flag
                MOV     DI,OFFSET UPDNLOAD_COM  ;DI -> 1st char of upload com
                ADD     DI,10D                  ;DI -> 10th char of upl. com
                MOV     AL,SPACE                ;Reset com string w/ 89
                MOV     CX,89D                  ; blanks
MUC_10:
                STOSB                           ;Fill sting with AL
                LOOP    MUC_10                  ;89 chars complete ?
                SUB     DI,89D                  ;DI -> 10th char of upl. com
                MOV     SI,OFFSET UPDN_STRING   ;SI -> 1st char of upl. file
                MOV     CX,UP_LENGTH            ;CX = length of string
MUC_20:
                LODSB                           ;AL = char from filename
                STOSB                           ;Upl. com gets char
                LOOP    MUC_20                  ;All chars copied ?
                MOV     AL,SPACE                ;Place a blank char at the
                STOSB                           ; end of filename
                MOV     SI,OFFSET REMOTE_COM    ;SI -> 1st char of comm var.
                MOV     CX,4D                   ;Pick off 4 characters
MUC_30:
                LODSB                           ;Pick off 1 char
                STOSB                           ;Place in upl. com string
                LOOP    MUC_30                  ;Entire string done ?
                MOV     AL,SPACE                ;Place a blank at end of comm
                STOSB                           ; string in upl. com string
                MOV     SI,OFFSET PARM_STRING   ;SI -> 1st char of parm
                ADD     SI,3D                   ;SI -> 4th char of parm
                MOV     CX,5D                   ;Pick off 5 chars from parm
MUC_40:
                LODSB                           ;Pick off 1 char
                STOSB                           ;Place char in upl. com
                LOOP    MUC_40                  ;6 characters done yet ?
MUC_50:
                MOV     AL,CR                   ;Carriage return
                STOSB
                MOV     AL,0                    ;0 to end string
                STOSB
;
; Length of UPDNLOAD_COM must be inserted in the first character of the
; string ("?/c zcopy..."), therefore the length must be determined.
;
                MOV     SI,OFFSET UPDNLOAD_COM  ;SI-> 1st char of string
                CALL    GET_STRING_LEN          ;CX = Length of string
                MOV     AL,CL                   ;Ready AL for STOSB instruct.
                MOV     DI,OFFSET UPDNLOAD_COM  ;DI-> 1st char of string
                STOSB                           ;Store AL in string
MUC_RET:
                POPF                            ;Restore registers & flags
                POP     SI
                POP     DI
                POP     ES
                POP     DS
                POP     CX
                POP     AX
                RET
MAKE_UP_COM     ENDP

;----------------------------------------------------------------------
; Creates the download command string that the REMOTE computer EXECS
;----------------------------------------------------------------------
MAKE_DN_COM     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Store registers
                PUSH    CX
                PUSH    DS
                PUSH    ES
                PUSH    DI
                PUSH    SI
                PUSHF
                PUSH    CS
                PUSH    CS
                POP     DS                      ;DS -> CS
                POP     ES                      ;ES -> CS
                CLD                             ;Forward direction flag
                MOV     DI,OFFSET UPDNLOAD_COM  ;DI -> Download command string
                ADD     DI,10D                  ;DI -> 11th char of command
                MOV     AL,SPACE                ;Reset command w/ blanks
                MOV     CX,89D                  ; 89 of them...
MDC_10:
                STOSB                           ;Store the blank in command
                LOOP    MDC_10                  ;56 chars. complete ?
                SUB     DI,89D                  ;Reset DI to 11th char.
                MOV     SI,OFFSET REMOTE_COM    ;SI -> (C)OMx
                MOV     CX,4D                   ;Transfer COMx to com. line
MDC_20:
                LODSB                           ;Pick off a char into AL
                STOSB                           ;Move AL into command
                LOOP    MDC_20                  ;All 4 chars copied ?
                MOV     AL,SPACE                ;Move a blank in after COMx
                STOSB                           ;Store blank
                CMP     PATH_LENGTH,0           ;Does target directory exist?
                JE      MDC_40                  ;No, skip target dir. copy
                MOV     SI,OFFSET PATH_STRING   ;Yes, SI -> 1st char of dir.
                MOV     CX,PATH_LENGTH          ;CX = length of path string
MDC_30:
                LODSB                           ;Pick off one char
                STOSB                           ;Store char in command line
                LOOP    MDC_30                  ;All chars. copied ?
                MOV     AL,SPACE                ;Move in a blank after dir.
                STOSB                           ;Store blank in line
MDC_40:
                MOV     SI,OFFSET PARM_STRING   ;SI -> 1st char of parm
                ADD     SI,3D                   ;SI -> 4th char of parm
                MOV     CX,5D                   ;Move in last 5 characters
MDC_50:
                LODSB                           ;Pick off one char.
                STOSB                           ;Copy char in command
                LOOP    MDC_50                  ;All 6 chars copied ?
MDC_60:
                MOV     AL,CR                   ;Carriage return
                STOSB
                MOV     AL,0                    ;End string with a 0
                STOSB
;
; Length of download_com must be inserted in the first character of the
; string ("?/c zcopy..."), therefore the length must be determined.
;
                MOV     SI,OFFSET UPDNLOAD_COM  ;SI-> 1st char of string
                CALL    GET_STRING_LEN          ;CX = Length of string
                MOV     AL,CL                   ;Ready AL for STOSB instruct.
                MOV     DI,OFFSET UPDNLOAD_COM  ;DI-> 1st char of string
                STOSB                           ;Store al in string
MDC_RET:
                POPF                            ;Restore registers
                POP     SI
                POP     DI
                POP     ES
                POP     DS
                POP     CX
                POP     AX
                RET
MAKE_DN_COM     ENDP

;----------------------------------------------------------------------
; Get string length determines the length of the string that DS:SI is
; pointing to.  The length is returned in the CX register.
; Input - DS:SI points to the first character of the associated string.
; Output - CX = Length of the string.
; Changes - CX
;----------------------------------------------------------------------
GET_STRING_LEN  PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Store registers
                PUSH    SI
                PUSHF
                CLD                             ;Go forward in variables
                XOR     CX,CX                   ;Reset CX (counter) to zero
GSL_10:
                LODSB                           ;Pick off one character
                OR      AL,AL                   ;Is the character a zero ?
                JZ      GSL_20                  ;Yes, return w/ length = CX
                INC     CX                      ;No... bump character counter
                JMP     GSL_10                  ;Get another character
GSL_20:
                POPF                            ;Restore registers
                POP     SI
                POP     AX
                RET
GET_STRING_LEN  ENDP

;----------------------------------------------------------------------
; Get current cursor position
; Input - Nothing
; Output - DX contains current cursor position
; Changes - DX
;----------------------------------------------------------------------
GET_CURSOR      PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Push registers
                PUSH    BX
                MOV     BH,0                    ;First page
                MOV     AH,3
                INT     10H                     ;DOS call
                POP     BX                      ;Restore registers
                POP     AX
                RET
GET_CURSOR      ENDP

;--------------------------------------------------------------------
; Places cursor in the DX position.
; Input - DX
; Output - Cursor placement
; Changes - Nothing
;--------------------------------------------------------------------
PUT_CURSOR      PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Push registers
                PUSH    BX
                MOV     AH,2
                MOV     BH,0                    ;First page
                INT     10H                     ;DOS call
                POP     BX                      ;Restore registers
                POP     AX
                RET
PUT_CURSOR      ENDP

;--------------------------------------------------------------------
; Inputs a character string from the keyboard.  Characters will be
; accepted until the user presses the return, ESC or CX chars. have
; been received from keyboard.
; Input - ES:DI Points to the buffer that is to receive the string.
;         CX    Contains the length of the input buffer.
; Output -CX    Contains the actual length of the character string.
;         AX    Contains the last keystroke
;         The specified buffer contains the character string that was
;         entered.
;---------------------------------------------------------------------
GETSTRING       PROC    NEAR
            ASSUME      CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    BX
                PUSH    DI                      ;Store Registers
                PUSH    DX
                PUSHF
                CLD                             ;Direction Flag Forward
                XOR     BH,BH                   ;Screen page 0
                XOR     DX,DX                   ;Reset character counter
GS_10:
                XOR     AH,AH                   ;Function 0 -INT16H- Read key
                INT     16H                     ;Read character from keys
                CMP     AL,CR                   ;Carriage Return ?
                JE      GS_30                   ;Yes
                CMP     AL,ESC_KEY              ;Escape key ?
                JE      GS_30                   ;Yes
                CMP     AL,BACKSPACE            ;Back space ?
                JNE     GS_14                   ;No
                CMP     DX,0                    ;Yes, but 1st char. ?
                JE      GS_10                   ;Yes, so skip BS & get a key
GS_14:
                STOSB                           ;No... Store Character
                INC     DX                      ;Bump character counter
                MOV     AH,14                   ;Function 14 -INT10H-
                INT     10H                     ;Display character
                CMP     AL,BACKSPACE            ;Was last char a backspace ?
                JNE     GS_20                   ;No
GS_15:
                SUB     DX,2                    ;Decrement char count twice
                SUB     DI,2                    ;Move PTR back 2 spaces
                ADD     CX,2                    ;Move in two spaces
                MOV     AH,14                   ;Display character interrupt
                MOV     AL,SPACE                ;Space
                INT     10H                     ;Display character
                MOV     AH,14                   ;Display character interrupt
                MOV     AL,BACKSPACE            ;Backspace
                INT     10H                     ;Display character
GS_20:
                LOOP    GS_10                   ;Size = # of characters
GS_30:
                MOV     CX,DX                   ;Save # of chars. in CX
                POPF                            ;Restore Registers
                POP     DX
                POP     DI
                POP     BX
                RET
GETSTRING       ENDP

;------------------------------------------------------------
; Input - DS:SI must point to the 1st char of the string to be
;         sent to the host computer.
; Output - The entire string is sent to the host computer until
;          a ZERO (end of string) is found.
; Changes - SI,AL
;------------------------------------------------------------
SEND_STRING     PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSHF                           ;Store Direction Flag
                CLD
SZ_10:
                LODSB                           ;Get AL
                CMP     AL,0                    ;Done with string ?
                JE      SZ_EXIT                 ;Yes...
                CALL    GET_SEND_CODE           ;No...get & send scan
                JMP     SZ_10                   ;Repeat until done
SZ_EXIT:
                POPF                            ;Restore Direction Flag
                RET
SEND_STRING     ENDP

;------------------------------------------------------------
; Input - AL contains the ASCII code of the character to be
;         sent to the host computer.
; Output - AX is sent to the host computer where
;               AL is ASCII code
;               AH is the associated scan code for the ASCII code
; Changes - Nothing
;------------------------------------------------------------
GET_SEND_CODE   PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Store Registers
                PUSH    BX
                PUSH    CX
                PUSH    SI
                PUSHF
                CMP     AL,'a'                  ;Small Case letter ?
                JL      GSC_10                  ;No ....
                CMP     AL,'z'
                JG      GSC_10                  ;No ....
                SUB     AL,32D                  ;Yes ....
GSC_10:
                MOV     BL,AL                   ;Save ASCII code
                MOV     CX,0                    ;Reset counter into var.
                MOV     SI,OFFSET ASCII_CODE    ;DS:SI -> ASCII variable
                CLD                             ;Forward
GSC_20:
                CMP     CX,70D                  ;Invalid ASCII ?
                JE      GSC_EXIT                ;Yes...
                INC     CX                      ;Bump counter
                LODSB                           ;AL = next char in var.
                CMP     AL,BL                   ;Find correct offset ?
                JNE     GSC_20                  ;No...
GSC_30:                                         ;Yes... so bump that far
                MOV     SI,OFFSET SCAN_CODE     ;into scan_code var and
                CLD                             ;pick off that scan_code.
GSC_40:
                LODSB
                LOOP    GSC_40
GSC_50:
                MOV     AH,AL                   ;AH = scan-code
                MOV     AL,BL                   ;AL = ASCII code
                CALL    SEND_KEYSTROKE          ;Send AX
GSC_EXIT:
                POPF                            ;Restore registers
                POP     SI
                POP     CX
                POP     BX
                POP     AX
                RET
GET_SEND_CODE   ENDP

;-----------------------------------------------------------------
; Execute ZCOPY (upload/download) or shell to DOS and return to
; the calling program.
;-----------------------------------------------------------------
EXEC_ZCOPY      PROC    NEAR
                CMP     NOEXEC,0FFH
                JNZ     EZ_10
                JMP     EZ_EXIT
EZ_10:
                PUSH    AX                      ;Store all registers
                PUSH    BX
                PUSH    CX
                PUSH    DX
                PUSH    SI
                PUSH    DI
                PUSH    DS
                PUSH    ES
                PUSH    BP
                PUSH    CS
                PUSH    CS
                POP     DS
                POP     ES
                CMP     EXEC_VAL,1D             ;Shelling to DOS ?
                JE      EZ_20                   ;Yes, display shell prompt
                CMP     EXEC_VAL,2D             ;Downloading from host ?
                JE      EZ_30                   ;Yes, display transfer prompt
                CMP     EXEC_VAL,3D             ;Uploading from host ?
                JE      EZ_30                   ;Yes, display transfer prompt
                JMP     EZ_60                   ;EXEC_VAL has wrong value !
EZ_20:
                MOV     AX,OFFSET NUL_TAIL      ;Shell to DOS
                MOV     SI,OFFSET SHELL_PROMPT  ;Notify user of shell
                JMP     EZ_40
EZ_30:
                MOV     AX,OFFSET UPDNLOAD_COM    ;Uploading
                MOV     SI,OFFSET TRANSFER_PROMPT ;Notify transferring
EZ_40:
                MOV     PAR_BLK+2,AX            ;EXEC
                CALL    CLEAR_SCREEN            ;Clear local screen
                CALL    GET_STRING_LEN          ;Get string length of DS:SI
                PUSH    ES                      ;Store registers & flags
                PUSH    DI
                PUSHF                           ;Store flag
                MOV     AX,VIDEO_SEGMENT        ;AX->1st char of video segment
                MOV     ES,AX                   ;ES->1st char of video segment
                XOR     DI,DI                   ;ES:DI-> 1st char of video seg
                CLD                             ;Forward in video segment
EZ_50:
                LODSB                           ;Pick off char in string
                CALL    PUT_VIDEO_DATA          ; and place in video segment
                MOV     AL,70H                  ;Place reverse video
                CALL    PUT_VIDEO_DATA          ; in video segment
                LOOP    EZ_50                   ;All characters displayed ?
                POPF                            ;Restore flags
                POP     DI                      ;Restore registers
                POP     ES
                MOV     DH,2D                   ;Row 2
                XOR     DL,DL                   ;Col 0
                CALL    PUT_CURSOR              ;Place cursor on screen
                MOV     CX,CURSOR_SHAPE         ;Restore cursor shape
                MOV     AH,1D                   ;  that we had on entry
                INT     10H                     ;  to manned program
                CALL    SET_ENV
                CMP     NOEXEC,0FFH
                JZ      EZ_60

        if enable_uart_fifos
                ; disable the RX and TX FIFO's (only applicable to 16550A)
                ; in case another program only expects an 8250 UART
                ;
                MOV     DX,COMM_PORT
                ADD     DX,2                    ;FIFO control register
                MOV     AL,0
                OUT     DX,AL
        endif

                ; resend the previous keyboard shift status
                ;
                ; this should undo the ALT-down state that
                ; we sent before we realized that the user
                ; wanted to shell to DOS or transfer a file,
                ;
                ; otherwise a user on the host PC will be confused
                ;
                MOV     AH,1                    ; Put into out buff
                MOV     AL,0FDH                 ; Sync, expect shift
                CALL    PUT_BUFF_DATA           ; send to host PC
                MOV     AL,PREV_SHIFT_STS       ; Get previous shift status
                CALL    PUT_BUFF_DATA           ; Heeere's the shift

                ; wait for the output buffer to be empty
                ; in case we are shelling to a comm program
                ;
                MOV     AH,1
                CALL    WAIT_BUFF_EMPTY

                PUSH    DS
                PUSH    ES
                MOV     OLD_SS,SS
                MOV     OLD_SP,SP
                MOV     BX,OFFSET PAR_BLK
                LDS     DX,CS:COMSPC
                MOV     AX,4B00H
                INT     21H
;
; return from DOS exec
;
                MOV     SS,CS:OLD_SS
                MOV     SP,CS:OLD_SP
                MOV     ES,CS:PAR_BLK
                MOV     AH,49H
                INT     21H
                POP     ES
                POP     DS
EZ_60:
                POP     BP
                POP     ES
                POP     DS
                POP     DI
                POP     SI
                POP     DX
                POP     CX
                POP     BX
                POP     AX
EZ_EXIT:
                RET
EXEC_ZCOPY      ENDP

;-----------------------------------------------------------------------
;  Find the string pointed to by DS:SI in the segment pointed to by ES
;      ES:0000 is assumed to contain null terminated stings.
;  Input - ES points to environment segment
;          SI points to search string
;  Output - Zero set - could not locate string
;           Zero reset - found the string, ES:DI points to end of string
;                        found.  DI is also count of characters up to
;                        end of string.
;  Changes -
;-----------------------------------------------------------------------
FIND_SI         PROC    NEAR
                XOR     DI,DI                   ;Zero environment pointer
FNDSI1:         MOV     BX,SI                   ;Save the source string
                CMP     BYTE PTR ES:[DI],00     ;End of environment space?
                JNZ     FNDSI2                  ; no
                RET                             ; yes, return with zero set
FNDSI2:
                MOV     AL,[BX]                 ;Get source string character
                OR      AL,AL                   ;End of string?
                JZ      FNDSI4                  ; yes
                CMP     AL,ES:[DI]              ; no, Same as environment char?
                JNZ     FNDSI3                  ; no, string doesn't match
                INC     BX                      ; yes, check the next char
                INC     DI                      ; in both strings
                JMP     FNDSI2
FNDSI3:
                XOR     AL,AL                   ;Null string terminator
                MOV     CX,-0001                ;Make sure to find string end
                CLD
                REPNZ   SCASB                   ;Find the end of this string
                JMP     FNDSI1                  ;Test the next string
FNDSI4:
                XOR     AX,AX                   ;Located the string
                INC     AX                      ;Reset the zero flag
                RET
FIND_SI         ENDP

;-----------------------------------------------------------------
;       SET NEW ENVIRONMENT
;
;-----------------------------------------------------------------

SET_ENV         PROC    NEAR
                PUSH    DS
                PUSH    ES
                MOV     ES,DS:[002CH]           ;Environment segment pointer
                MOV     SI,OFFSET PMTSTR        ;Pointer to "PROMPT="
                CALL    FIND_SI                 ;Is "PROMPT=" in environment?
                JNZ     GOTPMT                  ; yes
                XOR     DI,DI                   ; no, set length to 0
;
;  DI = number of characters in environment to end of "PROMPT=" string.
;
GOTPMT:
                MOV     BX,DI                   ;Save the count
                XOR     DI,DI
                MOV     AX,DI
                MOV     CX,0FFFFH
                CLD
                REPNZ   SCASW                   ;Locate the end of the string
                MOV     AX,PMTLGT
                ADD     AX,DI
                CMP     BX,0000
                JNZ     PMTHER
                ADD     AX,(OFFSET PCRSTR) - (OFFSET DFTPMT)
PMTHER:
                MOV     CL,04
                SHR     AX,CL
                INC     AX
                SUB     DI,BX
                PUSH    DI

                PUSH    BX
                MOV     BX,AX                   ;number of paragraphs request
                MOV     AH,48H                  ;malloc
                INT     21H
                JNC     NOERR
                MOV     NOEXEC,0FFH

NOERR:          MOV     PAR_BLK,AX              ;segment to malloc
                MOV     ES,AX                   ;segment to malloc
                XOR     DI,DI
                CLD
                POP     CX
                CMP     CX,0000
                JNZ     DOITAL
                MOV     CX,PMTLGT
                MOV     WORD PTR PMTLGT,0000
                ADD     CX,(OFFSET PCRSTR) - (OFFSET DFTPMT)
                MOV     SI,OFFSET DFTPMT
                REP     MOVSB
                MOV     SI,OFFSET GODPMT
                MOV     CX,(OFFSET PMTSTR) - (OFFSET GODPMT)
                REP     MOVSB
;
;       FROM ADD TO USERS PROMPT
;
DOITAL:
                MOV     AX,DS:[002CH]
                MOV     DS,AX
                XOR     SI,SI
                JCXZ    PMTDON
                REP     MOVSB
;
;       IF NO PROMPT=', NEW PROMPT STORED AND WHOLE ENV STORED NEXT
;
PMTDON:
                PUSH    DS
                PUSH    SI
                PUSH    CS
                POP     DS
                MOV     SI,OFFSET PCRSTR
                MOV     CX,PMTLGT
                JCXZ    NOPMT
                REP     MOVSB
;
;       ANY PROMPT STORED?
;
NOPMT:
                POP     SI
                POP     DS
                POP     CX
                REP     MOVSB
                POP     ES
                POP     DS
                RET
SET_ENV         ENDP

;----------------------------------------------------------------------
GET_COMSPEC     PROC    NEAR
                PUSH    ES
                MOV     PAR_BLK+04,DS
                MOV     AX,DS:[002CH]
                MOV     PAR_BLK,AX
                MOV     ES,AX
                MOV     SI,OFFSET COM_VAR
                CALL    FIND_SI
                MOV     CS:COM_SEG,ES
                MOV     CS:COM_OFS,DI
                JNZ     NO_FND
                MOV     NOEXEC,0FFH
NO_FND:
                POP     ES
                RET
GET_COMSPEC     ENDP

;----------------------------------------------------------------------
; Switch ES and DS
; Input - Nothing
; Output - ES is in DS and DS is in ES
; Changes - ES, DS
;----------------------------------------------------------------------
SWITCH_DS_ES    PROC    NEAR
                PUSH    ES                      ;Push ES
                PUSH    DS                      ;Push DS
                POP     ES                      ;Switch ES
                POP     DS                      ;with DS
                RET
SWITCH_DS_ES    ENDP

;----------------------------------------------------------------------
; Put the keystroke data into the output buffer to be sent to unattended
; system
; Input - AX - keystroke data
; Output - Nothing
; Changes - Nothing
;----------------------------------------------------------------------
SEND_KEYSTROKE  PROC    NEAR
        ASSUME  CS:CSEG,DS:CSEG,ES:NOTHING,SS:NOTHING
                PUSH    AX                      ;Save it
                MOV     AH,1                    ;Use the output buffer
                MOV     AL,0FEH                 ;Sync byte, to expect key data
                CALL    PUT_BUFF_DATA           ;Put AL into the output buffer
                POP     AX                      ;Get AL back
                PUSH    AX                      ;Save AH
                MOV     AH,1                    ;Use the output buffer
                CALL    PUT_BUFF_DATA           ;Put ASCII code in out buffer
                POP     AX                      ;Get AH back
                PUSH    AX                      ;Save keystroke
                MOV     AL,AH                   ;Move it to AL
                MOV     AH,1                    ;Use the output buffer
                CALL    PUT_BUFF_DATA           ;Put scan code in out buffer
                POP     AX                      ;Restore keystroke
                RET
SEND_KEYSTROKE  ENDP

;======================================================================
LAST_BYTE       EQU     $
CSEG            ENDS
                PUBLIC  START                   ;so mapsym doesn't bitch
                END     START
